# Copyright 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from confluent_kafka import KafkaError
from confluent_kafka.avro import AvroConsumer
from confluent_kafka.avro.serializer import SerializerError
from json import dumps as json_dumps, loads as json_loads
from netq_lib.common import configuration
from netq_apps.lib.netq import NetQ
from netq_apps.cmd.netq import NetqCLI
from netq_apps.modules.core.os.ntp import cli_check_ntp,ntp_check_report # noqa
from netq_apps.modules.core.l3.bgp import cli_check_bgp,bgp_check_report # noqa
from netq_apps.modules.core.l3.mtu import cli_check_mtu,mtu_check_report # noqa
from netq_apps.modules.core.l3.ospf import cli_check_ospf,ospf_check_report # noqa
from netq_apps.modules.core.nvo.evpn import cli_check_evpn,evpn_check_report # noqa
from netq_apps.modules.core.l2.clag import cli_check_clag,clag_check_report # noqa
from netq_apps.modules.core.l2.vlan import cli_check_vlan,vlan_check_report # noqa
from netq_apps.modules.core.trace.trace import cli_trace_path # noqa
from netq_apps.modules.core.platform.cables import cli_check_physical_links,cable_check_report #noqa
from netq_apps.modules.core.platform.inventory import cli_check_version, version_check_report # noqa
from netq_apps.modules.core.platform.sensors import cli_check_sensors,sensors_check_report # noqa
from netq_apps.modules.core.netq.agents import cli_check_agents,agent_check_report # noqa
from netq_apps.modules.core.nvo.vxlan import cli_check_vxlan,vxlan_check_report # noqa
from netq_lib.common.auth import GatewayAuth
import sys
import os
from io import StringIO

from confluent_kafka import avro
from confluent_kafka.avro import AvroProducer
import multiprocessing
import traceback
import time
import netq_lib.orm.avro.tc_response as tc_response
from netq_lib.common import utils
from network_docopt import NetworkDocopt


class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend("".join(self._stringio.getvalue()))
        sys.stdout = self._stdout


class NetQKafkaBridge(object):
    MSG_TYPE_TRACE = 0
    MSG_TYPE_VALIDATION = 1

    def __init__(self, conf_dict):
        self.logger = utils.get_logger("netq-cli", conf_dict)
        self.logger.info(conf_dict)
        self.bootstrap_server = os.environ['KAFKA_BOOTSTRAP_SERVERS']
        self.schema_registry_url = \
            "http://" + os.environ['SCHEMA_REGISTRY_URL']
        if ('TC_INPUT_TOPIC' in os.environ):
            self.input_topic = os.environ['TC_INPUT_TOPIC']
        else:
            self.input_topic = "netq_tc_request"
        if ('TC_OUTPUT_TOPIC' in os.environ):
            self.output_topic = os.environ['TC_OUTPUT_TOPIC']
        else:
            self.output_topic = "netq_tc_response"

        if (('netq-cli' in conf_dict) and ('server' in conf_dict['netq-cli'])):
            self.gateway_ip = conf_dict['netq-cli']['server']
        else:
            self.gateway_ip = os.environ['NETQ_API_GATEWAY_SERVICE_HOST']
        if (('netq-cli' in conf_dict) and ('port' in conf_dict['netq-cli'])):
            self.gateway_port = conf_dict['netq-cli']['port']
        else:
            self.gateway_port = os.environ['NETQ_API_GATEWAY_SERVICE_PORT']

        self.valid_check_functions = ["bgp", "evpn", "clag", "interfaces",
                                      "agents", "ntp", "sensors",
                                      "vxlan", "vlan", "mtu", "ospf", "version"]

    def kafka_send_msg_to_kafka(self, opid, rc, job_status, resJson, msg):
        respStr = ""
        respStr += ' '.join(json_dumps("".join(list(resJson)))
                            .strip('"').split())

        job_details = {"opid": opid, "jobid": msg.value()['jobid'],
                       "job_req_time": msg.value()['job_req_time'],
                       "job_start_time": msg.value()['job_start_time'],
                       "job_end_time": int(time.time() * 1000),
                       "job_status": job_status,
                       "status_details": str(rc),
                       "is_scheduled": msg.value()['is_scheduled'],
                       "request_type": msg.value()['request_type'],
                       "user_name": msg.value()['user_name'],
                       "params": msg.value()['params']}
        value_schema = avro.loads(tc_response.value)
        key_schema = avro.loads(tc_response.key)
        value = {"job_request": job_details, "response": respStr}
        key = {"opid": msg.value()['opid']}

        avro_producer = AvroProducer({
            'bootstrap.servers': self.bootstrap_server,
            'schema.registry.url': self.schema_registry_url
        }, default_key_schema=key_schema, default_value_schema=value_schema)
        avro_producer.produce(topic=self.output_topic, value=value, key=key)
        avro_producer.flush()
        print((value, key))

    def validate_and_doc_opt(self, params, msg_type, auth, opid):
        cli = NetqCLI(False, self.logger, auth, opid)
        testDict = {"agents": "agent",
                    "license": "lic",
                    "interfaces": "cable",
                    "mlag": "clag"}
        print(("Params: {}".format(params)))
        if msg_type == self.MSG_TYPE_TRACE:
            reqJson = ["netq", "trace", params['dst']]
            if int(params['vlan']) > 0:
                reqJson.extend(["vlan", params['vlan']])
            reqJson.extend(["from", params['src']])
            if params['vrf'] != "":
                reqJson.extend(["vrf", params['vrf']])
            reqJson.extend(["json"])
        elif msg_type == self.MSG_TYPE_VALIDATION:
            if (('modular_test_config' in params) and (params['modular_test_config'][0].get("modular_test_info") != None)):
                val_type = params['validation_type']
                # Get the list of modular tests to be run for this validation
                test_req = [i.get('name').lower() for i in params['modular_test_config'][0].get("modular_test_info")]
                if (val_type in testDict):
                    val_type = testDict.get(val_type)
                # check_report.tests will provide all the tests supported by respective validation
                tests = globals()["{}_check_report".format(val_type)].tests
                # Compare desired tests and retrive test ids from check_report
                test_parms_to_enable = ','.join([str(i.get('id')) for i in tests if (i.get('name').lower() in test_req)])
                # Pass comma seperated ids as input
                reqJson = ["netq", "check", ("cl-version" if (params['validation_type'] == "version") else params['validation_type'])]
                if (('hostnames' in params) and (len(params['hostnames']) > 0)):
                    reqJson.extend(['hostnames', params['hostnames']])
                if (('check_filter_id' in params) and (len(params['check_filter_id']) > 0)):
                    reqJson.extend(['check_filter_id', params['check_filter_id']])
                reqJson.extend(["include", test_parms_to_enable])
            else:
                reqJson = ["netq", "check", ("cl-version" if (params['validation_type'] == "version") else params['validation_type'])]
                if (('hostnames' in params) and (len(params['hostnames']) > 0)):
                    reqJson.extend(['hostnames', params['hostnames']])
                if (('check_filter_id' in params) and (len(params['check_filter_id']) > 0)):
                    reqJson.extend(['check_filter_id', params['check_filter_id']])
            reqJson.extend(["json"])
        print(("Params: {}".format(params)))
        NetworkDocopt.run(cli, reqJson)
        return cli.args

    def execute(self, msg_type, params, reqJson, netq):
        if msg_type == self.MSG_TYPE_TRACE:
            rc = cli_trace_path(reqJson, netq)
        elif msg_type == self.MSG_TYPE_VALIDATION:
            validation_type = params['validation_type']
            if validation_type in self.valid_check_functions:
                if validation_type == "interfaces":
                    rc = cli_check_physical_links(reqJson, netq)
                else:
                    rc = globals()["cli_check_{}".format(validation_type)](reqJson, netq)
            else:
                rc = 1
        return rc

    def consume(self, conf_dict):
        consumer = AvroConsumer({
            'bootstrap.servers': self.bootstrap_server,
            'group.id': 'group-1',
            'schema.registry.url': self.schema_registry_url})
        consumer.subscribe([self.input_topic])
        while True:
            try:
                msg = consumer.poll(10)
            except SerializerError as e:
                print(("Message deserialization failed for {}: {}".
                      format(msg, e)))
                break

            if (msg is None):
                continue

            if msg.error():
                if msg.error().code() == KafkaError._PARTITION_EOF:
                    continue
                elif msg.error().code() == KafkaError._MAX_POLL_EXCEEDED:
                    print(("Previous message took long time:{}"
                          .format(msg.error())))
                    continue
                else:
                    print((msg.error()))
                    break

            if (msg.value()['params'] is None):
                continue

            opid = msg.value()['opid']
            auth = GatewayAuth(logger=self.logger,
                               gateway_ip=self.gateway_ip,
                               gateway_port=self.gateway_port,
                               opid=opid)
            job_status = ""
            resJson = ""
            rc = 0

            if auth.conn is not None:
                msg_type = msg.value()['request_type']
                params = json_loads(msg.value()['params'])
                reqJson = self.validate_and_doc_opt(params, msg_type, auth, opid)
                start_time = None
                #if msg.value()['is_scheduled'] is False:
                #    start_time = msg.value()['job_req_time'] / 1000
                netq = NetQ(None, start_time, None, self.logger, None)
                print(("NetworkDocOpt: {}".format(reqJson)))

                try:
                    with Capturing() as resJson:
                        rc = self.execute(msg_type, params, reqJson, netq)
                    job_status = "Complete"
                except Exception:
                    rc = 1
                    job_status = "Complete with exceptions"
                    traceback.print_exc()
            else:
                rc = 1
                if job_status == "":
                    job_status = "Opid(" + str(opid) + ") Unauthorized"

            self.kafka_send_msg_to_kafka(opid, rc, job_status, resJson, msg)

            msg = None

        consumer.close()


if __name__ == "__main__":
    procs = 1
    jobs = []
    conf_dict = configuration.get_config("netq-cli")
    for i in range(0, procs):
        kafka = NetQKafkaBridge(conf_dict)
        process = multiprocessing.Process(target=kafka.consume,
                                          args=(conf_dict,))
        jobs.append(process)

    for j in jobs:
        j.start()

    for j in jobs:
        j.join()

    print("Exiting TraceScheduler")
