# Copyright 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    netq add validation type (ntp | interfaces | license | sensors | evpn | vxlan | agents | mlag | vlan | bgp | mtu | ospf) [alert-on-failure]
    netq add validation name <text-new-validation-name> type (ntp | interfaces | license | sensors | evpn | vxlan | agents | mlag | vlan | bgp | mtu | ospf) interval <text-time-min> [alert-on-failure]
    netq del validation <text-validation-name>
    netq show validation settings [name <text-validation-name>] [type ntp | type interfaces | type license | type sensors | type evpn | type vxlan | type agents | type mlag | type vlan | type bgp | type mtu | type ospf] [json]
    netq show validation summary [name <text-validation-name>] type (ntp | interfaces | license | sensors | evpn | vxlan | agents | mlag | vlan | bgp | mtu | ospf) [around <text-time-hr>] [json]

Options:
    validation                  : Schedule a validation check
    interval                    : Frequency which a validation occurs in minutes
    alert-on-failure            : Raise a notification when a failure has occurred
    settings                    : Validation configuration settings
    name                        : Validation name
    summary                     : Summary of validation results
    <text-new-validation-name>  : Name of new validation
    <text-validation-name>      : Name of existing validation
    <text-time-min>             : Cadence time format in minutes such as 1m, 15m, 30m ...
    <text-time-hr>              : Time format in hours such as 1h, 2h, 3h ...
"""


import time
import requests
import json
from netq_lib.common import utils
from netq_lib.common.enums import (
    HeaderLength,
)
from netq_apps.modules import NetqModule, RC_SUCCESS
from netq_apps.cmd.netq import netq_tabulate
from netq_apps.lib.completers import wrapMatcher, wrapCompleter
from netq_lib.orm.redisdb.models import Validation
app = NetqModule()
ON_DEMAND = -1
MILLISECONDS = 1000


@app.route('add validation')
def cli_add_validation(cli, netq):
    name = cli.get('<text-new-validation-name>')

    if cli.get('<text-multiple-types>'):
        validation_type = cli.get('<text-multiple-types>').split(',')
    else:
        validation_type = cli.get('ntp') or cli.get('interfaces') or cli.get('license') or cli.get('sensors') or \
                          cli.get('evpn') or cli.get('vxlan') or cli.get('agents') or \
                          cli.get('mlag') or cli.get('vlan') or cli.get('bgp') or cli.get('mtu') or cli.get('ospf')
        validation_type = [validation_type]

    interval = cli.get('<text-time-min>')
    alert = 1 if cli.get('alert-on-failure') else 0

    if name:
        try:
            response = Validation.query.post_request(path='schedule', validation_name=name,
                                                     validation_type=validation_type,
                                                     cadence_min=utils.time_format(interval, 'm'),
                                                     alert_on_failure=alert,
                                                     start_time=int(time.time() * MILLISECONDS), action='Add')
        except requests.exceptions.HTTPError as ex:
            print((utils.RED_T('Failed to add scheduled validation')))
            raise ex
        response_json = response.json()
        if isinstance(response_json, dict) and len(response_json) and response_json.get('Error'):
            raise Exception(response_json.get('Error'))
        else:
            print((utils.COLOR_GREEN('Successfully added {} running every {}'.format(name, interval))))
    else:
        try:
            response = Validation.query.post_request(path='schedule', validation_type=validation_type,
                                                     alert_on_failure=alert, start_time=ON_DEMAND, action='Add')
        except requests.exceptions.HTTPError as ex:
            print((utils.RED_T('Failed to add on-demand validation')))
            raise ex
        else:
            response_json = response.json()
            if response_json.get('Error'):
                raise Exception(response_json.get('Error'))
            else:
                for job in response_json.get('data'):
                    print((utils.COLOR_GREEN('Running job {} type {}'.format(job.get('jobid'),
                                                                            job.get('validation_type')))))

    return RC_SUCCESS


@app.route('del validation')
def cli_del_validation(cli, netq):
    name = cli.get('<text-validation-name>')
    response = Validation.query.post_request(path='schedule', validation_name=name, action='Delete')
    if response:
        response_json = response.json()
        if isinstance(response_json, list) and len(response_json) and response_json[0].get('Error'):
            raise Exception(response_json[0].get('Error'))
    print((utils.COLOR_GREEN('Successfully deleted validation {}'.format(name))))
    return RC_SUCCESS


@app.route('show validation')
def cli_show_validation(cli, netq):
    if cli.get('settings'):
        return validation_settings(cli, netq)
    else:
        name = cli.get('<text-validation-name>')
        validation_type = cli.get('ntp') or cli.get('interfaces') or cli.get('license') or cli.get('sensors') or \
                          cli.get('evpn') or cli.get('vxlan') or cli.get('agents') or \
                          cli.get('mlag') or cli.get('vlan') or cli.get('bgp') or cli.get('mtu') or cli.get('ospf')
        duration = utils.time_format(cli.get('<text-time-hr>', '24h'), 'h')
        output = []
        col_headers = []
        if cli.get('summary'):
            if cli.get('json'):
                json_out = dict()
                result = []
                json_response = Validation.query.request(json=True, by='scheduled', proto=validation_type,
                                                         time=int(time.time()), duration=duration)
                if json_response:
                    for entry in json_response.get('data', []):
                        if name and name != entry.get('validation_label'):
                            continue
                        result.append(entry.get('summary'))
                json_out['scheduled'] = result

                if name is None:
                    json_response = Validation.query.request(json=True, by='ondemand', proto=validation_type,
                                                             time=int(time.time()), duration=duration)
                    if json_response:
                        json_out['on-demand'] = [entry.get('summary') for entry in json_response.get('data', [])]
                    else:
                        json_out['on-demand'] = []
                print((json.dumps(json_out, sort_keys=True, indent=4)))
                return RC_SUCCESS

            for entry in Validation.query.request(by='scheduled', proto=validation_type,
                                                  time=int(time.time()), duration=duration):
                if name and entry.validation_label != name:
                    continue
                summary = entry.summary
                output.append([entry.validation_label, 'scheduled', entry.jobid, summary.get('checkedNodeCount'),
                               summary.get('failedNodeCount'),
                               summary.get('totalNodeCount') if 'totalNodeCount' in list(summary.keys()) else
                               summary.get('totalSessionCount'),
                               utils.absolute_date(entry.res_timestamp)])

            if name is None:
                for entry in Validation.query.request(by='ondemand', proto=validation_type,
                                                      time=int(time.time()), duration=duration):
                    summary = entry.summary
                    output.append(['', 'on-demand', entry.jobid, summary.get('checkedNodeCount'),
                                   summary.get('failedNodeCount'),
                                   summary.get('totalNodeCount') if 'totalNodeCount' in list(summary.keys()) else
                                   summary.get('totalSessionCount'),
                                   utils.absolute_date(entry.res_timestamp)])

            col_headers = ['Name', 'Type', 'Job ID', 'Checked Nodes', 'Failed Nodes',
                           'Total Nodes', 'Timestamp']
        elif cli.get('results'):
            if cli.get('json'):
                json_out = dict()
                result = []
                json_response = Validation.query.request(json=True, by='scheduled', proto=validation_type,
                                                         time=int(time.time()), duration=duration)
                if json_response:
                    for entry in json_response.get('data', []):
                        if name and name != entry.get('validation_label'):
                            continue
                        result.append(entry)
                json_out['scheduled'] = result

                if name is None:
                    json_response = Validation.query.request(json=True, by='ondemand', proto=validation_type,
                                                             time=int(time.time()), duration=duration)
                    if json_response:
                        json_out['on-demand'] = json_response.get('data', [])
                    else:
                        json_out['on-demand'] = []

                print((json.dumps(json_out, sort_keys=True, indent=4)))
                return RC_SUCCESS

            for entry in Validation.query.request(by='scheduled', proto=validation_type,
                                                  time=int(time.time()), duration=duration):
                if name and entry.validation_label != name:
                    continue
                if entry.failed_node_set:
                    for node in entry.failed_node_set:
                        output.append([entry.validation_label, 'scheduled', entry.jobid, node.get('hostname'),
                                       get_message(node),
                                       utils.absolute_date(entry.res_timestamp)])
                else:
                    output.append([entry.validation_label, 'scheduled', entry.jobid, '', '',
                                   utils.absolute_date(entry.res_timestamp)])
            if name is None:
                for entry in Validation.query.request(by='ondemand', proto=validation_type,
                                                      time=int(time.time()), duration=duration):
                    if entry.failed_node_set:
                        for node in entry.failed_node_set:
                            output.append(['', 'on-demand', entry.jobid, node.get('hostname'),
                                           get_message(node),
                                           utils.absolute_date(entry.res_timestamp)])
                    else:
                        output.append(['', 'on-demand', entry.jobid, '', '', utils.absolute_date(entry.res_timestamp)])
            col_headers = ['Name', 'Type', 'Job ID', 'Failed Nodes', 'Message', 'Timestamp']

        col_sizes = [HeaderLength.get_col_size(header) for header in col_headers]
        if len(output):
            netq_tabulate(output=output, headers=col_headers, colsizes=col_sizes, jsonify=False)
        else:
            print('\nNo matches found')
        return RC_SUCCESS


def get_message(session):
    output = []
    for key in session:
        if key == 'hostname':
            continue
        output.append('{}: {}'.format(key, session.get(key)))
    return ', '.join(output)


def validation_settings(cli, netq):
    name = cli.get('<text-validation-name>')
    validation_type = cli.get('ntp') or cli.get('interfaces') or cli.get('license') or cli.get('sensors') or \
                      cli.get('evpn') or cli.get('vxlan') or cli.get('agents') or cli.get('mlag') or \
                      cli.get('vlan') or cli.get('bgp') or cli.get('mtu') or cli.get('ospf')

    if cli.get('json'):
        result = []
        json_response = Validation.query.request(json=True, path='schedule', type='schedule')
        if json_response:
            for entry in json_response.get('data', []):
                if name and name != entry.get('validation_name') or \
                        validation_type and validation_type not in entry.get('params', dict()).get('validation_type', []):
                    continue
                result.append(entry)
        print((json.dumps(result, sort_keys=True, indent=4)))
        return RC_SUCCESS

    output = []
    for entry in Validation.query.request(path='schedule', type='schedule'):
        if name and name != entry.validation_name or \
                validation_type and validation_type not in entry.params.get('validation_type', []):
            continue
        cadence = '{}m'.format(entry.cadence_min)
        active = 'yes' if entry.is_active == 1 else 'no'
        creation_time = utils.absolute_date(entry.create_timestamp)
        start_time = utils.absolute_date(entry.start_time) if entry.start_time != -1 else creation_time
        output.append([entry.validation_name, __parse_types(entry.params), cadence, start_time, creation_time, active])

    col_headers = ['Name', 'Types', 'Cadence', 'Start Time', 'Creation Time', 'Active']
    col_sizes = [HeaderLength.get_col_size(header) for header in col_headers]

    if len(output):
        netq_tabulate(output=output, headers=col_headers, colsizes=col_sizes, jsonify=False)
    else:
        print('\nNo matches found')
    return RC_SUCCESS


def __parse_types(params):
    return ', '.join(params.get('validation_type', []))


def add_completers(cli):
    cli.add_tag_matcher('<text-validation-name>', name_match, True)
    cli.add_tab_complete_hook('<text-validation-name>', name_complete)


@wrapMatcher
def name_match(argv_word, use_fuzzy, tab_completing=False):
    return [entry.validation_name for entry in Validation.query.request(path='schedule', type='schedule')]


@wrapCompleter
def name_complete(args, ended_with_space, last_argv):
    out = []
    for entry in Validation.query.request(path='schedule', type='schedule'):
        if ended_with_space or entry.validation_name.startswith(last_argv):
            out.append((entry.validation_name, 'Validation Name'))

    if not out:
        out.append(('<text-validation-name>', 'Validation Name'))
    return out
