# Copyright 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    netq add trace <mac> [vlan <1-4096>] from (<src-hostname> | <ip-src>) [vrf <vrf>] [alert-on-failure]
    netq add trace name <text-new-trace-name> <mac> [vlan <1-4096>] from (<src-hostname> | <ip-src>) [vrf <vrf>] interval <text-time-min> [alert-on-failure]
    netq add trace <ip> from (<src-hostname> | <ip-src>) [vrf <vrf>] [alert-on-failure]
    netq add trace name <text-new-trace-name> <ip> from (<src-hostname> | <ip-src>) [vrf <vrf>] interval <text-time-min> [alert-on-failure]
    netq del trace <text-trace-name>
    netq show trace settings [name <text-trace-name>] [json]
    netq show trace summary [name <text-trace-name>] [around <text-time-hr>] [json]
    netq show trace results <text-job-id> [json]

Options:
    interval                    : Frequency which a trace occurs
    alert-on-failure            : Raise a notification when a failure has occurred
    settings                    : Trace configuration settings
    name                        : Trace name
    results                     : Show results of trace or validation scheduler
    <text-time-min>             : Cadence time format in minutes such as 1m, 15m, 30m ...
    <text-time-hr>              : Time format in hours such as 1h, 2h, 3h ...
    <text-trace-name>           : Name of existing trace scheduler
    <text-new-trace-name>       : Name of new trace scheduler
    <text-job-id>               : Job ID of trace or validation
"""


import requests
import time
import json
from netq_lib.common import utils
from netq_lib.common.enums import (
    HeaderLength,
)

from netq_apps.modules import NetqModule, RC_SUCCESS
from netq_apps.cmd.netq import netq_tabulate
from netq_lib.orm.redisdb.models import Trace
from netq_apps.lib.completers import wrapMatcher, wrapCompleter
app = NetqModule()
ON_DEMAND = -1
BUCKET = 1
MILLISECONDS = 1000


@app.route('add trace')
def cli_add_trace(cli, netq):
    name = cli.get('<text-new-trace-name>')
    dst = cli.get('<mac>') or cli.get('<ip>')
    vlan = cli.get('<1-4096>')
    src = cli.get('<src-hostname>') or cli.get('<ip-src>')
    vrf = cli.get('<vrf>')
    interval = cli.get('<text-time-min>')
    alert = 1 if cli.get('alert-on-failure') else 0

    data = {'trace_params': {'src': src, 'dst': dst},
            'alert_on_failure': alert,
            'action': 'Add'}

    if name:
        data['trace_name'] = name
        data['start_time'] = int(time.time() * MILLISECONDS)
        data['cadence_min'] = utils.time_format(interval, 'm')
    else:
        data['start_time'] = ON_DEMAND

    if vrf:
        data['trace_params']['vrf'] = vrf
    if vlan:
        data['trace_params']['vlan'] = int(vlan)

    try:
        for entry in Trace.query.request(json=True, path='schedule', type='schedule'):
            if isinstance(entry, dict) and name == entry.get('trace_name'):
                data['old_trace_name'] = '{}(old)'.format(name)
                data['action'] = 'Update'
                break
        response = Trace.query.post_request(path='schedule', data=[data])
    except requests.exceptions.HTTPError as ex:
        if name:
            print((utils.RED_T('Failed to add scheduled trace {}'.format(name))))
        else:
            print((utils.RED_T('Failed to add on-demand trace')))
        raise ex
    else:
        if name:
            print((utils.COLOR_GREEN('Successfully added/updated {} running every {}'.format(name, interval))))
        else:
            print((utils.COLOR_GREEN('Running job {} src {} dst {}'.format(response.json().get('jobid'), src, dst))))
    return RC_SUCCESS


@app.route('del trace')
def cli_del_trace(cli, netq):
    name = cli.get('<text-trace-name>')
    data = [{'trace_name': name, 'action': 'Delete'}]

    try:
        Trace.query.post_request(path='schedule', data=data)
    except requests.exceptions.HTTPError as ex:
        print((utils.RED_T('Failed to delete scheduled trace {}'.format(name))))
        raise ex
    else:
        print((utils.COLOR_GREEN('Successfully deleted schedule trace {}'.format(name))))
    return RC_SUCCESS


@app.route('show trace settings')
def cli_show_trace_settings(cli, netq):
    name = cli.get('<text-trace-name>')
    output = []

    if cli.get('json'):
        data = Trace.query.request(json=True, path='schedule', type='schedule')
        data = data.get('data', []) if isinstance(data, dict) else data
        for entry in data:
            if isinstance(entry, dict) and name and name != entry.get('trace_name'):
                continue
            else:
                output.append(entry)
        print((json.dumps(output, sort_keys=True, indent=4)))
        return RC_SUCCESS
    else:
        for entry in Trace.query.request(path='schedule', type='schedule'):
            if not hasattr(entry, 'trace_name') or (name and name != entry.trace_name):
                continue
            params = entry.trace_params
            output.append([entry.trace_name, params.get('src', ''), params.get('dst', ''), params.get('vrf', ''),
                           '' if params.get('vlan') == '-1' else params.get('vlan', ''),
                           'yes' if params.get('alert_on_failure') == '1' else 'no',
                           '{}m'.format(entry.cadence_min), 'yes' if entry.is_active else 'no',
                           utils.absolute_date(entry.start_time)])

        col_headers = ['Name', 'Source', 'Destination', 'VRF', 'VLAN', 'Alert on Failure', 'Cadence', 'Active',
                       'Start Time']
        col_sizes = [HeaderLength.get_col_size(header) for header in col_headers]

        if len(output):
            netq_tabulate(output=output, headers=col_headers, colsizes=col_sizes, jsonify=False)
        else:
            print('\nNo matches found')

        return RC_SUCCESS


@app.route('show trace summary')
def cli_show_trace_summary(cli, netq):
    duration = utils.time_format(cli.get('<text-time-hr>', '24h'), 'h')
    name = cli.get('<text-trace-name>')
    output = []

    if cli.get('json'):
        data = Trace.query.request(json=True, path='schedule', type='schedule_result',
                                   time=int(time.time()), duration=duration)
        data = data.get('data', []) if isinstance(data, dict) else data
        for entry in data:
            if isinstance(entry, dict) and name and name != entry.get('trace_name'):
                continue
            else:
                output.append(entry)
        print((json.dumps(output, sort_keys=True, indent=4)))
        return RC_SUCCESS

    for entry in Trace.query.request(path='schedule', type='schedule_result', time=int(time.time()), duration=duration):
        if not hasattr(entry, 'trace_name') or (name and name != entry.trace_name):
            continue
        else:
            output.append([entry.trace_name, entry.jobid, entry.status, entry.status_details,
                           utils.absolute_date(entry.job_start_time),
                           utils.absolute_date(entry.job_end_time)])

    col_headers = ['Name', 'Job ID', 'Status', 'Status Details', 'Start Time', 'End Time']
    col_sizes = [HeaderLength.get_col_size(header) for header in col_headers]
    if len(output):
        netq_tabulate(output=output, headers=col_headers, colsizes=col_sizes, jsonify=False)
    else:
        print('\nNo matches found')
    return RC_SUCCESS


@app.route('show trace results')
def cli_show_trace_result(cli, netq):
    jobid = cli.get('<text-job-id>')

    if cli.get('json'):
        print((json.dumps(Trace.query.request(json=True, path='jobid/{}/l3-route-summary'.format(jobid)),
                         sort_keys=True, indent=4)))
        return RC_SUCCESS
    else:
        output = []
        for entry in Trace.query.request(path='jobid/{}/l3-route-summary'.format(jobid)):
            if not hasattr(entry, 'route'):
                continue
            for id in entry.route:
                data = entry.route.get(id)
                output.append([id, data.get('errors'), data.get('warnings'), data.get('hops'), data.get('mtu')])

        col_headers = ['ID', 'Errors', 'Warnings', 'Hops', 'MTU']
        col_sizes = [HeaderLength.get_col_size(header) for header in col_headers]
        if len(output):
            netq_tabulate(output=output, headers=col_headers, colsizes=col_sizes, jsonify=False)
        else:
            print('\nNo matches found')
        return RC_SUCCESS


def add_completers(cli):
    cli.add_tag_matcher('<text-trace-name>', name_match, True)
    cli.add_tab_complete_hook('<text-trace-name>', name_complete)


@wrapMatcher
def name_match(argv_word, use_fuzzy, tab_completing=False):
    result = []
    for entry in Trace.query.request(path='schedule', type='schedule'):
        if hasattr(entry, 'trace_name'):
            result.append(entry.trace_name)
    return result


@wrapCompleter
def name_complete(args, ended_with_space, last_argv):
    out = []
    for entry in Trace.query.request(path='schedule', type='schedule'):
        if not hasattr(entry, 'trace_name'):
            continue
        if ended_with_space or entry.trace_name.startswith(last_argv):
            out.append((entry.trace_name, 'Trace Name'))

    if not out:
        out.append(('<text-trace-name>', 'Trace Name'))
    return out