# Copyright 2018, 2019, 2020, 2021 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

# GENERIC CONFIGURATION HELPERS




import os
import yaml
from netq_lib.common.enums import DefaultPaths, ConfKw, Platform
from netq_lib.common.utils import get_logger, RED_T, is_cl3, is_cl4
from configobj import ConfigObj

__GATEWAY = None
__AUTH = None


def debugger(output, service=ConfKw.NETQD):
    config = get_config(service)
    logger = get_logger(service, config)
    logger.info(str(output))
    return logger

def is_cli_gateway():
    global __GATEWAY
    if __GATEWAY is None:
        current_config = get_config()
        if current_config.get(ConfKw.NETQ_CLI) and \
                current_config[ConfKw.NETQ_CLI].get(ConfKw.ENDPOINT, ConfKw.GATEWAY) == ConfKw.GATEWAY:
            __GATEWAY = True
        else:
            __GATEWAY = False
    return __GATEWAY


def container_env():
    return 'KAFKA_BOOTSTRAP_SERVERS' in os.environ


def is_opta():
    return os.path.isfile(DefaultPaths.OPTA_FILE)


def is_cloud_appliance():
    conf = ConfigObj('/etc/app-release')
    return conf.get('APPLIANCE_NAME') == "NetQ Cloud Appliance"


def represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')


def update_yaml_configuration(conf_dict, path, preamble=None):
    yaml.add_representer(type(None), represent_none)
    try:
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        with open(path, 'w') as fp:
            if preamble:
                fp.write(preamble)
            yaml.dump(conf_dict, fp, default_flow_style=False, default_style='')
    except PermissionError as ex:
        get_logger('_netq_configuration_').exception(ex)
        print(RED_T("Permission denied: Failed to write configuration"))
        return 1  # Coincidentally RC_FAIL
    except Exception as ex:
        get_logger('_netq_configuration_').exception(ex.message)
        return 1  # Coincidentally RC_FAIL
    return 0


def init_default_yaml_configuration():

    _conf = dict()
    _conf[ConfKw.BACKEND] = dict()
    _conf[ConfKw.BACKEND][ConfKw.VRF] = ConfKw.DEFAULT_VRF

    _conf[ConfKw.NETQ_CLI] = dict()
    _conf[ConfKw.NETQ_CLI][ConfKw.ENDPOINT] = ConfKw.GATEWAY
    _conf[ConfKw.NETQ_CLI][ConfKw.OPID] = ConfKw.DEFAULT_OPID
    _conf[ConfKw.NETQ_CLI][ConfKw.PORT] = ConfKw.DEFAULT_GATEWAY_PORT
    _conf[ConfKw.NETQ_CLI][ConfKw.COUNT] = ConfKw.DEFAULT_BATCH_SIZE
    _conf[ConfKw.NETQ_CLI][ConfKw.VRF] = ConfKw.DEFAULT_VRF
    _conf[ConfKw.NETQ_CLI][ConfKw.API_LOGGING] = ConfKw.DEFAULT_API_LOGGING

    _conf[ConfKw.NETQ_AGENT] = dict()
    _conf[ConfKw.NETQ_AGENT][ConfKw.ENABLE_OPTA_DISCOVERY] = ConfKw.DEFAULT_ENABLE_OPTA_DISCOVERY
    _conf[ConfKw.NETQ_AGENT][ConfKw.IS_OPTA_ENABLED] = ConfKw.DEFAULT_IS_OPTA_ENABLED
    _conf[ConfKw.NETQ_AGENT][ConfKw.IS_REDIS_ENABLED] = ConfKw.DEFAULT_IS_REDIS_ENABLED
    _conf[ConfKw.NETQ_AGENT][ConfKw.AGENT_PORT] = ConfKw.DEFAULT_AGENT_PORT
    _conf[ConfKw.NETQ_AGENT][ConfKw.PORT] = ConfKw.DEFAULT_OPTA_PORT
    _conf[ConfKw.NETQ_AGENT][ConfKw.VRF] = ConfKw.DEFAULT_VRF
    _conf[ConfKw.NETQ_AGENT][ConfKw.SERVER] = ConfKw.DEFAULT_SERVER
    _conf[ConfKw.NETQ_AGENT][ConfKw.CPU_LIMIT] = ConfKw.DEFAULT_CPU_LIMIT
    _conf[ConfKw.NETQ_AGENT][ConfKw.IS_GNMI_ENABLED] = ConfKw.DEFAULT_IS_GNMI_DISABLED
    _conf[ConfKw.NETQ_AGENT][ConfKw.NETQ_STREAM_PORT] = ConfKw.DEFAULT_NETQ_STREAM_PORT
    _conf[ConfKw.NETQ_AGENT][ConfKw.NETQ_STREAM_ADDRESS] = ConfKw.DEFAULT_NETQ_STREAM_ADDRESS
    return _conf


def is_default_backend(config_dict):
    default_config = init_default_yaml_configuration()
    return config_dict[ConfKw.BACKEND] == default_config[ConfKw.BACKEND]


def is_default_netq_cli(config_dict):
    default_config = init_default_yaml_configuration()
    return config_dict[ConfKw.NETQ_CLI] == default_config[ConfKw.NETQ_CLI]


def is_default_netq_agent(config_dict):
    default_config = init_default_yaml_configuration()
    return config_dict[ConfKw.NETQ_AGENT] == default_config[ConfKw.NETQ_AGENT]


def check_yaml(config_dict):
    netq_agent = config_dict.get(ConfKw.NETQ_AGENT)
    netq_cli = config_dict.get(ConfKw.NETQ_CLI)

    if not netq_agent:
        raise ValueError('netq-agent is not defined in {}'.format(DefaultPaths.CONF_FILE))
    if not netq_cli:
        raise ValueError('netq-cli is not defined in {}'.format(DefaultPaths.CONF_FILE))

    if not is_default_netq_agent(config_dict):
        if netq_agent.get(ConfKw.SERVER) is None:
            raise ValueError('netq-agent server is not defined in {}'.format(DefaultPaths.CONF_FILE))

    if not is_default_netq_cli(config_dict):
        if netq_cli.get(ConfKw.COUNT) > ConfKw.MAX_BATCH_SIZE:
            raise ValueError('netq-cli count must be less than equal to 5000')

    if netq_cli.get(ConfKw.SERVER) is None:
        netq_cli[ConfKw.SERVER] = netq_agent.get(ConfKw.SERVER, ConfKw.DEFAULT_SERVER)


def get_yaml_configuration(proc_name='', logger=None, paths=None):
    '''aggregate configuration file values, evaluate default paths alphabetically, provided paths exact order'''

    main_conf = init_default_yaml_configuration()
    conf = dict()

    paths = paths
    for conf_file in paths:
        error, msg = load_configuration(conf_file)
        if error:
            if logger:
                logger.error(msg)
            else:
                print(msg)
        else:
            conf = msg

        if not conf:
            continue

        for key, value in list(conf.items()):
            if not main_conf.get(key):
                main_conf[key] = value
            else:
                if isinstance(main_conf[key], dict) and isinstance(value, dict):
                    # Merge values in original dict with ones from new dict
                    for key2, value2 in list(value.items()):
                        main_conf[key][key2] = value2
                else:
                    # Obliterate original value
                    main_conf[key] = value

    try:
        check_yaml(main_conf)
    except ValueError as ex:
        print((RED_T(ex.message)))
    return main_conf


def save_configuration(conf_file, settings):
    dir_path = os.path.dirname(conf_file)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    with open(conf_file, 'w') as fd:
        yaml.safe_dump(settings, fd)


def load_configuration(conf_file):
    try:
        sfile = open(conf_file, 'r')
    except IOError as ioe:
        return True, ('Unable to open configuration file: %s, error: %s' % (conf_file, str(ioe)))
    else:
        with sfile:
            try:
                return False, yaml.safe_load(sfile)
            except yaml.YAMLError as yex:
                return True, ('Unable to parse configuration file: %s, error: %s' % (conf_file, str(yex)))


def safe_load_all(conf_file):
    try:
        sfile = open(conf_file, 'r')
    except IOError as ioe:
        return True, ('Unable to open configuration file: %s, error: %s' % (conf_file, str(ioe)))
    else:
        with sfile:
            try:
                return False, list(yaml.safe_load_all(sfile))
            except yaml.YAMLError as yex:
                return True, ('Unable to parse configuration file: %s, error: %s' % (conf_file, str(yex)))


def get_config(proc_name='', log=False, paths=None):
    '''aggregate configuration file values, evaluate default paths alphabetically, provided paths exact order'''

    # TODO: Move
    logger = get_logger(proc_name) if log else None

    conf = get_yaml_configuration(proc_name, logger, paths=get_paths(paths))

    if logger:
        for hdlr in logger.handlers:
            logger.removeHandler(hdlr)
    del logger

    return conf


def get_paths(paths):
    if paths is None:
        return __get_configuration_paths()
    return paths if isinstance(paths, list) else [paths]


def __get_configuration_paths():
    conf_files = []

    if os.path.exists(DefaultPaths.CONF_FILE):
        conf_files.append(DefaultPaths.CONF_FILE)

    for root, dirs, files in os.walk(DefaultPaths.CONF_DIR):
        for filename in sorted(files):
            # only look for files with .yml or .yaml extensions
            if filename.endswith('.yml') or filename.endswith('.yaml'):
                conf_files.append(os.path.join(root, filename))

    return conf_files


def process_service_file(sfile, quota, add=True):
    service_file = '/lib/systemd/system/netq-agent.service'
    match = '[Service]'
    buffer = ""
    pattern = 'CPUQuota='
    for line in sfile.readlines():
        if match in line and add:
            buffer += line + pattern + str(quota) + "%\n"
        elif pattern in line:
            continue
        else:
            buffer += line
    try:
        ofile = open(service_file, 'w')
    except IOError as ioe:
        print(('Unable to open configuration file: %s, error: %s' % (service_file, str(ioe))))
        return False
    ofile.write(buffer)
    ofile.close()
    return True


def update_cpu_quota(quota, add=True):
    service_file = '/lib/systemd/system/netq-agent.service'
    if ((quota < 40) or (quota > 60)) and add is True:
        print("Allowed range (40-60). Aborting...")
        return False
    # The unsetting of CPU quota doesn't reset the
    # CPUQuota parameter, offsetting the bug by setting quota to 100
    elif not add:
        add = True
        quota = 100
    try:
        sfile = open(service_file, 'r')
    except IOError as ioe:
        print(('Unable to open configuration file: %s, error: %s' % (service_file, str(ioe))))
        return False
    else:
        with sfile:
            try:
                ret = process_service_file(sfile, quota, add)
            except Exception as ex:
                print(('Unable to parse configuration file: %s, error: %s' % (service_file, str(ex))))
                return False
    return ret
