# Copyright 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import shutil
import os
from collections import defaultdict
# from netq_lib.common.configuration import DefaultPaths, CommandParams
from netq_lib.common.agent_stats_monitor import AgentStatsMon
from netq_lib.common.enums import DefaultPaths, CommandParams
from netq_lib.common.utils import get_platform_type
from netq_lib.common.configuration import load_configuration, save_configuration
from netq_lib.orm.redisdb.models import BgpSession, EvpnVniInfo, OspfIf, OspfNbr, ClagSession, PackageInfo, Ptm, \
    MstpInfo, License, Port, AgentStats, Lldp, ForwardingResource, ACLResource, ProcNetDevInterface, ConfigMon, \
    RunningConfigMon, ClSupport, ResourceUtil, Temp, Fan, PSU, Power, SsdUtil, BtrFsInfo, Dom, Roce, RoceConfig


class NetqCommandInfo(object):

    def __init__(self, service, key, period, command, is_active, parser, callback=None):
        self.__service = service
        self.__service_key = key
        self.__parser = parser
        self.__command = command
        self.__isactive = is_active
        self.__period = period
        self.__callback = callback

    def __eq__(self, other):
        if (self.__service == other.__service)  and\
                (self.__service_key == other.__service_key) and \
                (self.__isactive == other.__isactive) and\
                (self.__period == other.__period):
            return True
        else:
            return False

    def to_dict(self):
        cmd = {}
        cmd[CommandParams.SERVICE] = self.__service
        cmd[CommandParams.SERVICEKEY] = self.__service_key
        cmd[CommandParams.PERIOD] = self.__period
        cmd[CommandParams.COMMAND] = self.__command
        cmd[CommandParams.ISACTIVE] = self.__isactive
        cmd[CommandParams.PARSER] = self.__parser
        return cmd

    @classmethod
    def build(cls, service, key, parser, command):
        yield cls(service, key, parser, command)

    @property
    def service(self):
        return self.__service

    @service.setter
    def service(self, value):
        self.__service = value

    @property
    def service_key(self):
        return self.__service_key

    @service_key.setter
    def service_key(self, value):
        self.__service_key = value

    @property
    def command(self):
        return self.__command

    @command.setter
    def command(self, value):
        self.__command = value

    @property
    def period(self):
        return self.__period

    @period.setter
    def period(self, value):
        self.__period = value

    @property
    def is_active(self):
        return self.__isactive

    @is_active.setter
    def is_active(self, value):
        if value is None:
            value = True
        self.__isactive = value

    @property
    def parser(self):
        return self.__parser

    @parser.setter
    def parser(self, value):
        self.__parser = value


class NetqSupportedCommands(object):

    def __init__(self):
        self.__supported_commands = {
            'bgpd': [
                {
                    "period": "15",
                    "key": "bgp-neighbors",
                    "command": ["/usr/bin/vtysh", "-c",
                                "show ip bgp vrf all neighbors json"],
                    "callback": "cumulus.bgp.process_bgp_neighbors",
                    "model": [BgpSession]
                },
                {
                    "period": "15",
                    "key": "evpn-vni",
                    "command": ["/usr/bin/vtysh", "-c",
                                "show bgp l2vpn evpn vni json"],
                    "callback": "cumulus.bgp.process_bgp_evpn_vni",
                    "model": [EvpnVniInfo]
                },
            ],
            'ospfd': [
                {
                    "period": "15",
                    "key": "ospf-neighbor-json",
                    "command": ["/usr/bin/vtysh", "-c",
                                "show ip ospf vrf all neighbor detail json"],
                    "callback": "cumulus.ospf.process_ospf_neighbors",
                    "model":[OspfIf]
                },
                {
                    "period": "15",
                    "key": "ospf-interface-json",
                    "command": ["/usr/bin/vtysh", "-c",
                                "show ip ospf vrf all interface json"],
                    "callback": "cumulus.ospf.process_ospf_if",
                    "model":[OspfNbr]
                }

            ],
            'clagd': [
                {
                    "period": "15",
                    "key": "clagctl-json",
                    "command": "/usr/bin/clagctl -j",
                    "callback": "cumulus.clag.process_clagctl",
                    "model":[ClagSession]
                }
            ],
            'dpkg': [
                {
                    "period": "21600",
                    "key": "dpkg-query",
                    "command": "dpkg-query --show -f ${Package},${Version},${Status}\\n",
                    "callback": "cumulus.dpkg.process_dpkg",
                    "model":[PackageInfo]
                }
            ],
            'ptmctl': [
                {
                    "period": "30",
                    "key": "ptmctl-json",
                    "command": "/usr/bin/ptmctl -d -j",
                    "callback": "cumulus.ptm.process_ptmctl",
                    "model":[Ptm]
                }
            ],
            'mstpd': [
                {
                    "period": "15",
                    "key": "mstpctl-bridge-json",
                    "command": "/sbin/mstpctl showall json",
                    "callback": "cumulus.mstp.process_mstpall",
                    "model":[MstpInfo]
                }
            ],
            "misc": [
                {
                    "period": "30",
                    "key": "cl-license",
                    "command": "/usr/sbin/switchd -lic",
                    "callback": "cumulus.platform.process_cl_license",
                    "model":[License]
                },
                {
                    "period": "3600",
                    "key": "ports",
                    "command": None,
                    "callback": "cumulus.ports.process_ports",
                    "model": [Port]
                },
                {
                    "period": "30",
                    "key": "proc-net-dev",
                    "command": None,
                    "callback": "cumulus.proc.process_proc_net_dev",
                    "model":[ProcNetDevInterface]
                },
                {
                    "period": "60",
                    "key": "roce",
                    "command": "/usr/lib/cumulus/mlxcmd --json roce counters",
                    "callback": "cumulus.roce.process_roce",
                    "model":[Roce]
                },
                {
                    "period": "30",
                    "key": "roce-config",
                    "command": None,
                    "callback": "cumulus.roce.process_roce_config",
                    "model":[RoceConfig]
                },
                {
                    "period": "300",
                    "key": "dom",
                    "command": None,
                    "callback": "cumulus.dom.process_dom",
                    "model":[Dom]
                },
            ],
            "agentStats": [
                {
                    "period": "300",
                    "key": "agent_stats",
                    "command": None,
                    "callback": "cumulus.agentStats.process_agent_stats",
                    "model": [AgentStats]
                },
                {
                    "period": "15",
                    "key": "agent_util_stats",
                    "command": None,
                    "callback": "cumulus.agentStats.process_agent_util_stats",
                    "model": [AgentStatsMon]
                }
            ],
            "lldpd": [
                {
                    "period": "30",
                    "key": "lldp-json",
                    "command": "/usr/sbin/lldpctl -f json",
                    "callback": "cumulus.lldpd.process_lldpctl",
                    "model": [Lldp]
                }
            ],
            "tcam": [
                {
                    "period": "30",
                    "key": "tcam-resource-json",
                    "command": "/usr/cumulus/bin/cl-resource-query -j",
                    "callback": "cumulus.tcam_resource.process_tcam_resource",
                    "model": [ForwardingResource, ACLResource]
                }
            ],
            "btrfs": [
                {
                    "period": "1800",
                    "key": "btrfs-json",
                    "command": "/sbin/btrfs fi usage -b /",
                    "callback": "cumulus.btrfs.process_btrfs",
                    "model":[BtrFsInfo]
                }
            ],
            "configMon": [
                {
                    "period": "30",
                    "key": "config-mon-json",
                    "command": None,
                    "callback": "cumulus.configMon.process_config_mon",
                    "model":[ConfigMon]
                }
            ],
            "runningConfigMon": [
                {
                    "period": "30",
                    "key": "running-config-mon-json",
                    "command": None,
                    "callback": "cumulus.runningConfigMonitor.process_running_config_mon",
                    "model":[RunningConfigMon]
                }
            ],
            "clSupport": [
                {
                    "period": "60",
                    "key": "cl-support-json",
                    "command": None,
                    "callback": "cumulus.clSupport.process_cl_support",
                    "model":[ClSupport]
                }
            ],
            "resourceUtil": [
                {
                    "period": "30",
                    "key": "resource-util-json",
                    "command": "findmnt / -n -o FS-OPTIONS",
                    "callback": "cumulus.resourceUtil.process_resource_util",
                    "model":[ResourceUtil]
                }
            ],
            "smond": [
                {
                    "period": "30",
                    "key": "smonctl-json",
                    "command": "/usr/sbin/smonctl -j",
                    "callback": "cumulus.sensors.process_sensors",
                    "model":[Temp, Fan, PSU]
                },
                {
                    "period": "30",
                    "key": "sensors-json",
                    "command": "sensors -u",
                    "callback": "cumulus.sensors.process_power",
                    "model":[Power]
                }
            ],
            "ssdUtil": [
                {
                    "period": "86400",
                    "key": "ssd-util-json",
                    "command": "/usr/sbin/smartctl -a /dev/sda",
                    "callback": "cumulus.ssdUtil.process_ssd_util",
                    "model":[SsdUtil]
                }
            ],
        }

    @property
    def services(self):
        return self.__supported_commands.keys()

    @property
    def service_keys(self):
        service_key = []
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                service_key.append(cmd['key'])
        service_key

    def service_keys(self, service_name):
        service_key = []
        for key, commands in self.__supported_commands.items():
            if (service_name == key):
                for cmd in commands:
                    service_key.append(cmd['key'])
        service_key

    def command(self, service, service_key):
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                if cmd['key'] == service_key:
                    return cmd['command']
        return None

    def get_models(self, service_key):
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                if cmd['key'] == service_key:
                    return cmd['model']
        return None

    def service_callback(self, service_key):
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                if cmd['key'] == service_key:
                    return cmd['callback']
        return None

    @property
    def service_info(self):
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                callback = cmd['callback'] if 'callback' in cmd else None
                yield NetqCommandInfo(key, cmd[CommandParams.SERVICEKEY],
                                      cmd[CommandParams.PERIOD] if CommandParams.PERIOD in cmd else None,
                                      cmd[CommandParams.COMMAND],
                                      cmd[CommandParams.ISACTIVE] if CommandParams.ISACTIVE in cmd else None,
                                      cmd[CommandParams.PARSER] if CommandParams.PARSER in cmd else None)

    def get_command_info(self, service_key):
        for key, commands in self.__supported_commands.items():
            for cmd in commands:
                if service_key == cmd[CommandParams.SERVICEKEY]:
                    callback = cmd['callback'] if 'callback' in cmd else None
                    return NetqCommandInfo(key, cmd[CommandParams.SERVICEKEY],
                                           cmd[CommandParams.PERIOD],
                                           cmd[CommandParams.COMMAND],
                                           cmd[CommandParams.ISACTIVE] if CommandParams.ISACTIVE in cmd else None,
                                           cmd[CommandParams.PARSER] if CommandParams.PARSER in cmd else None)
        return None


class NetqFileManager(object):

    def __init__(self, conf=None, logger=None):
        self._conf = conf
        self._logger = logger
        self.__supported_cmds = NetqSupportedCommands()
        self.__file_mod_time = 0


    def __update_mod_time(self):
        if os.path.exists(DefaultPaths.WORKING_COMMANDS):
            self.__file_mod_time = os.path.getmtime(DefaultPaths.WORKING_CMD_FILE)

    def is_file_modified(self):
        curr_mtime = os.path.getmtime(DefaultPaths.WORKING_CMD_FILE)
        if not curr_mtime:
            self.__file_mod_time = curr_mtime
            return False
        else:
            if curr_mtime > self.__file_mod_time:
                self.__file_mod_time = curr_mtime
                return True
            else:
                return False

    def get_supported_keys(self):
        return self.__supported_cmds.service_keys()

    def get_supported_services(self):
        return self.__supported_cmds.service_info()

    def get_supported_keys(self, service_name):
        return self.__supported_cmds.service_keys(service_name)

    def __reset_running_commands(self, cmd_file):
        src_file = DefaultPaths.DEFAULT_COMMANDS + cmd_file
        dst_file = DefaultPaths.WORKING_CMD_FILE
        self._logger.info("Loading default file {}".format(src_file))
        # Create directories if it doesn't exist
        os.makedirs(DefaultPaths.WORKING_COMMANDS, exist_ok=True)

        if (os.path.exists(src_file)):
            shutil.copy(src_file, dst_file)
            return True
        else:
            self._logger.error("Unable to find file {}".format(src_file))
            return False

    def update_working_files(self):
        ptype = get_platform_type()
        command_file = "{}-netq-commands.yml".format(ptype)
        self._logger.info("Resetting Running Commands Source : {}".format(command_file))
        return self.__reset_running_commands(command_file)

    def __get_working_files(self, predicate):
        cmd_file = None
        cmd_files = []
        update = True
        self._logger.debug("Loading config for platform type {}".format(get_platform_type()))
        if os.path.exists(DefaultPaths.WORKING_COMMANDS):
            if os.path.exists(DefaultPaths.WORKING_CMD_FILE):
                cmd_file = DefaultPaths.WORKING_CMD_FILE
                update = False
                if predicate:
                    self._logger.info("Loading existing file {}".format(cmd_file))
                else:
                    self.update_working_file_with_new_commands()
        if update:
            if self.update_working_files():
                cmd_file = DefaultPaths.WORKING_CMD_FILE
            else:
                self._logger.error("Unable to load commands")

        if cmd_file:
            cmd_files.append(cmd_file)
        return cmd_files

    def get_configured_commands(self, predicate=False):
        cmd_files = self.__get_working_files(predicate)
        main_cmds = dict()

        for conf_file in cmd_files:
            error, msg = load_configuration(conf_file)
            if error:
                self._logger.error(msg)
                break
            else:
                conf = msg

            if not conf:
                continue

            for key, value in list(conf.items()):
                if not main_cmds.get(key):
                    main_cmds[key] = value
                else:
                    if isinstance(main_cmds[key], dict) and isinstance(value, dict):
                        for key2, value2 in list(value.items()):
                            main_cmds[key][key2] = value2
                    else:
                        main_cmds[key] = value

        self.__update_mod_time()
        return main_cmds

    def get_commands(self, command_file):
        main_cmds = dict()
        error, msg = load_configuration(command_file)
        if error:
            self._logger.error(msg)
            return main_cmds
        else:
            existing_conf = msg
        if existing_conf:
            for key, value in list(existing_conf.items()):
                if not main_cmds.get(key):
                    main_cmds[key] = value
                else:
                    if isinstance(main_cmds[key], dict) and isinstance(value, dict):
                        for key2, value2 in list(value.items()):
                            main_cmds[key][key2] = value2
                    else:
                        main_cmds[key] = value
        return main_cmds

    def get_command_keys(self, main_cmds):
        cmds = defaultdict(list)
        for entry in main_cmds[CommandParams.NETQCOMANDS]:
            service = entry.get(CommandParams.SERVICE, 'misc')
            commands = entry.get(CommandParams.COMMANDS, [])
            for command in commands:
                if command.get(CommandParams.ISACTIVE):
                    cmds[service].append(command)
        return cmds

    def get_all_command_keys(self, main_cmds):
        cmds = defaultdict(list)
        for entry in main_cmds[CommandParams.NETQCOMANDS]:
            service = entry.get(CommandParams.SERVICE, 'misc')
            commands = entry.get(CommandParams.COMMANDS, [])
            for command in commands:
                cmds[service].append(command)
        return cmds

    def update_working_file_with_new_commands(self):
        """
        1) Get the Relevant Default File.
        2) Check if there is new command added in default file.
        3) Write new commands to working file.
        """

        ptype = get_platform_type()
        command_file = DefaultPaths.DEFAULT_COMMANDS + "{}-netq-commands.yml".format(ptype)
        self._logger.debug("Default File getting checked for new command : {}".format(command_file))
        new_cmds = self.get_commands(command_file)
        if not new_cmds:
            return
        default_cmd_set = self.get_command_keys(new_cmds)

        working_cmds = self.get_commands(DefaultPaths.WORKING_CMD_FILE)
        if not working_cmds:
            return
        working_cmd_set = self.get_all_command_keys(working_cmds)

        newly_added_cmds = defaultdict(list)
        for service, commands in default_cmd_set.items():
            if service in working_cmd_set:
                existing_keys = [cmd[CommandParams.SERVICEKEY] for cmd in working_cmd_set[service]]
                for cmd in commands:
                    if cmd[CommandParams.SERVICEKEY] not in existing_keys:
                        newly_added_cmds[service].append(cmd)
            else:
                newly_added_cmds[service] = commands
        self._logger.debug("Newly added commands list {}".format(newly_added_cmds))
        cmds = working_cmds[CommandParams.NETQCOMANDS]
        delta_svc = []
        for service_name, new_cmds in newly_added_cmds.items():
            service_found = False
            for service_obj in cmds:
                if service_name == service_obj[CommandParams.SERVICE]:
                    service_found = True
                    for cmd in new_cmds:
                        service_obj[CommandParams.COMMANDS].append(cmd)
            if not service_found:
                service_obj = dict()
                service_obj[CommandParams.SERVICE] = service_name
                service_obj[CommandParams.COMMANDS] = new_cmds
                delta_svc.append(service_obj)

        working_cmds[CommandParams.NETQCOMANDS] = cmds + delta_svc
        save_configuration(DefaultPaths.WORKING_CMD_FILE, working_cmds)
