# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt



import atexit
import os
import signal
import socket
import subprocess
import sys
from abc import ABCMeta, abstractmethod

import greenlet
import eventlet
from daemon import daemon

from netq_lib.common import utils


class Service(object, metaclass=ABCMeta):  # flake8: noqa
    """ Abstract base class.
    """

    def __init__(self, conf, proc_name, enable_monkey_patch=True, shutdown_callbacks=None):
        self._conf = conf
        # Set up signal handlers before daemonizing.
        self._run_gt = None
        self.__setup_signal_handlers()
        self.__daemon_context = daemon.DaemonContext(umask=0o022, working_directory='/var/run')
        # Patch system modules to be greenthread-friendly. Must be after
        # daemonizing.
        if enable_monkey_patch:
            eventlet.monkey_patch(thread=False, select=False)
        self._pool = eventlet.GreenPool(size=1000)
        # Set up logging. Must be after daemonizing.
        self._logger = utils.get_logger(proc_name, self._conf)

        self.__pidfd = utils.Pidfile('/var/run/%s.pid' % proc_name, proc_name)
        self.__proc_name = proc_name

        #Callbacks to be invoked from __term_handler
        self.__shutdown_callbacks = shutdown_callbacks

    @abstractmethod
    def _run(self):
        """ Run method to be implemented by the derived class.
        """
        raise NotImplementedError

    def _serve(self, sock, handle, pool=None, bufsize=None, err_cbs=None):
        """ Runs a udp listener on the supplied socket. Calls the function
        *handle* in a separate greenthread for every incoming datagram.
        *handle* takes two arguments: packet buffer and client address.
        """
        # pylint: disable=too-many-arguments
        err_cbs = err_cbs or {}
        pool = pool or self._pool
        while True:
            try:
                buf, addr = sock.recvfrom(bufsize)
                green_thread = pool.spawn(handle, buf, addr)
                green_thread.link(self._stop_checker)
            except socket.error as ex:
                (err, _) = ex.args
                if err in err_cbs:
                    err_cbs[err]()
                    return
                else:
                    raise
            except eventlet.StopServe:
                return

    def _serve_command(self, cmd, handle, pool=None):
        """ FIXME: Runs a tcp server on the supplied socket. Calls the
        function *handle* in a separate greenthread for every incoming client
        connection. *handle* takes two arguments: the client socket object,
        and the client address.
        """
        pool = pool or self._pool
        bpm = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        while True:
            try:
                line = bpm.stdout.readline()
                if not line:
                    raise Exception('Command %s terminated unexpectedly' % cmd)
                green_thread = pool.spawn(handle, line)
                green_thread.link(self._stop_checker)
            except eventlet.StopServe:
                return
            finally:
                try:
                    bpm.kill()
                except Exception:  # pylint: disable=broad-except
                    pass

    def _serve_tcp(self, sock, handle, pool=None):
        """ Runs a tcp server on the supplied socket. Calls the function
        *handle* in a separate greenthread for every incoming client
        connection. *handle* takes two arguments: the client socket object,
        and the client address.
        """
        pool = pool or self._pool
        while True:
            try:
                client_sock, addr = sock.accept()
                green_thread = pool.spawn(handle, client_sock, addr)
                green_thread.link(self._stop_checker)
            except eventlet.StopServe:
                return

    def _stop_checker(self, green_thread):
        """ Propagates exceptions to the run green thread.
        """
        try:
            green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except Exception:  # pylint: disable=broad-except
            eventlet.kill(self._run_gt, *sys.exc_info())

    def __delpid(self):
        """ Removes the pidfile when the process exits gracefully.
        """
        try:
            os.remove('/var/run/%s.pid' % self.__proc_name)
        except OSError:
            self._logger.debug('Unable to remove pid file on exit')

    def __setup_signal_handlers(self):
        """ Set up signal handlers.
        """
        signal.signal(signal.SIGINT, self.__term_handler)
        signal.signal(signal.SIGTERM, self.__term_handler)
        signal.signal(signal.SIGHUP, self.__term_handler)

    def __term_handler(self, signum, _):
        """ Signal handler that terminates the process.
        """
        msg = 'Caught signal %d' % signum
        self._logger.debug(msg)
        # _run_gt is initialized at init but stuffed in run().
        # But we could get here before that from a SIG.

        # Call shutdown callbacks
        # TODO: call CBs with eventlet sensistive timeout
        if self.__shutdown_callbacks:
            for cb in self.__shutdown_callbacks:
                fcn_info = utils.get_function_info(cb)
                try:
                    cb()
                    self._logger.info('Processed shutdown function %s' % fcn_info)
                except Exception:
                    self._logger.exception('Exception calling shutdown function %s' % fcn_info)

        if self._run_gt is not None:
            eventlet.kill(self._run_gt, SystemExit(msg))
        else:
            raise SystemExit(msg)


    def get_logger(self):
        """ Returns the logger object for the instance.
        """
        return self._logger

    def unsubscribe(self):
        """ Unsubscribe from Pubsub, if you are using it """
        pass

    def run(self):
        """ Once upon a time...
        """
        try:
            # Make sure that I am not already running.
            pid = os.getpid()
            self.__pidfd.lock()
            self.__pidfd.write(pid)
            atexit.register(self.__delpid)
            self._logger.info('Starting (pid %d) ...', pid)
            atexit.register(self._logger.info,
                            'Terminating (pid %d)' % pid)
            self._run_gt = self._pool.spawn(self._run)
            self._run_gt.wait()
        except SystemExit:
            pass
        except Exception:  # pylint: disable=broad-except
            self._logger.exception('Error: ')
            raise
        finally:
            self.__daemon_context.close()
