# Copyright 2018, 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt



import subprocess
import signal
import copy
import fcntl
import inspect
import itertools
import ujson as json
import logging
import logging.handlers
import os
import re
import socket
import sys
import time
import yaml
from datetime import datetime, timedelta
from functools import partial, wraps
from json import JSONEncoder
from logging.handlers import SysLogHandler
from netq_lib.common.enums import ConfKw, DefaultPaths, BgColors, Platform, PlatformType
import ipaddress as ipaddr
from netq_lib.common.netq_pq import NetQPriorityQueue, QueueElem
from configobj import ConfigObj
from functools import wraps

try:
    from io import StringIO
except Exception:
    try:
        from io import StringIO
    except ImportError:
        from io import StringIO
from subprocess import CalledProcessError
import configparser
from logging import INFO, WARNING, ERROR, CRITICAL

def strip_color(string):
    return re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]').sub('', string)


def color_wrap(color, text):
    # http://stackoverflow.com/questions/287871/
    # print-in-terminal-with-colors-using-python

    if not is_color_output():
        return text

    if color == BgColors.GREEN:
        return BgColors.GREEN + text + BgColors.ENDC
    elif color == BgColors.RED:
        return BgColors.RED + text + BgColors.ENDC
    elif color == BgColors.BLUE:
        return BgColors.BLUE + text + BgColors.ENDC
    elif color == BgColors.YELLOW:
        return BgColors.YELLOW + text + BgColors.ENDC
    elif color == BgColors.GRAY:
        return BgColors.GRAY + text + BgColors.ENDC
    elif color == BgColors.BOLD:
        return BgColors.BOLD + text + BgColors.ENDC
    elif color == BgColors.ENDC:
        return text
    else:
        raise RuntimeError('%s is not a supported color' % color)


RED_T = partial(color_wrap, BgColors.RED)
COLOR_GREEN = partial(color_wrap, BgColors.GREEN)

__USE_COLOR = 'unconfigured'


def print_out(connection, msg):
    output = {'rc': 0, 'output': '{}'.format(msg)}
    connection.sendall(bytes('{}\n'.format(json.dumps(output)), 'utf-8'))


def idempotent(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            return None
    return wrapper


def is_color_output():
    global __USE_COLOR

    if __USE_COLOR != 'unconfigured':
        return __USE_COLOR
    __USE_COLOR = True
    '''Default behavior is to print out color'''
    try:
        sfile = open(DefaultPaths.CONF_FILE, 'r')
    except IOError as ioe:
        print(('Unable to open configuration file: %s, error: %s' % (DefaultPaths.CONF_FILE, str(ioe))))
    else:
        with sfile:
            try:
                current_config = yaml.safe_load(sfile)
                if not current_config:
                    current_config = {ConfKw.NETQD: {'color': True}}

                if ConfKw.NETQD in list(current_config.keys()) and \
                        'color' in list(current_config[ConfKw.NETQD].keys()):
                    if isinstance(current_config[ConfKw.NETQD]['color'], bool) and \
                            current_config[ConfKw.NETQD]['color'] is False or \
                            isinstance(current_config[ConfKw.NETQD]['color'], str) and \
                            current_config[ConfKw.NETQD]['color'].lower() == 'no':
                        __USE_COLOR = False
            except yaml.YAMLError:
                pass
    return __USE_COLOR


def is_valid_ip(ip_address):
    """Returns True if this is a valid IPv4/v6 address"""
    try:
        if '/' in ip_address:
            ipaddr.ip_network(ip_address, strict=False)
        else:
            ipaddr.ip_address(ip_address)
        return True
    except Exception:
        return False


def is_valid_mac(mac_address):
    """Returns true if this is a valid MAC address"""
    val = mac_address.replace(':', '').replace('.', '').strip().lower()
    nmac = ':'.join(val[i:i + 2] for i in range(0, len(val), 2))
    return re.match(r'((?:[\da-fA-F]{2}[:\-]){5}[\da-fA-F]{2})', nmac)


def time_format(input_time, suffix):
    input_time = str(input_time)
    suffix = suffix.lower()
    TIME_REGEX = re.compile(
        r'(?:(?:(?P<days>[\d|.]+)d)|'
        r'(?:(?P<hours>[\d|.]+)h)|'
        r'(?:(?P<minutes>[\d|.]+)m)|'
        r'(?:(?P<seconds>[\d|.]+)s))'
    )

    if suffix == 'm':
        if TIME_REGEX.match(input_time) and 'm' in input_time:
            return int(input_time.strip('m'))
        else:
            raise RuntimeError('Time format {} must be in minutes'.format(input_time))
    elif suffix == 'h':
        if TIME_REGEX.match(input_time) and 'h' in input_time:
            return int(input_time.strip('h'))
        else:
            raise RuntimeError('Time format {} must be in hours'.format(input_time))
    else:
        raise RuntimeError('Time format must be in {}m for minutes or {}h for hours'.format(input_time, input_time))


def absolute_date(input_time):
    """Converts UNIX timestamp to human readable format"""
    if input_time > ConfKw.BILLION_SEC_IN_MILLISEC:
        input_time = float(input_time) / 1000.0
    return datetime.fromtimestamp(input_time).ctime()


def pretty_date(input_time):
    """Get a datetime object or a int() Epoch timestamp and return a
    pretty string date. The time string changes the closer the delta
    is to the event. So if the event occurred less than an hour back,
    the time is displayed with granularity is msecs, else h:m:s or
    d:h:m:s
    """
    if not input_time:
        return "never"

    then = datetime.utcfromtimestamp(input_time)
    now = datetime.utcnow()
    if then > now:
        return 'Time warp'

    try:
        diff = then.ctime()
    except Exception:
        if isinstance(time, datetime):
            diff = now - then
        else:
            raise RuntimeError('Type of time (%s) is neither a timestamp or '
                               'datetime' % type(input_time))
    return diff


def msecs_to_dhms(msecs):
    try:
        secs, _ = divmod(msecs, 1000)
        dlt = timedelta(seconds=secs)
        out = [0] * 4
        out[0] = dlt.days
        out[1], fracmins = divmod(dlt.seconds, 3600)
        out[2], out[3] = divmod(fracmins, 60)
        fmts = ['{}d:', '{}h:', '{}m:', '{}s']
        for i in range(4):
            if out[i]:
                fmt = ''.join(fmts[i:])
                return fmt.format(*out[i:])
    except Exception:
        pass

    return None


def strip_colors(text, colors=BgColors.COLORS):
    if not isinstance(text, str) or not colors:
        return text
    for color_idx, color in enumerate(colors):
        idx = text.find(color)
        if idx != -1:
            next_idx = text.find(BgColors.ENDC, idx)
            if next_idx != -1:
                return strip_colors(text[:idx] +
                                    text[idx + len(color):next_idx] +
                                    text[next_idx + len(BgColors.ENDC):],
                                    colors[color_idx + 1:])
    return text


def prtree(tree, shift, top, path, final_json):
    """print a tree when provided input in the form of:
    ['a', ['b', ['c', 'd']]]
    output looks like:
    a -- b -- c
           -- d
    STP Topo, MAC Path trace and others produce output in the form
    used as input by this routine
    """
    prefixstr = ''
    is_leaf_node = True
    host = ''
    port = ''
    entry = {}
    jsonify = False
    if final_json is not None:
        jsonify = True

    if 'hostname' in tree:
        fld = tree['hostname']
        if ':' in fld:
            words = fld.split(':')
            if words[1] == 'mac':
                host = words[0]
                port = ':'.join(words[1:])
            else:
                host = words[0]
                port = ':'.join(words[1:])
        else:
            host = fld
            port = ''
    else:
        if not jsonify:
            print(('Invalid formatting of input {}'.format(tree)))
            return False
        else:
            print('{"error": "Invalid formatting of data"}')
            return False

    if jsonify:
        if 'hostname' in tree:
            for key in tree:
                if key == 'nexthop' or key == 'subtree':
                    continue  # We copy this separately
                if key == 'info' and tree[key] and 'Blocked' in tree[key]:
                    if 'nexthop' in tree:
                        entry['ingress'] = tree['nexthop']
                    else:
                        entry['ingress'] = port
                    entry[key] = tree[key]
                elif key == 'hostname':
                    pass
                else:
                    entry[key] = tree[key]
        if 'nexthop' in tree:
            if 'ingress' not in entry:
                entry['nexthop'] = tree['nexthop']
        elif port != '':
            entry['nexthop'] = port

        if 'hostname' in tree and tree['hostname']:
            entry['hostname'] = host

        path.extend([entry])
    else:
        if top:
            printstr = fld
            if 'info' in tree and tree['info']:
                if 'Error' in tree['info']:
                    printstr = color_wrap(BgColors.RED, printstr)
                elif 'Blocked' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY,
                                          '%s:Blocked(%s)' % (host, port))
                elif 'ClagIsl' in tree['info']:
                    printstr = color_wrap(BgColors.YELLOW,
                                          '%s:ClagIsl(%s)' % (host, port))
                elif 'Root' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN,
                                          'Root(%s)' % printstr)
                elif 'Edge' in tree['info']:
                    printstr = color_wrap(BgColors.BLUE,
                                          '%s:EdgeIntf(%s)' % (host, port))
                elif 'Routed' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN, printstr)
                elif 'Blackhole' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY, printstr)

            sys.stdout.write(prefixstr + printstr)
            shift += len(strip_colors(printstr))
            top = False
        else:
            printstr = ''
            if 'info' in tree and tree['info']:
                if 'Error' in tree['info']:
                    printstr = ' -- ' + color_wrap(BgColors.RED, fld)
                elif 'Blocked' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY,
                                          ' -- %s:Blocked(%s)' % (host, port))
                elif 'ClagIsl' in tree['info']:
                    printstr = color_wrap(BgColors.YELLOW,
                                          ' -- %s:ClagIsl(%s)' % (host, port))
                elif 'Root' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN,
                                          ' -- Root(%s)' % fld)
                elif 'Edge' in tree['info']:
                    printstr = color_wrap(BgColors.BLUE,
                                          ' -- %s:EdgeIntf(%s)' % (host, port))
                elif 'Routed' in tree['info']:
                    printstr = color_wrap(BgColors.GREEN, ' -- %s' % fld)
                elif 'Blackhole' in tree['info']:
                    printstr = color_wrap(BgColors.GRAY, ' -- %s' % fld)

            if 'vtep' in tree:
                printstr = color_wrap(BgColors.BLUE, ' -- %s' % fld)

            if printstr == '':
                printstr = ' -- ' + fld

            sys.stdout.write(prefixstr + printstr)
            shift += len(strip_colors(printstr))

    for fld in tree['subtree']:
        is_leaf_node = False
        if not jsonify:
            sys.stdout.write(prefixstr)
            new_path = []
        else:
            # We're going to branch here. Copy existing path to new
            new_path = copy.deepcopy(path)

        prtree(fld[0], shift, top, new_path, final_json)

        if not jsonify:
            prefixstr = '\n' + ' ' * shift

    if jsonify:
        # In JSON, we print the entire path at the final terminating node
        if is_leaf_node:
            final_json.append(path)
        return final_json
    else:
        return True


class ClassEncoder(JSONEncoder):

    def default(self, o):  # pylint: disable=method-hidden
        return o.__dict__


class Pidfile(object):
    """ Checks if another instance of the program is running.
    Explanation from http://stackoverflow.com/questions/220525/
    ensure-a-single-instance-of-an-application-in-linux
    The Right Thing is advisory locking using flock(LOCK_EX); in Python,
    this is found in the fcntl module.
    Unlike pidfiles, these locks are always automatically released when
    your process dies for any reason, have no race conditions exist
    relating to file deletion (as the file doesn't need to be deleted to
    release the lock), and there's no chance of a different process
    inheriting the PID and thus appearing to validate a stale lock.
    If you want unclean shutdown detection, you can write a marker
    (such as your PID, for traditionalists) into the file after
    grabbing the lock, and then truncate the file to 0-byte status
    before a clean shutdown (while the lock is being held); thus, if the
    lock is not held and the file is non-empty, an unclean shutdown is
    indicated.
    """

    # pylint: disable=missing-docstring
    def __init__(self, pidfile, procname, uuid=None):
        try:
            self.__pidfd = open(pidfile, 'ab+')
        except IOError as err:
            err.extra_info = 'Failed to open pidfile: %s' % pidfile
            raise
        self.__pidfile = pidfile
        self.__procname = procname
        self.__uuid = uuid

    def __str__(self):
        return self.__pidfile

    def is_running(self):
        pid = self.read()
        if pid is None:
            return False
        cmdline = '/proc/%s/cmdline' % pid
        try:
            with open(cmdline, 'r') as pidfd:
                exec_out = pidfd.readline()
            return self.__procname in exec_out and (not self.__uuid or
                                                    self.__uuid in exec_out)
        except IOError:
            return False

    def lock(self):
        try:
            fcntl.flock(self.__pidfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise RuntimeError('Another instance of this program is already '
                               'running.')

    def read(self):
        try:
            self.__pidfd.seek(0)
            return int(self.__pidfd.readline().strip())
        except Exception:  # pylint: disable=broad-except
            return

    def unlock(self):
        try:
            fcntl.flock(self.__pidfd, fcntl.LOCK_UN)
        except IOError:
            raise RuntimeError('Unable to unlock pid file.')

    def write(self, pid):
        self.__pidfd.truncate(0)
        self.__pidfd.write(bytes('{}\n'.format(pid), 'utf-8'))
        self.__pidfd.flush()


def get_logger(proc_name, conf=None):
    logger = logging.getLogger(proc_name)
    if len(logger.handlers):
        # Logger already configured
        return logger

    loglevel = 'info'
    loglevel_dict = {'debug': logging.DEBUG,
                     'info': logging.INFO,
                     'warning': logging.WARNING,
                     'error': logging.ERROR}
    if conf:
        proc_conf = conf.get(proc_name, None)
        if proc_conf:
            loglevel = conf[proc_name].get('log_level', 'info')

    logging_level = loglevel_dict.get(loglevel, logging.INFO)
    if 'QA_CLI_AUTOMATION' in os.environ:
        # Push logs to stdout, in case netqd is running in pod for cli automation
        handler = logging.StreamHandler(sys.stdout)
    else:
        handler = SysLogHandler(address='/dev/log', facility=SysLogHandler.LOG_USER)
    formatter = logging.Formatter(proc_name + '[%(process)d]: %(levelname)s: %(message)s')
    handler.setLevel(logging_level)
    handler.setFormatter(formatter)
    logger.setLevel(logging_level)
    logger.addHandler(handler)
    return logger

def list_to_ranges(in_list):
    '''Convert a list into a compact hyphenated string'''
    # This routine was copied from stackoverflow:
    # http://stackoverflow.com/questions/9847601/convert-list-of-numbers-to-string-ranges
    ret = []
    for val in sorted(in_list):
        if not ret or ret[-1][-1] + 1 != val:
            ret.append([val])
        else:
            ret[-1].append(val)
    return ",".join([str(x[0])
                     if len(x) == 1 else str(x[0]) + "-" + str(x[-1])
                     for x in ret])

def glob_to_numbers(glob):
    """
    Given a string containing single numbers and ranges, return a sorted
    list of deduplicated integers.

    glob - A string of digits and ranges

    >>> glob_to_numbers('3-4,7,10-12,17,22,4001-4003,7777,8000-8004')
    [3, 4, 7, 10, 11, 12, 17, 22, 4001, 4002, 4003, 7777, 8000, 8001, 8002, 8003, 8004]
    """

    assert isinstance(glob, str), "glob={0}".format(glob)

    # Using split(',') instead of the replacement could yield empty strings in the result.
    glob_list = glob.replace(',', ' ').split()

    numbers = set()
    range_re = re.compile(r"""^(\d+)-(\d+)$""")  # ex. 4-6

    for x in glob_list:
        if x.isdigit():
            numbers.add(int(x))
        else:
            range_match = range_re.match(x)

            if range_match is None:
                # The substring is neither a digit nor a range.
                print(("Globs must consist of numbers or ranges, but {0} is neither. We were given glob '{1}'.".format(x, glob)))
                return []
            else:
                min_range = int(range_match.group(1))
                max_range = int(range_match.group(2))

                if max_range >= min_range:
                    numbers.update(list(range(min_range, max_range + 1)))
                else:
                    print(("Glob \"{0}\" contains the invalid range \"{1}\".".format(glob, x)))  # ex. 6-4
                    return []

    return sorted(numbers)  # A sorted list

def safe_run_checkoutput(cmd):
    '''Safely run check_output command, return output good or bad'''
    if isinstance(cmd, str):
        cmd = cmd.split()

    with open(os.devnull, 'w') as fnull:
        try:
            return 0, cmd, subprocess_wrapper(cmd)
        except CalledProcessError as cpe:
            return cpe.returncode, cpe.cmd, cpe.output
        except Exception as ex:
            return -1, cmd, ex.message


__VALID_HOSTNAME_RE = '(?!-)[A-Z\d-]{1,63}(?<!-)$'
__VALID_SEG = re.compile(__VALID_HOSTNAME_RE, re.IGNORECASE)


def is_valid_hostname(hostname):
    '''Check hostname conforms with RFC'''
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]
    return all(__VALID_SEG.match(x) for x in hostname.split("."))


__IPV46 = re.compile(r'(?<=inet)6{0,1} (.*)(?=\/)', re.M)
__IPCMD = '/sbin/ip -o %s addr show scope global up'


def ips_on_local():
    '''Discover all the local IPs via ip addr'''
    ips = set()
    for cmd in [__IPCMD % f for f in ['-4', '-6']]:
        retcode, _, outp = safe_run_checkoutput(cmd)
        if retcode == 0:
            ips |= set(__IPV46.findall(outp))
    return ips


def find_ip_address(msg):
    '''Find IP:PORT address in a string'''
    prts = re.findall(
        r'(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d):([0-9]{4,5})',
        msg)
    return ''.join(prts[:-1]), prts[-1:] if prts and len(prts) == 5 else '', ''


def load_environment(loc):
    try:
        section = 'main'
        with open(loc, 'r') as props_f:
            enviro_f = StringIO(''.join(itertools.chain(['[%s]\n' % section], props_f.read(1024))))

            config = configparser.RawConfigParser()
            config.optionxform = str
            config.readfp(enviro_f)
            props = {k: v for k, v in list(dict(config.items(section)).items()) if not os.environ.get(k)}
            os.environ.update(props)

    except Exception:
        # Ignore exceptions and rely on os.environ for good or bad
        pass


def sorted_hosts(raw_ips):
    if raw_ips:
        not_ips = []
        ips = []
        for ip in raw_ips:
            try:
                socket.inet_pton(socket.AF_INET, ip)
                ips.append(ip)
            except Exception:
                not_ips.append(ip)
        ips = sorted_ips(ips)
        return (ips if ips else []) + (sorted(not_ips) if not_ips else [])
    else:
        return raw_ips


def sorted_ips(raw_ips):
    '''Sort IPs in natural order'''
    if not raw_ips:
        return raw_ips
    ips = [socket.inet_pton(socket.AF_INET, ip) for ip in raw_ips]
    ips.sort()
    return [socket.inet_ntop(socket.AF_INET, ip) for ip in ips]


def obj_delta(prev_obj, cur_obj):
    event_dict = dict()

    if type(prev_obj) == type(cur_obj):
        for key, prev_val in list(prev_obj.__dict__.items()):
            if key.endswith('timestamp'):
                continue
            cur_val = cur_obj.__dict__.get(key)
            if prev_val != cur_val:
                event_dict[key[2:]] = (prev_val, cur_val)
    return event_dict


#########
# SYSTEMD
#########

# TODO: consolidate with netq_apps/modules/core/netq/daemon.pyx
def systemd_service(cmd, service):
    '''Common systemctl commands function for daemons'''
    legit_cmds = ['start', 'stop', 'status', 'restart', 'reload', 'disable', 'enable', 'is-active']
    systemctl = '/bin/systemctl'

    if cmd not in legit_cmds:
        raise NotImplemented('systemd command %s not recognized.' % cmd)

    return safe_run_checkoutput([systemctl, cmd, service])


def status_systemd_svc(service, full=False):
    rc, cmd, msg = systemd_service('status', service)
    if full or rc == -1:
        return msg

    for l in msg.splitlines():
        if 'Active:' in l:
            return ''.join(l.split(':')[1:])


def status_systemd_svcs(services, full=False):
    return [(service, status_systemd_svc(service, full)) for service in services]


class SystemdService(object):
    def __init__(self, service):
        self.service = service

    def disable(self):
        return systemd_service('stop', self.service)

    def enable(self):
        return systemd_service('start', self.service)

    def status(self, full=False):
        return status_systemd_svc(self.service, full)

    def reload(self):
        return systemd_service('reload', self.service)


def convert_severity(severity):
    sev_dict = {'INFO': INFO, 'WARNING': WARNING, 'WARN': WARNING,
                'ERROR': ERROR, 'CRITICAL': CRITICAL, 'CRIT': CRITICAL}
    return sev_dict.get(severity.upper(), INFO)


def get_function_info(fcn):
    try:
        return inspect.getmodule(fcn).__name__ + '.' + fcn.__name__
    except Exception:
        return 'unknown'


def safe_json_out(inp):
    try:
        return json.dumps(inp, indent=4, escape_forward_slashes=False)
    except Exception as ex:
        return json.dumps({'error': ex.message}, escape_forward_slashes=False)


def validate_ips(ips):
    if not ips:
        return None

    servers = set()
    for ip in str(ips).split():
        try:
            ipaddr.ip_network(str(ip), strict=False)
            servers.add(ip)
        except ValueError as ex:
            raise ValueError('Invalid IP %s ' % ex.message)
    return list(servers)


def find_host_ip_in_list(ips=None):
    if ips is None:
        raise Exception('Cannot find if local host IP is valid without a list of target IPS')

    local_ips = ips_on_local()
    host = local_ips & set(ips) if ips and len(ips) else None
    if not host:
        print((color_wrap(BgColors.RED,
            'Cannot find local host IP(s) %s within %s' %
            (','.join(ips) if ips else '', ','.join(local_ips) if local_ips else '')
        )))
        return None
    return host.pop()


def retry_on_timeout(retry_count=5, delay=5, callback_on_exception=None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = None
            for _ in range(10000000):
                try:
                    result = func(*args, **kwargs)
                    if not result:
                        return result
                except Exception as e:
                    trace_back = sys.exc_info()[2]
                    kwargs['exception'] = e
                    kwargs['traceback'] = trace_back
                    if callback_on_exception:
                        callback_on_exception(*args, **kwargs)
                time.sleep(delay)
            return result
        return wrapper
    return decorator

def retry_with_backoff(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            num_tries, num_delay = tries, delay
            while num_tries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck:
                    time.sleep(num_delay)
                    num_tries -= 1
                    num_delay *= backoff
            return f(*args, **kwargs)
        return wrapper
    return decorator


def signal_handler(signal, frame):
    pass


def stop_svc_old_vrf_context(name, vrf):
    '''Stop/disable the service in the old VRF context'''

    systemctl = '/bin/systemctl'
    fnull = open(os.devnull, 'w')

    # TODO: Use utils version of systemd
    # Check if this service is enabled in a different VRF
    if not vrf or vrf == 'None':
        check_svc = '%s@%s' % (name, 'mgmt')
    else:
        check_svc = name
    try:
        subprocess_wrapper([systemctl, 'is-active', check_svc])
        cur_active = True
    except subprocess.CalledProcessError:
        cur_active = False

    try:
        subprocess_wrapper([systemctl, 'is-enabled', check_svc])
        cur_enabled = True
    except subprocess.CalledProcessError:
        cur_enabled = False

    if cur_enabled or cur_active:
        try:
            subprocess_wrapper([systemctl, 'disable', check_svc])
            subprocess_wrapper([systemctl, 'stop', check_svc])
        except subprocess.CalledProcessError:
            pass
        return True

    return False


def process_svc_command(name, cmd_issued, vrf):
    systemctl = '/bin/systemctl'
    fnull = open(os.devnull, 'w')
    RC_SUCCESS = 0  # since failed to import from netq_apps
    RC_FAIL = 1
    rc = RC_SUCCESS

    if vrf:
        if vrf == 'default':
            vrfsvcname = name
        else:
            vrfsvcname = '%s@%s' % (name, vrf)
    else:
        vrfsvcname = name
    svcname = name

    if cmd_issued == 'start':
        if os.getuid() != 0:
            print('Please run this command as root')
            rc = RC_FAIL

        stop_svc_old_vrf_context(name, vrf)
        try:
            subprocess_wrapper([systemctl, 'daemon-reload'])
            subprocess_wrapper([systemctl, 'enable', svcname])
            subprocess_wrapper([systemctl, 'reset-failed', svcname])
            subprocess_wrapper([systemctl, 'start', svcname])
            print(color_wrap(BgColors.GREEN,
                                   'Starting {}... Success!'.format(svcname)))
        except subprocess.CalledProcessError:
            print(('Failed to start %s' % svcname))
            rc = RC_FAIL
        finally:
            return rc

    elif cmd_issued == 'stop':
        if os.getuid() != 0:
            print('Please run this command as root')
            rc = RC_FAIL
        try:
            subprocess_wrapper([systemctl, 'stop', svcname])
            subprocess_wrapper([systemctl, 'disable', svcname])
            print(color_wrap(BgColors.GREEN,
                                   'Stopping {}... Success!'.format(svcname)))
        except subprocess.CalledProcessError:
            print(('Failed to stop %s' % svcname))
            rc = RC_FAIL
        finally:
            return rc

    elif cmd_issued == 'status':
        try:
            signal.signal(signal.SIGCHLD, signal_handler)
            if name == 'netqd':
                subprocess_wrapper([systemctl, 'status', vrfsvcname])
                print((color_wrap(BgColors.GREEN, 'NetQ CLI... Running')))
            else:
                subprocess_wrapper([systemctl, 'status', svcname])
                print((color_wrap(BgColors.GREEN,
                                       '{}... Running'.format(svcname))))
        except subprocess.CalledProcessError:
            try:
                subprocess_wrapper([systemctl, 'is-failed', svcname])
                if name == 'netqd':
                    print('NetQ CLI...Failed')
                else:
                    print(('{}...Failed'.format(svcname)))
            except subprocess.CalledProcessError:
                if name == 'netqd':
                    print('NetQ CLI...stopped')
                else:
                    print(('{}...stopped'.format(svcname)))
            finally:
                rc = RC_FAIL
        finally:
            signal.signal(signal.SIGCHLD, signal.SIG_IGN)
            return rc

    elif cmd_issued == 'restart':
        if os.getuid() != 0:
            print('Please run this command as root')
            rc = RC_FAIL

        stop_svc_old_vrf_context(name, vrf)
        try:
            if name == 'netqd':
                subprocess_wrapper([systemctl, 'daemon-reload'])
                subprocess_wrapper([systemctl, 'enable', vrfsvcname])
                subprocess_wrapper([systemctl, 'reset-failed', vrfsvcname])
                subprocess_wrapper([systemctl, 'restart', vrfsvcname])
                print((color_wrap(BgColors.GREEN,
                                       'Restarting NetQ CLI... Success!')))
            else:
                subprocess_wrapper([systemctl, 'daemon-reload'])
                subprocess_wrapper([systemctl, 'enable', svcname])
                subprocess_wrapper([systemctl, 'reset-failed', svcname])
                subprocess_wrapper([systemctl, 'restart', svcname])
                print((color_wrap(BgColors.GREEN,
                                       'Restarting {}... Success!'.format(svcname))))
        except subprocess.CalledProcessError as cpe:
            if cpe.returncode == -15:
                if name == 'netqd':
                    print((color_wrap(BgColors.GREEN,
                                       'Restarting NetQ CLI... Success!')))
                else:
                    print((color_wrap(BgColors.GREEN,
                                       'Restarting {}... Success!'.format(svcname))))
            else:
                if name == 'netqd':
                    print(('Failed to restart NetQ CLI %s' % cpe.returncode))
                else:
                    print(('Failed to restart %s' % name))
            rc = RC_FAIL
        finally:
            return rc

    else:
        print(('Invalid command %s for %s' % (cmd_issued, name)))
        rc = RC_FAIL
        return rc


def save_fullstate(message_array, msg_type="dummy", is_full_state=True):

    pq = NetQPriorityQueue.PQ
    if pq is None:
        return False

    if len(message_array) <= 0:
        priority = 1
        message_type = msg_type
        message_data = []
    else:
        priority = message_array[0].get_priority()
        message_type = message_array[0].message_type
        message_data = [queue_data.to_object_json() for queue_data in message_array]

    try:
        queue_elem = QueueElem(priority, message_type, is_full_state, message_data)
        pq.putq(queue_elem)
    except Exception:
        return False

    return True


def is_cloud_opta():
    cmd = "/bin/cat /etc/app-release"
    cloud_env_key = "APPLIANCE_NAME"
    cloud_env_value = "NetQ Cloud Appliance"
    try:
        output = subprocess_wrapper(cmd)
        # Reference output
        # $ cat /etc/app-release
        # APPLIANCE_VERSION=2.2.0
        # APPLIANCE_MANIFEST_HASH=9c1c0f1
        # APPLIANCE_NAME="NetQ Cloud Appliance"
        if output:
            for line in output.splitlines():
                if line.startswith(cloud_env_key):
                    name, value = line.split('=')
                    if value and value.strip("\"") == cloud_env_value:
                        return True
    except subprocess.CalledProcessError:
        return False

    return False


def get_cpu_quota():
    try:
        cmd = ["systemctl", "show", "netq-agent", "-p", "CPUQuotaPerSecUSec"]
        output = subprocess_wrapper(cmd)
        cpuquotams = output.split("=")[1]
        if "ms" in cpuquotams:
            cpuquota = float(cpuquotams.replace("ms", "")) / 1000 * 100
            return cpuquota
        else:
            return None
    except Exception as e:
        print (e)
        return None


def get_cpu_architecture():
    try:
        #cmd="dpkg --print-architecture"
        cmd = "lscpu"
        output = subprocess_wrapper(cmd)
        for param in output.split("\n"):
            key_value = param.strip().split(":")
            if 'Architecture' in key_value[0]:
                return key_value[1].strip(' ')
        return None
    except Exception:
        return None


def get_ip_address(ifname="eth0"):
    try:
        cmd = '/sbin/ip -4 -br addr show {}'.format(ifname)
        output = subprocess_wrapper(cmd.split())
        ip_address = ipaddr.ip_network(output.split()[2], strict=False)
        return ip_address.ip
    except subprocess.CalledProcessError:
        return None
    return None


def get_cl_version():
    """Get the platform name of this device. Skip if its not Cumulus.
    """
    if os.path.exists(DefaultPaths.OS_RELEASE):
        conf = ConfigObj(DefaultPaths.OS_RELEASE)
        if conf.get('NAME') == Platform.CL:
            version = conf.get('VERSION', '0.0.0').lstrip(Platform.CL)
            if '~' in version:
                idx = version.index('~')
                version = version[:idx]
            return version
    return None


def is_cpulimit_installed():
    cmd = "/usr/bin/dpkg -s cpulimit"
    retcode, _, outp = safe_run_checkoutput(cmd)
    is_installed = False
    if outp:
        for line in outp.split("\n"):
            if "Status" in line and "installed" in line:
                is_installed = True
                break
    return is_installed


def check_version(found_version_str, supported_version_str):

    found_version = [int(ver) for ver in found_version_str.split(".")]
    supported_version = [int(ver) for ver in supported_version_str.split(".")]

    #Test Major
    if found_version[0] == supported_version[0]:
        #Test Minor
        if found_version[1] > supported_version[1]:
            return True
        #Test Release
        elif found_version[1] == supported_version[1]:
            if found_version[2] >= supported_version[2]:
                return True
    return False


def is_cpuquota_supported():

    supported_versions = ['3.7.12', '4.1.0']
    try:
        version = get_cl_version()
        for supported_version in supported_versions:
            if check_version(version, supported_version):
                return True
        return False
    except Exception:
        return False


def get_all_ip_address(ifname):
    try:
        cmd = '/sbin/ip -4 -br addr show {}'.format(ifname)
        output = subprocess_wrapper(cmd.split())
        list_ip = list()
        for ip in output.split()[2:]:
            ip_address = ipaddr.ip_interface(ip)
            list_ip.append(ip_address)
        return list_ip
    except subprocess.CalledProcessError:
        return None


class SettingsException(Exception):
    pass


def is_host_server():
    os_info = dict()
    file_path = '/etc/os-release'
    if os.path.isfile(file_path):
        with open(file_path, 'r') as fp:
            for line in fp.readlines():
                pair = line.split('=')
                if len(pair) >= 2:
                    os_info[pair[0]] = pair[1].strip('\"\n')
        return 'NAME' in list(os_info.keys()) and 'cumulus' not in os_info['NAME'].lower()
    else:
        raise SettingsException('ERROR: file {} not found'.format(file_path))


def _signal_handler(signal, frame):
    pass

def check_cl_version(major):
    """Get the platform name of this device. Skip if its not Cumulus.
    """
    if os.path.exists(DefaultPaths.OS_RELEASE):
        conf = ConfigObj(DefaultPaths.OS_RELEASE)
        if conf.get('NAME') == Platform.CL:
            version = conf.get('VERSION', '0.0.0').lstrip(Platform.CL)
            if '~' in version:
                idx = version.index('~')
                version = version[:idx]
            cl_major = int(version.split('.')[0])
            # check if major versions are same
            return cl_major == major
    return False


def is_opta():
    return os.path.isfile('/etc/app-release')


def is_cl4():
    return check_cl_version(major=4) or check_cl_version(major=5)


def is_cl3():
    return check_cl_version(major=3)


def is_sonic():
    return os.path.isdir('/etc/sonic')


def is_chassis_card():
    file_path = '/usr/lib/cumulus/chassis-info'
    status = False
    if os.path.isfile(file_path):
        try:
            signal.signal(signal.SIGCHLD, _signal_handler)
            subprocess.check_call([file_path], stdout=False)
        except subprocess.CalledProcessError:
            status = False
        else:
            status = True
        finally:
            signal.signal(signal.SIGCHLD, signal.SIG_IGN)
    return status


is_subprocess_run_enabled = None


def subprocess_wrapper(cmd_list):
    supported_codec = ['utf-8', 'ISO-8859-1']
    global is_subprocess_run_enabled
    if is_subprocess_run_enabled is None:
        try:
            from subprocess import run  # flake8: noqa
            is_subprocess_run_enabled = True
        except ImportError:
            is_subprocess_run_enabled = False
    if is_subprocess_run_enabled is False:
        cmd_out = subprocess.check_output(cmd_list, stderr=subprocess.STDOUT)
    else:
        cmd_out = subprocess.run(cmd_list, stdout=subprocess.PIPE, \
            stderr=subprocess.PIPE, check=True).stdout

    for codec in supported_codec:
        try:
            cmd_out = cmd_out.decode(codec).strip()
            return cmd_out
        except:
            pass
    return None


def get_platform_type():

    if is_opta():
        return PlatformType.OPTA
    elif "arm" in get_cpu_architecture():
        return PlatformType.ARMEL
    elif is_cl3():
        return PlatformType.CL3
    elif is_cl4():
        return PlatformType.CL4
    elif is_sonic():
        return PlatformType.SONIC
    else:
        return PlatformType.DEFAULT


def timeit(f):
    @wraps(f)
    def wrap(*args, **kw):
        start_time = time.time()
        result = f(*args, **kw)
        end_time = time.time()
        print("func:{} took: {:2.6f} sec".format(f.__name__, end_time-start_time))
        return result
    return wrap

