#!/usr/bin/env python

# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
# Author: Daniel Walton, dwalton@cumulusnetworks.com
#

import logging
import re
import sys
import subprocess
from network_docopt.match import (
    custom_matchers,
    loose_matching_variables,
)
from network_docopt.tab import (
    merge_dict,
    tab_words_expand,
    tab_words_scrub_loose_options,
    tab_words_print_fish,
    tab_words_print_bash,
    tree_tab_words,
)

from network_docopt.node import Node, tree_nodes, red
from datetime import datetime
from pprint import pformat

log = logging.getLogger(__name__)


re_help_key_value = re.compile(r"""^(.*?):(.*)$""")


def get_bash_prompt():
    """
    This takes 40-50ms, we have not found a faster way to get the user's prompt

    This is a duplicate copy of tab.py:get_bash_prompt(), this copy is here
    until NetQ changes to import the copy in tab.py instead of this copy.
    """
    try:
        subprocess.check_output("/bin/echo /bin/false | /bin/bash -i",
                                stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        for line in e.output.splitlines():
            line = line.decode(encoding='UTF-8').strip()
            line = line.replace('/bin/false', '')

            if not line:
                continue

            if 'stty: standard input: Inappropriate ioctl for device' in line:
                continue

            return line
    return None


def get_network_docopt_info(argv):
    """
    Analyze argv and return (print_options, ended_with_space, argv)
    - print_options is True if TAB complete options should be printed.
    - ended_with_space is True if the user hit the spacebar prior to hitting TAB
    - argv is the original argv but with the '--completions X' at the end removed

    >>> get_network_docopt_info([])
    (False, False, [])

    >>> get_network_docopt_info(['nes'])
    (False, False, ['nes'])

    >>> get_network_docopt_info(['nes', 'show', 'games'])
    (False, False, ['nes', 'show', 'games'])

    >>> get_network_docopt_info(['nes', 'show', 'games', '--completions', '2'])
    (True, False, ['nes', 'show', 'games'])

    >>> get_network_docopt_info(['nes', 'show', 'games', '--completions', '3'])
    (True, True, ['nes', 'show', 'games'])
    """

    # Note if the last arg was '--completions'
    print_options = False
    ended_with_space = False
    len_argv = len(argv)

    if len_argv >= 2 and argv[-2] == '--completions':
        print_options = True
        init_completion_cword = int(argv[-1])

        # If the user did:
        #   "iface swp1 add <TAB>"
        #
        # then argv will be:
        #   ['./iface', 'swp1', 'add', '--completions', '3']
        #
        # but if they did
        #   "iface swp1 add<TAB>"
        #
        # then argv will be:
        #   ['./iface', 'swp1', 'add', '--completions', '2']
        #
        # We need to know how many keywords they entered so
        # subtract 1 for 'iface' and 2 for '--completions 2'
        len_argv_pre_tab_complete = len_argv - 3

        if init_completion_cword > len_argv_pre_tab_complete:
            ended_with_space = True
        else:
            ended_with_space = False

        argv = argv[0:-2]

    return (print_options, ended_with_space, argv)


def sort_for_humans(list_foo):
    """
    Sort list_foo in the way that humans expect.
    http://nedbatchelder.com/blog/200712/human_sorting.html

    >>> sort_for_humans(['swp10', 'swp2', 'swp1', 'swp20', 'swp0', 'swp11'])
    ['swp0', 'swp1', 'swp2', 'swp10', 'swp11', 'swp20']
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    list_foo.sort(key=alphanum_key)
    return list_foo


def sort_ifaces(ifnames):
    """
    Given a list of interface names, return a list of those names sorted
    in the following order:
        lo -> eth -> swp -> bonds, bridges, etc

    >>> sort_ifaces(['bonds', 'bridges', 'swp10', 'eth0', 'swp1', 'swp20', 'swp2', 'eth1', 'lo'])
    ['lo', 'eth0', 'eth1', 'swp1', 'swp2', 'swp10', 'swp20', 'bonds', 'bridges']
    """
    sorted_variables = []
    sorted_interfaces_lo = []
    sorted_interfaces_eth = []
    sorted_interfaces_swp = []
    sorted_interfaces_nonswp = []

    for x in sort_for_humans(ifnames):
        if x.startswith('<'):
            sorted_variables.append(x)

        elif x.startswith('lo'):
            sorted_interfaces_lo.append(x)

        elif x.startswith('eth'):
            sorted_interfaces_eth.append(x)

        elif x.startswith('swp'):
            sorted_interfaces_swp.append(x)

        else:
            sorted_interfaces_nonswp.append(x)

    return (sorted_variables +
            sorted_interfaces_lo +
            sorted_interfaces_eth +
            sorted_interfaces_swp +
            sorted_interfaces_nonswp)


def time_delta_ms(delta):
    """
    Return the number of milliseconds in a datetime.timedelta object
    """

    # 1 day is 86400000ms (60 seconds * 60 minutes * 24 hours * 1000ms)
    return (delta.days * 86400000) + (delta.seconds * 1000) + (delta.microseconds/1000)


def _tree_ascii(node, nodes_to_print, nodes_to_highlight, color):
    result = []

    if nodes_to_print is None or node in nodes_to_print:

        if nodes_to_highlight is not None and node in nodes_to_highlight:
            result.append(node.ascii(color=color))
        else:
            result.append(node.ascii())

        for child_node_key in sorted(node.children.keys()):
            child_node = node.children[child_node_key]
            result.extend(_tree_ascii(child_node, nodes_to_print, nodes_to_highlight, color))

    return result

def tree_ascii(node, nodes_to_print=None, nodes_to_highlight=None, color=None):
    return '\n'.join(_tree_ascii(node, nodes_to_print, nodes_to_highlight, color))


def tree_match_leaf_nodes(node, argv, argv_index, use_fuzzy):
    """
    Return a list of the leaf Nodes for all matching branches. This is used to find
    the branches that match argv.
    """
    leaf_nodes = []

    # 'node' matches the word at argv[argv_index]
    if node.match(argv[argv_index], use_fuzzy, keywords_only=False, for_tab_complete=False):

        # Is this the last word in argv?
        if argv_index == len(argv) - 1:

            # If this is the last node in the command OR this node has a child
            # branch where all nodes are optional then this branch matched
            if node.leaf_node() or node.ENTER_ok or not node.required:
                leaf_nodes.append(node)

        else:
            # <wildcard> nodes are special, do not advance to the next Node on a <wildcard> match
            if node.match_word_text.startswith('<wildcard'):
                leaf_nodes.extend(tree_match_leaf_nodes(node, argv, argv_index + 1, use_fuzzy))
            else:
                for child_node in list(node.children.values()):
                    leaf_nodes.extend(tree_match_leaf_nodes(child_node, argv, argv_index + 1, use_fuzzy))

    # 'node' does not match the word at argv[argv_index]. If the Node is not
    # required then we keep exploring. If the Node is required then we stop
    # exploring and return an empty leaf_nodes list.
    elif not node.required:
        for child_node in list(node.children.values()):
            leaf_nodes.extend(tree_match_leaf_nodes(child_node, argv, argv_index, use_fuzzy))

    return leaf_nodes


def tree_last_matching_argv_index(node, argv, argv_index, use_fuzzy):
    """
    This will only be called when the parse tree does not contain a matching
    branch. It is used to print the "^ Invalid value here" message.

    Walk the tree and return the argv_index of the last argv that matched a Node.
    """
    max_argv_index = 0

    try:
        argv_word = argv[argv_index]
    except IndexError:
        return argv_index - 1

    # 'node' matches the word at argv[argv_index]
    if node.match(argv_word, use_fuzzy, keywords_only=False, for_tab_complete=False):

        for child_node in list(node.children.values()):
            max_argv_index = max(max_argv_index, tree_last_matching_argv_index(child_node, argv, argv_index + 1, use_fuzzy))

    # 'node' does not match the word at argv[argv_index]. If the Node is not
    # required then we keep exploring. If the Node is required then we stop
    # exploring and return the previous argv_index as the last one that matched.
    else:

        if node.required:
            return argv_index - 1
        else:
            for child_node in list(node.children.values()):
                max_argv_index = max(max_argv_index, tree_last_matching_argv_index(child_node, argv, argv_index, use_fuzzy))

    return max_argv_index


class NetworkDocopt(object):
    """
    Heavily influenced by docopt, but designed to be a little more like
    a Networking CLI with partial word acceptance, IPv4 sanity checking, etc.
    """

    def __init__(self, docstring, daemon=False):
        self.docstring = docstring
        self.tab_complete_hook = {}
        self.tab_complete_policer = None
        self.helptext = {
            'global': {},
            'per-command': {},
        }
        self.program = None
        self.daemon = daemon
        self.root_node = None

        self.args = {}
        self.argv_expanded = []
        self.match = []

        self.build_parse_tree()

    def build_parse_tree(self):
        """
        Parse self.docstring to build a parse tree.
        """

        start_time1 = datetime.now()
        state = None
        found_help_option = False
        usage_commands = []
        help_line = None

        # Create a Command object for each line in the Usage section. We store
        # the lines in usage_commands so we can sort them...just to make
        # troubleshooting easier later
        for line in self.docstring.splitlines():
            line = line.strip()

            if line.startswith('Usage'):
                state = 'USAGE'
                continue

            elif line.startswith('Help'):
                state = 'HELP'
                continue

            elif line.startswith('Options'):
                state = 'HELP'
                continue

            # Ignore comments in the docstring
            elif line.startswith('#'):
                continue

            elif line == '':
                continue

            if state == 'USAGE':
                usage_commands.append(line)

                if self.program is None:
                    self.program = line.split()[0]
                    help_line = '%s help' % self.program

                if line == help_line:
                    found_help_option = True

            elif state == 'HELP':
                re_line = re_help_key_value.match(line)

                if re_line:
                    keyword = re_line.group(1).strip()
                    keyword_help = re_line.group(2).strip()

                    if keyword and keyword_help:
                        self.helptext['global'][keyword] = keyword_help

        if not found_help_option:
            usage_commands.append(help_line)

        end_time1 = datetime.now()
        start_time2 = end_time1

        # Build a tree of Node objects
        self.root_node = Node(0, None, self.program)

        for command in usage_commands:
            self.root_node.add_branch(command)

        # Build a tuple of nodes and leaf nodes
        self.nodes = tree_nodes(self.root_node)
        self.leaf_nodes = tuple([node for node in self.nodes if node.leaf_node()])

        # Now that we have constructed the tree we can populate what
        # the TAB choices are for each Node, etc.
        for node in self.nodes:
            node.populate_tab_words()

        end_time2 = datetime.now()
        log.debug("%dms to build the parse tree (%d Nodes), %dms to parse the __doc__, %dms to create/link the Nodes" %
            (time_delta_ms(end_time2 - start_time1), Node.count,
             time_delta_ms(end_time1 - start_time1),
             time_delta_ms(end_time2 - start_time2)))

    def re_init(self):
        """
        Re-init the parse tree
        """
        self.args = {}
        self.argv_expanded = []
        self.match = []

        for node in self.nodes:
            node.reset()

    def add_tag_matcher(self, keyword, matcher, is_loose_matcher):
        global custom_matchers
        global loose_matching_variables
        custom_matchers[keyword] = matcher

        if is_loose_matcher:
            loose_matching_variables.insert(0, keyword)

    def del_tag_matcher(self, keyword):
        global custom_matchers
        global loose_matching_variables

        if keyword in custom_matchers:
            del custom_matchers[keyword]

        if keyword in loose_matching_variables:
            loose_matching_variables.remove(keyword)

    def add_tab_complete_hook(self, keyword, completer):
        self.tab_complete_hook[keyword] = completer

    def del_tab_complete_hook(self, keyword):
        if keyword in self.tab_complete_hook:
            del self.tab_complete_hook[keyword]

    def get(self, keyword):
        return self.args.get(keyword)

    def get_keyword_help(self, keywords):
        assert isinstance(keywords, (tuple, list)), "keywords '%s' is %s, it must be a list or tuple" % (pformat(keywords), type(keywords))
        assert keywords, "keywords cannot be an empty list or tuple"
        keyword_leaf_nodes = []

        # Sort the keywords from longest to shortest
        keywords.sort(key=len, reverse=True)

        # Find all leaf_nodes whose command has all keywords
        for leaf_node in self.leaf_nodes:
            if leaf_node.command_has_keywords(keywords):
                keyword_leaf_nodes.append(leaf_node)

        if keyword_leaf_nodes:
            output = []

            for leaf_node in keyword_leaf_nodes:
                output.append('    ' + leaf_node.command())

            output = sorted(output)
            output.insert(0, 'The following commands contain keyword(s) \'%s\'\n' % "', '".join(keywords))
            output.append('')
            output.append('')
            return '\n'.join(output)
        else:
            return "ERROR: There are no commands with keyword(s) '%s'" % "', '".join(keywords)

    def tree_ascii_leaf_nodes(self, leaf_nodes, color=None):
        """
        Return the ascii tree for all branches to "leaf_nodes".
        """

        nodes_to_print = []
        nodes_to_highlight = []

        for leaf_node in leaf_nodes:
            nodes_to_print.append(leaf_node)

            if color is not None:
                nodes_to_highlight.append(leaf_node)

            nodes_to_print.extend(leaf_node.parent_nodes())

        nodes_to_print = set(nodes_to_print)
        return tree_ascii(self.root_node, nodes_to_print, nodes_to_highlight, color)

    def tree_ascii_keyword(self, keyword):
        """
        Return the ascii tree for all branches going through Node(s) with 'keyword'
        """
        nodes_to_print = []
        nodes_to_highlight = []

        for node in self.nodes:
            if node.has_word(keyword):
                nodes_to_highlight.append(node)
                nodes_to_print.append(node)
                nodes_to_print.extend(node.parent_nodes())
                nodes_to_print.extend(node.child_nodes())

        nodes_to_print = set(nodes_to_print)
        nodes_to_highlight = set(nodes_to_highlight)

        return tree_ascii(self.root_node, nodes_to_print, nodes_to_highlight)

    def add_per_command_helptext(self, command, command_helptext):

        # This should always be there but on the off chance that the
        # user deleted it, put it back.
        if 'per-command' not in self.helptext:
            self.helptext['per-command'] = {}
            log.warning("'per-command' was missing from self.helptext")

        if command in self.helptext['per-command']:
            self.helptext['per-command'][command] = merge_dict(self.helptext['per-command'][command], command_helptext)
        else:
            self.helptext['per-command'][command] = command_helptext

    def run(self, argv=None):
        """
        Given a command as a list of strings or None, attempt to match it with a valid command
        syntax.  Upon success, populate the dictionary self.args.

        Return a two-tuple of (bool, str).  The boolean indicates success or failure.  The
        string is a message for upstream consumption.
        """

        assert not (self.daemon and argv is None), "argv cannot be None when running as a daemon."
        assert argv is None or isinstance(argv, list)

        self.args = {}
        self.argv_expanded = []
        self.match = []

        self.argv = argv if (argv is not None) else sys.argv

        # argv[0] is the program name.
        if '/' in self.argv[0]:
            # The program name is "foo", for example, but the user ran it as "./foo".
            # Remove everything before the "/".
            self.argv[0] = self.argv[0].split('/')[-1]

        len_argv = len(self.argv)

        if len_argv == 1:
            # If the user did not provide any options, output the docstring.
            return (True, self.docstring)

        help_tuple = ('-h', '--help', 'help', 'hel', 'he', 'h')

        if len_argv == 2 and self.argv[1] in help_tuple:
            # net help
            self.args['help'] = True
            return (True, self.docstring)
        elif self.argv[1] in help_tuple:
            # net help KEYWORDS
            self.args['help'] = True
            return (True, self.get_keyword_help(self.argv[2:]))
        elif self.argv[-1] in help_tuple:
            # net KEYWORDS help
            self.args['help'] = True
            return (True, self.get_keyword_help(self.argv[1:-1]))
        else:

            #log.info("TREE:\n%s\n" % self.tree_ascii_keyword('summary'))
            matching_leaf_nodes = tree_match_leaf_nodes(self.root_node, self.argv, 0, use_fuzzy=False)

            if matching_leaf_nodes:
                #log.info("TREE:\n%s\n" % self.tree_ascii_leaf_nodes(matching_leaf_nodes))

                """
                Consider the commands:
                    foobar show interface <interface>
                    foobar show interface pluggables

                The input "foobar show int plug" matches both commands.  Loop over the matching
                branches, and keep the branches that are the least ambiguous.  Ideally, there
                will be one clear winner.
                """
                if len(matching_leaf_nodes) > 1:
                    min_ambiguity_score = None
                    min_ambiguity_score_leaf_nodes = []

                    for leaf_node in matching_leaf_nodes:
                        ambiguity_score = leaf_node.command_ambiguity_score()
                        #log.info("MATCH (%d): %s" % (ambiguity_score, leaf_node.command()))

                        if min_ambiguity_score is None or ambiguity_score < min_ambiguity_score:
                            min_ambiguity_score = ambiguity_score
                            min_ambiguity_score_leaf_nodes = [leaf_node, ]
                        elif ambiguity_score == min_ambiguity_score:
                            min_ambiguity_score_leaf_nodes.append(leaf_node)

                    #log.info("TREE:\n%s\n" % self.tree_ascii_leaf_nodes(matching_leaf_nodes))
                    #log.info("TREE:\n%s\n" % self.tree_ascii_leaf_nodes(min_ambiguity_score_leaf_nodes))
                    matching_leaf_nodes = min_ambiguity_score_leaf_nodes

                if len(matching_leaf_nodes) == 1:
                    leaf_node = matching_leaf_nodes[0]

                    # self.match must be populated with a list of the matching Node objects
                    self.match = list(leaf_node.parent_nodes_matching_only())
                    self.match.append(leaf_node)

                    # populate self.args and self.argv_expanded
                    leaf_node.command_populate_args(self.args)
                    self.argv_expanded = []

                    for node in self.match:
                        try:
                            self.argv_expanded.append(str(node.match_value))
                        except UnicodeEncodeError:
                            self.argv_expanded.append(node.match_value.encode('utf-8'))

                    #log.info("argv_expanded: %s" % self.argv_expanded)
                    #log.info("args:\n%s\n" % pformat(self.args))

                # We could not declare a clear winner.
                else:
                    log.warning("Ambiguous parse tree for \"{0}\"\n{1}".format(' '.join(self.argv), self.tree_ascii_leaf_nodes(matching_leaf_nodes)))
                    response = "Ambiguous parse tree\n\n{0}\n".format(self.tree_ascii_leaf_nodes(matching_leaf_nodes, color=red))
                    return (False, response)

            else:
                # No match.

                response = "Command not found.\n\n"

                # Output the given command and "^ Invalid value here" at the first word that didn't match.
                last_matching_argv_index = tree_last_matching_argv_index(self.root_node, self.argv, 0, use_fuzzy=False)
                response += "    {0}\n".format(' '.join(self.argv))
                response += "    {0}^ Invalid value here.\n".format(' ' * (len(' '.join(self.argv[0:last_matching_argv_index + 1])) + 1))

                # Try fuzzy matching to generate "Did you mean...?" suggestions.
                fuzzy_leaf_nodes = tree_match_leaf_nodes(self.root_node, self.argv, 0, use_fuzzy=True)

                if fuzzy_leaf_nodes:
                    # Found some fuzzy matches.
                    response += "\nDid you mean one of the following?\n"
                    for leaf_node in fuzzy_leaf_nodes:
                        response += leaf_node.command_did_you_mean() + '\n'
                else:
                    # We have no clue what command the user was trying to run.

                    help_output = self.get_keyword_help(self.argv[1:])

                    if "There are no commands with keyword" in help_output:
                        response += "\nUse \"{0} help KEYWORD(s)\" to list all options that use KEYWORD(s).\n".format(self.program)
                    else:
                        response += '\n' + help_output

                return (False, response)

        # Successfully parsed the command.
        return (True, '')

    def print_options(self, ended_with_space, options_style="bash", argv=None):
        """
        Print TAB complete options
        """

        self.argv = argv if (argv is not None) else sys.argv

        # argv[0] will be the program name, if the program name is foo but the
        # user ran it as ./foo remove everything before (and including) the '/'.
        if '/' in self.argv[0]:
            self.argv[0] = self.argv[0].split('/')[-1]

        last_argv = self.argv[-1]

        (tab_words, self.args, ENTER_ok, tab_complete_leaf_nodes) =\
            tree_tab_words(self.root_node, self.argv, self.args, ended_with_space)

        # Build a list of all of the leaf_node commands that were used for TAB
        # complete.  This is needed to provide per-command specific help text.
        leaf_node_commands = [leaf_node.command() for leaf_node in tab_complete_leaf_nodes]

        if tab_words:

            # Expand <interface> to a list of interfaces, etc.
            (tab_words, hook_helptext) = tab_words_expand(tab_words, ended_with_space, last_argv, self.tab_complete_hook, self.args)

            # The tab_complete_hook may need to add entries to our helptext dictionary.
            if hook_helptext:
                self.helptext["global"] = merge_dict(self.helptext["global"], hook_helptext)

            # Uncomment to log a warning for each keyword that does not have help text
            # for word in tab_words_without_helptext(tab_words, self.helptext['global']):
            # log.warning("%s does not have help text." % word)

            # Our child class may have defined a policer to give them final say on
            # what TAB words are presented.
            if self.tab_complete_policer:
                tab_words = self.tab_complete_policer(self.args, tab_words, ended_with_space, last_argv)

            tab_words = tab_words_scrub_loose_options(tab_words, ended_with_space, last_argv)

        if tab_words or ENTER_ok:

            # TAB choices must be printed differently for different shells.
            if options_style == "fish":
                tab_words_print_fish(tab_words, self.helptext, ENTER_ok)
            elif options_style == "bash":
                tab_words_print_bash(self.argv, last_argv, tab_words, ended_with_space,
                                     self.helptext, self.daemon, ENTER_ok,
                                     leaf_node_commands)
            else:
                raise RuntimeError("TAB complete is not supported for {0}.".format(options_style))


if __name__ == "__main__":
    import doctest

    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(filename)16s %(levelname)8s: %(message)s")
    log = logging.getLogger(__name__)

    # Color the errors and warnings in red
    logging.addLevelName(logging.ERROR, "\033[91m   %s\033[0m" % logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.WARNING, "\033[91m %s\033[0m" % logging.getLevelName(logging.WARNING))

    doctest.testmod()
