#!/usr/bin/env python

# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
# Author: Daniel Walton, dwalton@cumulusnetworks.com
#

import logging
import re
from ipaddress import ip_address, ip_network
from pprint import pformat

log = logging.getLogger(__name__)

TCP_PORT_MIN = 0
TCP_PORT_MAX = 65535
UDP_PORT_MIN = 0
UDP_PORT_MAX = 65535

loose_matching_variables = ['<interface>', '<text>', '<regex>', '<wildcard>', '']
custom_matchers = {}

tcp_flags = {
    'SYN' : 'Synchronize sequence numbers',
    'ACK' : 'Acknowledgment bit',
    'FIN' : 'No more data from sender',
    'RST' : 'Reset the connection',
    'URG' : 'Urgent pointer field is significant',
    'PSH' : 'Push function. Asks to push the buffered data to the receiving application',
    'ALL' : 'All TCP flags',
    'NONE': 'No TCP flags',
}

tcp_port_mapping = {
    'ftp'         : 'File Transfer Protocol',
    'ssh'         : 'Secure Shell',
    'telnet'      : 'Telnet',
    'smtp'        : 'Simple Mail Transfer Protocol',
    'domain'      : 'DNS - Domain Name System',
    'bootps'      : 'Bootstrap Protocol (BOOTP) server',
    'bootpc'      : 'Bootstrap Protocol (BOOTP) client',
    'dhcp-server' : 'Dynamic Host Configuration Protocol server',
    'dhcp-client' : 'Dynamic Host Configuration Protocol client',
    'http'        : 'Hypertext Transfer Protocol',
    'pop3'        : 'Post Office Protocol version 3',
    'imap2'       : 'Internet Message Access Protocol',
    'snmp'        : 'Simple Network Management Protocol',
    'snmp-trap'   : 'Simple Network Management Protocol trap',
    'bgp'         : 'Border Gateway Protocol',
    'ldap'        : 'Lightweight Directory Access Protocol',
    'https'       : 'Hypertext Transfer Protocol over TLS/SSL',
    'ldaps'       : 'Lightweight Directory Access Protocol over TLS/SSL',
    'clag'        : 'Multi-Chassis Link Aggregation',
    'msdp'        : 'Multicast Source Discovery Protocol',
}

udp_port_mapping = {
    'domain'      : 'DNS - Domain Name System',
    'bootps'      : 'Bootstrap Protocol (BOOTP) server',
    'bootpc'      : 'Bootstrap Protocol (BOOTP) client',
    'dhcp-server' : 'Dynamic Host Configuration Protocol server',
    'dhcp-client' : 'Dynamic Host Configuration Protocol client',
    'tftp'        : 'Trivial File Transfer Protocol',
    'ntp'         : 'Network Time Protocol',
    'snmp'        : 'Simple Network Management Protocol',
    'snmp-trap'   : 'Simple Network Management Protocol trap',
    'ldap'        : 'Lightweight Directory Access Protocol',
    'ldaps'       : 'Lightweight Directory Access Protocol over TLS/SSL',
    'bfd'         : 'Bidirectional Forwarding Detection',
    'bfd-echo'    : 'Bidirectional Forwarding Detection echo',
    'bfd-multihop': 'Bidirectional Forwarding Detection multihop',
    'lnv'         : 'Lightweight Network Virtualization',
}

dscp_mapping = {
    'best-effort' : 'drop probability N/A, routine',
    'af11' : 'drop probability low, priority',
    'af12' : 'drop probability medium, priority',
    'af13' : 'drop probability high, priority',
    'af21' : 'drop probability low, immediate',
    'af22' : 'drop probability medium, immediate',
    'af23' : 'drop probability high, immediate',
    'af31' : 'drop probability low, flash',
    'af32' : 'drop probability medium, flash',
    'af33' : 'drop probability high, flash',
    'af41' : 'drop probability low, flash override',
    'af42' : 'drop probability medium, flash override',
    'af43' : 'drop probability high, flash override',
    'cs1'  : 'cs1',
    'cs2'  : 'cs2',
    'cs3'  : 'cs3',
    'cs4'  : 'cs4',
    'cs5'  : 'cs5',
    'cs6'  : 'cs6',
    'cs7'  : 'cs7',
    'ef'   : 'drop probability N/A, critical',
}

re_mac_type1 = re.compile(r"""^(?:[0-9a-f]{2}[:-]){5}[0-9a-f]{2}$""", re.IGNORECASE)
re_mac_type2 = re.compile(r"""^[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}$""", re.IGNORECASE)
re_number_range = re.compile(r"""^(\d+)-(\d+)$""")
re_number_range_variable = re.compile(r"""^<(\d+)-(\d+)>$""")
re_hex_range_variable = re.compile(r"""^<0x(\w+)-0x(\w+)>$""")


class InvalidVariable(Exception):
    """
    Raised if a Node has a variable type that we do not support. For example
    say the Node attempts to use a <nes-characters> variable but support has
    not been added for <nes-characters>, this exception will be raised in that
    scenario.
    """
    pass


def is_loose_matching_variable(word):
    """
    There are variable types that will match on just about
    anything...<text>, <wildcard>, etc.

    Return True if `word` is considered a loose matching variable.

    >>> is_loose_matching_variable('<text>')
    True

    >>> is_loose_matching_variable('<interface>')
    True

    >>> is_loose_matching_variable('<wildcard>')
    True

    >>> is_loose_matching_variable('<regex>')
    True

    >>> is_loose_matching_variable('<text-foo>')
    True

    >>> is_loose_matching_variable('<interface-foo>')
    True

    >>> is_loose_matching_variable('<wildcard-foo>')
    True

    >>> is_loose_matching_variable('<regex-foo>')
    True

    >>> is_loose_matching_variable('homer')
    False

    >>> is_loose_matching_variable('<homer>')
    False
    """
    if word in loose_matching_variables:
        return True

    # <text-foo> counts as loose because <text> is loose
    for x in loose_matching_variables:

        # Chop the > at the end and add a '-'
        x = x[0:-1] + '-'

        if word.startswith(x):
            return True

    return False


def swap_letter_pairs(name):
    """
    Return a list of all variations of 'name' that can be made by swapping
    single pairs of letters.

    >>> swap_letter_pairs('zelda')
    ['ezlda', 'zleda', 'zedla', 'zelad']
    """
    names = []
    name = list(name)

    for index in range(len(name) - 1):
        char_at_index = name[index]
        char_at_index_plus_one = name[index+1]
        tmp = name[:]
        tmp[index] = char_at_index_plus_one
        tmp[index+1] = char_at_index
        names.append(''.join(tmp))
    return names


def drop_each_letter(name):
    """
    Return a list of all variations of 'name' that can be made by dropping
    each letter.

    >>> drop_each_letter('zelda')
    ['elda', 'zlda', 'zeda', 'zela', 'zeld']
    """
    names = []
    name = list(name)

    for index in range(len(name)):
        tmp = name[:]
        del tmp[index]
        names.append(''.join(tmp))
    return names


def match_keyword(keyword, text):
    """
    Return True if 'keyword' matches 'text'

    >>> match_keyword('zelda', 'zelda')
    True

    >>> match_keyword('zelda', 'mario')
    False

    >>> match_keyword('zelda', 'zel')
    False
    """
    assert isinstance(keyword, str), "keyword '%s' is %s, it must be str or unicode" % (pformat(keyword), type(keyword))
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return keyword == text


def match_keyword_startswith(keyword, text):
    """
    Return True if 'keyword' startswith 'text'

    >>> match_keyword_startswith('zelda', 'zelda')
    True

    >>> match_keyword_startswith('zelda', 'zeldaII')
    False

    >>> match_keyword_startswith('zelda', 'mario')
    False

    >>> match_keyword_startswith('zelda', 'zel')
    True

    >>> match_keyword_startswith('zel', 'zelda')
    False
    """
    assert isinstance(keyword, str), "keyword '%s' is %s, it must be str or unicode" % (pformat(keyword), type(keyword))
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return keyword.startswith(text)


def match_keyword_contains(keyword, text):
    """
    >>> match_keyword_contains('zelda', 'eld')
    True

    >>> match_keyword_contains('zelda', 'zel')
    True

    >>> match_keyword_contains('zelda', 'ario')
    False
    """
    return text in keyword


def match_keyword_fuzzy_lower(keyword, text):
    """
    Return True if 'keyword' matches 'text' if we lowercase 'text'

    >>> match_keyword_fuzzy_lower('zelda', 'zelda')
    True

    >>> match_keyword_fuzzy_lower('zelda', 'zElda')
    True

    >>> match_keyword_fuzzy_lower('zelda', 'mario')
    False
    """
    assert isinstance(keyword, str), "keyword '%s' is %s, it must be str or unicode" % (pformat(keyword), type(keyword))
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return match_keyword(keyword, text.lower())



def match_keyword_fuzzy_swap_letter_pairs(keyword, text):
    """
    Return True if 'keyword' matches 'text' by swapping pairs of letters.

    >>> match_keyword_fuzzy_swap_letter_pairs('zelda', 'zleda')
    True

    >>> match_keyword_fuzzy_swap_letter_pairs('zelda', 'lezda')
    False

    >>> match_keyword_fuzzy_swap_letter_pairs('zelda', 'zelda')
    False
    """
    assert isinstance(keyword, str), "keyword '%s' is %s, it must be str or unicode" % (pformat(keyword), type(keyword))
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    for typoed_word in swap_letter_pairs(keyword):
        if text == typoed_word:
            return True

    return False


def match_keyword_fuzzy_drop_each_letter(keyword, text):
    """
    Return True if 'keyword' matches 'text' by dropping a letter.

    >>> match_keyword_fuzzy_drop_each_letter('zelda', 'elda')
    True

    >>> match_keyword_fuzzy_drop_each_letter('zelda', 'zeld')
    True

    >>> match_keyword_fuzzy_drop_each_letter('zelda', 'zlda')
    True

    >>> match_keyword_fuzzy_drop_each_letter('zelda', 'mario')
    False

    >>> match_keyword_fuzzy_drop_each_letter('zelda', 'zelda')
    False
    """
    assert isinstance(keyword, str), "keyword '%s' is %s, it must be str or unicode" % (pformat(keyword), type(keyword))
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    for typoed_word in drop_each_letter(keyword):
        if text == typoed_word:
            return True

    return False


def match_ipv4_address(text):
    """
    Return True if text is an IPv4 address

    >>> match_ipv4_address('0.0.0.0')
    True

    >>> match_ipv4_address('1.2.3.4')
    True

    >>> match_ipv4_address('255.255.255.255')
    True

    >>> match_ipv4_address('0.0.0.0/32')
    False

    >>> match_ipv4_address('1.2.3.4/32')
    False

    >>> match_ipv4_address('255.255.255.255/32')
    False

    >>> match_ipv4_address('1.2.3.4.5')
    False

    >>> match_ipv4_address('256.255.255.255')
    False

    >>> match_ipv4_address('::')
    False

    >>> match_ipv4_address('0:0:0:0:0:0:0:0')
    False

    >>> match_ipv4_address('2001::1')
    False

    >>> match_ipv4_address('2001:0db8:85a3:0000:0000:8a2e:0370:7334')
    False

    >>> match_ipv4_address('2001:db8:85a3:0:0:8a2e:370:7334')
    False

    >>> match_ipv4_address('2001:db8:85a3::8a2e:370:7334')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    try:
        ip_address(text)
        return True
    except ValueError:
        return False


def match_ipv4_network(text):
    """
    Return True if text is an IPv4 network

    >>> match_ipv4_network('0.0.0.0')
    False

    >>> match_ipv4_network('1.2.3.4')
    False

    >>> match_ipv4_network('255.255.255.255')
    False

    >>> match_ipv4_network('0.0.0.0/32')
    True

    >>> match_ipv4_network('1.2.3.4/32')
    True

    >>> match_ipv4_network('255.255.255.255/32')
    True

    >>> match_ipv4_network('1.2.3.4.5/32')
    False

    >>> match_ipv4_network('256.255.255.255/32')
    False

    >>> match_ipv4_network('2001::1/0')
    False

    >>> match_ipv4_network('2001::1/64')
    False

    >>> match_ipv4_network('2001::1/128')
    False

    >>> match_ipv4_network('::/128')
    False

    >>> match_ipv4_network('0:0:0:0:0:0:0:0/64')
    False

    >>> match_ipv4_network('2001:0db8:85a3:0000:0000:8a2e:0370:7334/128')
    False

    >>> match_ipv4_network('2001:db8:85a3:0:0:8a2e:370:7334/128')
    False

    >>> match_ipv4_network('2001:db8:85a3::8a2e:370:7334/128')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if '/' not in text:
        return False

    if match_ipv4_address(text):
        return False

    try:
        ip_network(text, strict=False)
        return True
    except ValueError:
        return False


def match_ipv6_address(text):
    """
    Return True if text is an IPv6 address

    >>> match_ipv6_address('::')
    True

    >>> match_ipv6_address('0:0:0:0:0:0:0:0')
    True

    >>> match_ipv6_address('2001::1')
    True

    >>> match_ipv6_address('2001:0db8:85a3:0000:0000:8a2e:0370:7334')
    True

    >>> match_ipv6_address('2001:db8:85a3:0:0:8a2e:370:7334')
    True

    >>> match_ipv6_address('2001:db8:85a3::8a2e:370:7334')
    True

    >>> match_ipv6_address('2001::1/64')
    False

    >>> match_ipv6_address('0:0:0:0:0:0:0:0:0')
    False

    >>> match_ipv6_address('0.0.0.0')
    False

    >>> match_ipv6_address('1.2.3.4')
    False

    >>> match_ipv6_address('255.255.255.255')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    try:
        ip_address(text)
        return True
    except ValueError:
        return False


def match_ipv6_network(text):
    """
    Return True if text is an IPv6 network

    >>> match_ipv6_network('2001::1/0')
    True

    >>> match_ipv6_network('2001::1/64')
    True

    >>> match_ipv6_network('2001::1/128')
    True

    >>> match_ipv6_network('::/128')
    True

    >>> match_ipv6_network('0:0:0:0:0:0:0:0/64')
    True

    >>> match_ipv6_network('2001:0db8:85a3:0000:0000:8a2e:0370:7334/128')
    True

    >>> match_ipv6_network('2001:db8:85a3:0:0:8a2e:370:7334/128')
    True

    >>> match_ipv6_network('2001:db8:85a3::8a2e:370:7334/128')
    True

    >>> match_ipv6_network('::')
    False

    >>> match_ipv6_network('0:0:0:0:0:0:0:0')
    False

    >>> match_ipv6_network('2001::1')
    False

    >>> match_ipv6_network('2001::1/129')
    False

    >>> match_ipv6_network('2001:0db8:85a3:0000:0000:8a2e:0370:7334')
    False

    >>> match_ipv6_network('2001:db8:85a3:0:0:8a2e:370:7334')
    False

    >>> match_ipv6_network('2001:db8:85a3::8a2e:370:7334')
    False

    >>> match_ipv6_network('0:0:0:0:0:0:0:0:0')
    False

    >>> match_ipv6_network('0.0.0.0/32')
    False

    >>> match_ipv6_network('1.2.3.4/32')
    False

    >>> match_ipv6_network('255.255.255.255/32')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if '/' not in text:
        return False

    if match_ipv6_address(text):
        return False

    try:
        ip_network(text, strict=False)
        return True
    except ValueError:
        return False


def match_ip_address(text):
    """
    Return True if text is an IPv4 or IPv6 address
    """
    return match_ipv4_address(text) or match_ipv6_address(text)


def match_ip_network(text):
    """
    Return True if text is an IPv4 or IPv6 network
    """
    return match_ipv4_network(text) or match_ipv6_network(text)


def match_mac_address(text):
    """
    Return True if text is a MAC address

    The first regex will match
        3D:F2:C9:A6:B3:4F
        3D-F2-C9-A6-B3-4F

    The second regex will match
        1234.abcd.1234

    >>> match_mac_address('3D:F2:C9:A6:B3:4F')
    True

    >>> match_mac_address('3D-F2-C9-A6-B3-4F')
    True

    >>> match_mac_address('1234.abcd.1234')
    True

    >>> match_mac_address('3D:F2:C9:A6:B3:4F:AA')
    False

    >>> match_mac_address('3D-F2-C9-A6-B3-4F-AA')
    False

    >>> match_mac_address('1234.abcd.1234.abcd')
    False

    >>> match_mac_address('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return (re_mac_type1.match(text) is not None) or (re_mac_type2.match(text) is not None)


def match_dscp(text):
    """
    Return True if text is a valid dscp value

    >>> match_dscp('af11')
    True

    >>> match_dscp('ef')
    True

    >>> match_dscp('foo')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return text in dscp_mapping


def match_dscp_startswith(text):
    """
    Return True if text startswith valid dscp value

    >>> match_dscp_startswith('af11')
    True

    >>> match_dscp_startswith('af1')
    True

    >>> match_dscp_startswith('a')
    True

    >>> match_dscp_startswith('ef')
    True

    >>> match_dscp_startswith('e')
    True

    >>> match_dscp_startswith('x')
    False

    >>> match_dscp_startswith('xyz')
    False

    >>> match_dscp_startswith('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    if match_dscp(text):
        return True

    for dscp_option in dscp_mapping:
        if dscp_option.startswith(text):
            return True

    return False


def match_tcp_flag(text):
    """
    Return True if text is a valid TCP flag

    >>> match_tcp_flag('SYN')
    True

    >>> match_tcp_flag('ACK')
    True

    >>> match_tcp_flag('foo')
    False

    >>> match_tcp_flag('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return text in tcp_flags


def match_tcp_flags(text):
    """
    Return True if text is a comma separated list of valid TCP flags

    >>> match_tcp_flags('SYN')
    True

    >>> match_tcp_flags('SYN,ACK')
    True

    >>> match_tcp_flags('SYN,foo,ACK')
    False

    >>> match_tcp_flags('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    for flag in text.split(','):
        if not match_tcp_flag(flag):
            return False

    return True


def match_tcp_flag_startswith(text):
    """
    Return True if text startswith a valid TCP flag

    >>> match_tcp_flag_startswith('SYN')
    True

    >>> match_tcp_flag_startswith('SY')
    True

    >>> match_tcp_flag_startswith('S')
    True

    >>> match_tcp_flag_startswith('FOO')
    False

    >>> match_tcp_flag_startswith('s')
    False

    >>> match_tcp_flag_startswith('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    if match_tcp_flag(text):
        return True

    for tcp_flag_option in tcp_flags:
        if tcp_flag_option.startswith(text):
            return True

    return False


def match_tcp_port_mapping(text):
    """
    Return True if text is a valid TCP port mapping

    >>> match_tcp_port_mapping('bgp')
    True

    >>> match_tcp_port_mapping('zelda')
    False

    >>> match_tcp_port_mapping('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return text in tcp_port_mapping


def match_tcp_port_mapping_startswith(text):
    """
    Return True if text startswith a valid TCP port mapping

    >>> match_tcp_port_mapping_startswith('bgp')
    True

    >>> match_tcp_port_mapping_startswith('bg')
    True

    >>> match_tcp_port_mapping_startswith('zelda')
    False

    >>> match_tcp_port_mapping_startswith('zel')
    False

    >>> match_tcp_port_mapping_startswith('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    if match_tcp_port_mapping(text):
        return True

    for tcp_port_option in tcp_port_mapping:
        if tcp_port_option.startswith(text):
            return True

    return False


def match_udp_port_mapping(text):
    """
    Return True if text is a valid UDP port mapping

    >>> match_udp_port_mapping('tftp')
    True

    >>> match_udp_port_mapping('zelda')
    False

    >>> match_udp_port_mapping('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return text in udp_port_mapping


def match_udp_port_mapping_startswith(text):
    """
    Return True if text startswith a valid UDP port mapping

    >>> match_udp_port_mapping_startswith('tftp')
    True

    >>> match_udp_port_mapping_startswith('tft')
    True

    >>> match_udp_port_mapping_startswith('zelda')
    False

    >>> match_udp_port_mapping_startswith('zel')
    False

    >>> match_udp_port_mapping_startswith('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    if match_udp_port_mapping(text):
        return True

    for udp_port_option in udp_port_mapping:
        if udp_port_option.startswith(text):
            return True

    return False


def match_number_in_range(text, min_value, max_value):
    """
    Return True if text is a number in the range defined by min_value/max_value

    >>> match_number_in_range('0', 0, 50)
    True

    >>> match_number_in_range('37', 0, 50)
    True

    >>> match_number_in_range('50', 0, 50)
    True

    >>> match_number_in_range('-1', 0, 50)
    False

    >>> match_number_in_range('-1', -10, 10)
    True

    >>> match_number_in_range('0', -10, 10)
    True

    >>> match_number_in_range('0x600', 1536, 2000)
    True

    >>> match_number_in_range('0x600', 1537, 2000)
    False

    >>> match_number_in_range('0xFFFF', 0, 65535)
    True

    >>> match_number_in_range('0xFFFF', 0, 65534)
    False

    >>> match_number_in_range('57', 0, 50)
    False

    >>> match_number_in_range('0.0.0.0', 0, 50)
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    assert isinstance(min_value, int), "min_value '%s' is %s, it must be an int" % (pformat(min_value), type(min_value))
    assert isinstance(max_value, int), "max_value '%s' is %s, it must be an int" % (pformat(max_value), type(min_value))

    if text.startswith('0x'):
        text = str(int(text, 16))

    if text.lstrip('-').isdigit():
        return min_value <= int(text) <= max_value
    else:
        return False


def match_number_range(text):
    """
    Return True if text is a positive numeric range

    >>> match_number_range('1-1')
    True

    >>> match_number_range('1-10')
    True

    >>> match_number_range('01-0010')
    True

    >>> match_number_range('10-1')
    False

    >>> match_number_range('foo')
    False

    >>> match_number_range('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    match = re_number_range.match(text)

    if match is not None:
        start = int(match.group(1))
        stop = int(match.group(2))

        return start <= stop

    return False


def match_number_range_fuzzy(text):
    """
    The same as match_number_range just without the check that start <= stop

    >>> match_number_range_fuzzy('1-1')
    True

    >>> match_number_range_fuzzy('1-10')
    True

    >>> match_number_range_fuzzy('10-1')
    True

    >>> match_number_range_fuzzy('foo')
    False

    >>> match_number_range_fuzzy('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return re_number_range.match(text) is not None


def match_number_range_list(text):
    """
    Return True if text is a comma separated list of numbers and/or number
    ranges such as 10,20-30,40,50,60-70

    >>> match_number_range_list('')
    False

    >>> match_number_range_list('10')
    True

    >>> match_number_range_list('20-30')
    True

    >>> match_number_range_list('40-30')
    False

    >>> match_number_range_list('foo')
    False

    >>> match_number_range_list('10,20-30,40,50,60-70')
    True

    >>> match_number_range_list('10,40-30,40,50,60-70')
    False

    >>> match_number_range_list('10,20-30,40,foo,60-70')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    for x in text.split(','):
        if not x.isdigit() and not match_number_range(x):
            return False

    return True


def match_wildcard(text):
    """
    Return True, <wildcard> can match anything

    >>> match_wildcard('zelda')
    True
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return True


def match_regex(text):
    """
    Return True if text is a valid regular expression

    >>> match_regex('zelda')
    True

    >>> match_regex('[0-9]++')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    try:
        re.compile(text)
        return True
    except Exception:
        return False


def match_mac_ipnetwork_list(text):
    """
    Return True if text is a MAC address followed by a list of IP networks

    >>> match_mac_ipnetwork_list('00:11:22:33:44:01 10.0.0.1/24 10.0.1.2/24')
    True

    >>> match_mac_ipnetwork_list('00:11:22:33:44:01 10.0.0.1/24')
    True

    >>> match_mac_ipnetwork_list('00:11:22:33:44:01 10.0.0.1 10.0.1.2')
    False

    >>> match_mac_ipnetwork_list('00:11:22:33:44:01 10.0.0.1')
    False

    >>> match_mac_ipnetwork_list('10.0.0.1 00:11:22:33:44:01')
    False

    >>> match_mac_ipnetwork_list('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    words = text.split()

    if len(words) >= 2:
        first_word = words[0]

        if match_mac_address(first_word):

            for word in words[1:]:
                if not match_ip_network(word):
                    return False

            return True

    return False


def match_number_ipv4address_list(text):
    """
    Return True if text is a list of number=ipv4_address

    >>> match_number_ipv4address_list('100=10.0.100.1 101=10.101.1.1')
    True

    >>> match_number_ipv4address_list('100=10.0.100.1,101=10.101.1.1')
    True

    >>> match_number_ipv4address_list('100=10.0.100.1')
    True

    >>> match_number_ipv4address_list('100=foo 101=bar')
    False

    >>> match_number_ipv4address_list('100=10.0.100.1/24 101=10.101.1.1/32')
    False

    >>> match_number_ipv4address_list('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    for x in text.replace(',', ' ').split():
        if '=' in x:
            (number, ip) = x.split('=')

            if not number.isdigit():
                return False

            if not match_ipv4_address(ip):
                return False

        else:
            return False

    return True


def match_ipv4address_vrf_text(text):
    """
    "1.1.1.1 vrf FOO" but the "vrf FOO" is optional

    >>> match_ipv4address_vrf_text('1.1.1.1 vrf FOO')
    True

    >>> match_ipv4address_vrf_text('1.1.1.1')
    True

    >>> match_ipv4address_vrf_text('1.1.1.1/32 vrf FOO')
    False

    >>> match_ipv4address_vrf_text('1.1.1.1/32')
    False

    >>> match_ipv4address_vrf_text('vrf FOO')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    words = text.split()

    if len(words) == 1:
        return match_ipv4_address(words[0])

    elif len(words) == 3:
        return (match_ipv4_address(words[0]) and
                (match_keyword('vrf', words[1]) or match_keyword_startswith('vrf', words[1])) and
                match_text(words[2]))

    return False


def ifname_is_valid(ifname, honor_if_namesize=True):
    """
    Return True if ifname is a valid interface name as defined in the kernel's dev_valid_name().
    Valid interface names can have up to 15 visible characters plus a null terminator.
        http://stackoverflow.com/questions/24932172/what-length-can-a-network-interface-name-have#29398765

    honor_if_namesize - If "Honor Interface Namesize" is True we enforce the
    IF_NAMESIZE character limit.

    >>> ifname_is_valid('lo')
    True

    >>> ifname_is_valid('eth0')
    True

    >>> ifname_is_valid('swp7')
    True

    >>> ifname_is_valid('swp7.1')
    True

    >>> ifname_is_valid('swp7s0')
    True

    >>> ifname_is_valid('swp7s0.1')
    True

    # IF_NAMESIZE characters
    >>> ifname_is_valid('1234567890abcde')
    True

    # More than IF_NAMESIZE characters
    >>> ifname_is_valid('1234567890abcdef')
    False

    >>> ifname_is_valid('1234567890abcdef', False)
    True

    # ILLEGAL_CHARACTERS
    >>> ifname_is_valid('foo/bar')
    False

    >>> ifname_is_valid('foo bar')
    False

    # ILLEGAL_NAMES
    >>> ifname_is_valid('.')
    False

    >>> ifname_is_valid(':')
    False

    >>> ifname_is_valid('37')
    False

    >>> ifname_is_valid('1.1.1.1')
    False

    >>> ifname_is_valid('1.1.1.1/32')
    False

    >>> ifname_is_valid('1234.1234.1234')
    False
    """
    assert isinstance(ifname, str), "ifname '%s' is %s, it must be str or unicode" % (pformat(ifname), type(ifname))

    if not ifname:
        return False

    IF_NAMESIZE = 15
    ILLEGAL_CHARACTERS = ('/', ' ', '%', '*', '?', '^', '$', '+', '[', ']', '{', '}', '(', ')', '\\')
    ILLEGAL_NAMES = ('.', ':')  # Illegal names comprised of legal characters

    if ifname in ILLEGAL_NAMES:
        log.debug("'{0}' is an invalid interface name (cannot be '.' nor ':').".format(ifname))
        return False

    for character in ILLEGAL_CHARACTERS:
        if character in ifname:
            log.debug("'{0}' is an invalid interface name (cannot use '{1}').".format(ifname, "', '".join(ILLEGAL_CHARACTERS)))
            return False

    if honor_if_namesize and len(ifname) > IF_NAMESIZE:
        log.debug("'{0}' is an invalid interface name (more than {1} characters).".format(ifname, IF_NAMESIZE))
        return False

    # While these are technically valid interface names in linux, for the
    # purposes of network-docopt we want to treat these as illegal
    if ifname.isdigit():
        return False

    if match_ip_address(ifname):
        return False

    if match_ip_network(ifname):
        return False

    if match_mac_address(ifname):
        return False

    return True


def ifname_is_glob(ifname):
    """
    Return True if the given name refers to multiple interfaces (a.k.a. ifname is a glob).
    Otherwise, return False.

    >>> ifname_is_glob('')
    False

    >>> ifname_is_glob('-')
    False

    >>> ifname_is_glob(',')
    False

    >>> ifname_is_glob('not a glob')
    False

    >>> ifname_is_glob('vni-10')
    False

    >>> ifname_is_glob('peer-group')
    False

    >>> ifname_is_glob('swp1-5')
    True

    >>> ifname_is_glob('swp1,5')
    True

    >>> ifname_is_glob('swp1-5,8')
    True

    >>> ifname_is_glob('swp1,swp2,swp3')
    True

    >>> ifname_is_glob('swp1-4,6,10-12')
    True

    >>> ifname_is_glob('swp-5')
    False

    >>> ifname_is_glob('peer-group')
    False

    >>> ifname_is_glob('lo')
    False

    >>> ifname_is_glob('swp7')
    False

    >>> ifname_is_glob('swp7.1')
    False

    >>> ifname_is_glob('swp7.1,3')
    True

    >>> ifname_is_glob('swp7.1-3')
    True

    >>> ifname_is_glob('swp7s1')
    False

    >>> ifname_is_glob('swp7s1-3')
    True

    >>> ifname_is_glob('swp7s1.1-3')
    True

    >>> ifname_is_glob('lo,bridge')
    True

    >>> ifname_is_glob('bridge,lo')
    True

    >>> ifname_is_glob('vxlan10010-10013')
    True

    >>> ifname_is_glob('vni-1010,1020,1030,1040,1050')
    True

    >>> ifname_is_glob('bridge,vxlan10010-10013')
    True

    >>> ifname_is_glob('host4_bond')
    False

    >>> ifname_is_glob('host4-bond')
    False

    >>> ifname_is_glob('host4-5bond')
    False

    >>> ifname_is_glob('')
    False

    >>> ifname_is_glob('server09,server11,13')
    True

    >>> ifname_is_glob('po1ntnxprodnd09,po1ntnxprodnd11,13')
    True

    >>> ifname_is_glob('chassis01-lc1-1')
    False

    >>> ifname_is_glob('chassis01-lc1-4')
    True

    >>> ifname_is_glob('chassis01-lc4-1')
    False

    >>> ifname_is_glob('bond-leaf-a1-2')
    True

    >>> ifname_is_glob('bond-leaf-1-2')
    False

    >>> ifname_is_glob('bond-leaf-01-02')
    False
    """

    assert isinstance(ifname, str), "ifname={0}".format(ifname)

    # The empty string and strings with spaces are not globs.
    if not ifname or ' ' in ifname:
        return False

    if re.search(r"""\S,\S""", ifname) is not None:
        # Strings with comma-separated components are always a glob.
        return True

    # Strings with hyphens might be globs.
    re_range = re.search(r"""(?<!-)(\d+)-(\d+)(,|$)""", ifname)

    if re_range is not None:
        start_range = re_range.group(1)
        end_range = re_range.group(2)

        if (len(start_range) > 1 and start_range.startswith('0')) or end_range.startswith('0'):
            # Valid ranges do not contain lead zeros.  '0' is not valid as the end range.
            return False

        if int(end_range) > int(start_range):
            return True

    return False


def match_interface(text):
    return ifname_is_valid(text) or ifname_is_glob(text)


def match_interface_fuzzy(text):
    """
    The same as match_interface but with the 15 character
    IF_NAMESIZE restriction relaxed
    """
    return ifname_is_valid(text, False) or ifname_is_glob(text)


def match_interface_list(text):
    """
    An <interface-list> is another name for <interface> given
    our support of globs for <interface>
    """
    return ifname_is_valid(text) or ifname_is_glob(text)


def match_interface_number_list(text):
    """
    Return True if text is list of interface=number

    >>> match_interface_number_list('swp1=100 swp2=100')
    True

    >>> match_interface_number_list('swp1=100,swp2=100')
    True

    >>> match_interface_number_list('swp1=foo swp2=100')
    False

    >>> match_interface_number_list('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    for x in text.replace(',', ' ').split():
        if '=' in x:
            (ifname, number) = x.split('=')
            ifname = ifname.strip()
            number = number.strip()

            if not match_interface(ifname):
                return False

            if not number.isdigit():
                return False

        else:
            return False


    return True


def match_number_interface_list(text):
    """
    Return True if text is a number followed by a list of interfaces

    >>> match_number_interface_list('4 swp1 swp2')
    True

    >>> match_number_interface_list('swp0 swp1 swp2')
    False

    >>> match_number_interface_list('4')
    False

    >>> match_number_interface_list('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    words = text.replace(',', ' ').split()

    if len(words) >= 2:
        first_word = words[0]

        if first_word.isdigit():

            for ifname in words[1:]:
                if not match_interface(ifname):
                    return False

            return True

    return False


def match_interface_key_value_list(text, valid_values):
    """
    Return True if text is in valid_values.
    swp1=yes swp2=no swp3=1 swp4=0

    Return True if every token in text (split by whitespace) is of the
    form 'key=value' where value is in valid_values.

    If neither of those is the case, return False.

    >>> match_interface_key_value_list('yes', ['no', 'yes', 'auto'])
    False

    >>> match_interface_key_value_list('zelda', ['no', 'yes', 'auto'])
    False

    >>> match_interface_key_value_list('swp1=yes', ('no', 'yes', 'auto'))
    True

    >>> match_interface_key_value_list('swp1=yes,swp2=no', ('no', 'yes', 'auto'))
    True

    >>> match_interface_key_value_list('swp1=yes swp2=no', ('no', 'yes', 'auto'))
    True

    >>> match_interface_key_value_list('swp1=zelda', ('no', 'yes', 'auto'))
    False

    >>> match_interface_key_value_list('swp1=yes swp2=zelda', ('no', 'yes', 'auto'))
    False

    >>> match_interface_key_value_list('', ('no', 'yes', 'auto'))
    False

    >>> match_interface_key_value_list('swp1=yes swp2=zelda', ())
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    assert isinstance(valid_values, (list, tuple))

    if not text:
        return False

    if not valid_values:
        return False

    for x in text.replace(',', ' ').split():
        if '=' in x:
            (ifname, value) = x.split('=')

            if not match_interface(ifname):
                return False

            if value not in valid_values:
                return False

        else:
            return False

    return True


def match_vrf(text):
    return ifname_is_valid(text)


def match_decimal(text):
    """
    Return True if text is a decimal number

    >>> match_decimal('37')
    True

    >>> match_decimal('0')
    True

    >>> match_decimal('0037')
    False

    >>> match_decimal('-10')
    True

    >>> match_decimal('0x1234')
    False

    >>> match_decimal('zelda')
    False

    >>> match_decimal('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
    return text.lstrip('-').isdigit() and (text == '0' or not text.startswith('0'))


def match_hex(text):
    """
    Return True if text is a hex number

    >>> match_hex('37')
    True

    >>> match_hex('0x1234')
    True

    >>> match_hex('fe')
    True

    >>> match_hex('zelda')
    False

    >>> match_hex('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    try:
        int(text, 16)
        return True
    except ValueError:
        return False


def match_text(text):
    """
    Return True if text does not contain a whitespace

    >>> match_text('zelda')
    True

    >>> match_text('zelda mario')
    False

    >>> match_text('')
    False
    """
    assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))

    if not text:
        return False

    return ' ' not in text


def match_decimal_value(text1, text2):
    """
    Return True if the values of text1 and text2 are equal

    >>> match_decimal_value('37', '37')
    True

    >>> match_decimal_value('0', '0')
    True

    >>> match_decimal_value('-37', '-37')
    True

    >>> match_decimal_value('37', '38')
    False

    >>> match_decimal_value('', '')
    False

    >>> match_decimal_value('37', '')
    False

    >>> match_decimal_value('', '37')
    False
    """
    assert isinstance(text1, str), "text1 '%s' is %s, it must be str or unicode" % (pformat(text1), type(text1))
    assert isinstance(text2, str), "text2 '%s' is %s, it must be str or unicode" % (pformat(text2), type(text2))

    if (text1 == '' or
        text2 == '' or
        not text1.lstrip('-').isdigit() or
        not text2.lstrip('-').isdigit()):
        return False

    return match_decimal(text1) and match_decimal(text2) and int(text1) == int(text2)


def match_decimal_value_startswith(text1, text2):
    """
    Return True if text1 and text2 are decimals and text1 startswith text2

    >>> match_decimal_value_startswith('1234', '123')
    True

    >>> match_decimal_value_startswith('0', '0')
    True

    This should be False because 0123 fails match_decimal()
    >>> match_decimal_value_startswith('0123', '0')
    False

    >>> match_decimal_value_startswith('-1234', '-123')
    True

    >>> match_decimal_value_startswith('1234', '456')
    False
    """
    assert isinstance(text1, str), "text1 '%s' is %s, it must be str or unicode" % (pformat(text1), type(text1))
    assert isinstance(text2, str), "text2 '%s' is %s, it must be str or unicode" % (pformat(text2), type(text2))

    if (text1 == '' or
        text2 == '' or
        not text1.lstrip('-').isdigit() or
        not text2.lstrip('-').isdigit()):
        return False

    return match_decimal(text1) and match_decimal(text2) and text1.startswith(text2)


def is_tab_word(tab_words, argv_word):
    """
    Return True if argv_word matches any of tab_words.

    Example: Your parse tree has the following command
        nes show [<text>] [json]

    And the user enters 'net show json'.  We do not want '<text>' to match
    against 'json' so we need to exclude it from the list of things that
    can match for <text>. is_tab_word() returns True for this scenario, this
    allows us to match 'json' to the 'json' variable instead of to the '<text>'
    variable.

    >>> is_tab_word(['mario', 'zelda'], 'zelda')
    True

    >>> is_tab_word(['mario', 'zelda'], 'zel')
    True

    >>> is_tab_word(['mario', 'zelda'], 'megaman')
    False
    """

    for tab_word in tab_words:
        if not tab_word.startswith('<'):
            if match_keyword_startswith(tab_word, argv_word):
                return True

    return False


def match_word(text, argv_word, use_fuzzy, for_tab_complete, for_keyword_help, tab_words):
    """
    Evaluate argv_word against text where text can be
    - A variable: <ipv4>, <interface>, etc
    - A keyword: ip, address, bgp, etc
    - A number: 37, 100, etc

    Return True if argv_word matches text
    """

    matched = False
    value = None
    value_raw = None
    fuzzy = False

    if text.startswith('<'):

        # A client may want to use a new <foo> tag that has some special meaning.
        # We do not want to clutter network docopt with support for <foo>, so
        # provide a function hook.  'custom_matchers' is a global dictionary.
        if text in custom_matchers:
            matcher = custom_matchers[text]

            if matcher(argv_word, False, for_tab_complete):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and matcher(argv_word, True, for_tab_complete):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        # "1.1.1.1 vrf FOO" but the "vrf FOO" is optional
        elif text == '<ipv4-vrf-text>':

            if match_ipv4address_vrf_text(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # ipv4 or ipv6 address
        elif text == '<ip>' or text.startswith('<ip-'):

            if match_ip_address(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ip_network(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        # ipv4 or ipv6 address with prefix length
        elif text == '<ip/prefixlen>':

            if match_ip_network(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ip_address(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<ipv4>' or text.startswith('<ipv4-'):

            if match_ipv4_address(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ipv4_network(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<ipv4/prefixlen>':

            if match_ipv4_network(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ipv4_address(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<ipv6>' or text.startswith('<ipv6-'):

            if match_ipv6_address(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ipv6_network(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<ipv6/prefixlen>':

            if match_ipv6_network(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_ipv6_address(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<dscp>':

            if match_dscp(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif for_tab_complete and match_dscp_startswith(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<tcp-flags'):

            if match_tcp_flag(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif match_tcp_flags(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif for_tab_complete and match_tcp_flag_startswith(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<tcp-port'):

            if match_tcp_port_mapping(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif match_number_in_range(argv_word, TCP_PORT_MIN, TCP_PORT_MAX):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif for_tab_complete and match_tcp_port_mapping_startswith(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<udp-port'):

            if match_udp_port_mapping(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif match_number_in_range(argv_word, UDP_PORT_MIN, UDP_PORT_MAX):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif for_tab_complete and match_udp_port_mapping_startswith(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text == '<number-range>':

            if match_number_range(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif use_fuzzy and match_number_range_fuzzy(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

        elif text == '<number-range-list>' or text == '<number-comma-range-list>':

            if match_number_range_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<wildcard'):

            if match_wildcard(argv_word) and not is_tab_word(tab_words, argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<regex'):

            if match_regex(argv_word) and not is_tab_word(tab_words, argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # 00:11:22:33:44:01 10.0.0.1/24 10.0.1.2/24
        elif text == '<mac-ip/prefixlen-list>':

            if match_mac_ipnetwork_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text == '<mac>' or text.startswith('<mac-'):

            if match_mac_address(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # 100=10.0.100.1 101=10.101.1
        elif text == '<number-ipv4-list>':

            if match_number_ipv4address_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1 swp2 swp3
        elif text == '<interface-list>':

            if match_interface_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1=100 swp2=100
        elif text == '<interface-range-list>':

            if match_interface_number_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # 4 swp1 swp2
        elif text == '<number-interface-list>':

            if match_number_interface_list(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1=yes swp2=no
        elif text == '<interface-yes-no-list>':

            if match_interface_key_value_list(argv_word, ('yes', 'no')):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1=yes swp2=no swp3=1 swp4=0
        elif text == '<interface-yes-no-0-1-list>':

            if match_interface_key_value_list(argv_word, ('yes', 'no', '0', '1')):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1=on swp2=off
        elif text == '<interface-on-off-list>':

            if match_interface_key_value_list(argv_word, ('on', 'off')):
                value = argv_word
                value_raw = argv_word
                matched = True

        # swp1=yes swp2=no swp3=auto
        elif text == '<interface-yes-no-auto-list>':

            if match_interface_key_value_list(argv_word, ('yes', 'no', 'auto')):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text.startswith('<interface'):

            if not is_tab_word(tab_words, argv_word):
                if match_interface(argv_word):
                    value = argv_word
                    value_raw = argv_word
                    matched = True

                elif use_fuzzy and match_interface_fuzzy(argv_word):
                    value = argv_word
                    value_raw = argv_word
                    fuzzy = True
                    matched = True

        elif text == '<number>' or text.startswith('<number-'):

            if match_decimal(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

            elif match_hex(argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text == '<text>' or text.startswith('<text-'):

            if match_text(argv_word) and not is_tab_word(tab_words, argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        elif text == '<vrf>' or text.startswith('<vrf-'):

            if match_vrf(argv_word) and not is_tab_word(tab_words, argv_word):
                value = argv_word
                value_raw = argv_word
                matched = True

        else:
            # <0-65535>
            match_number_range_variable = re_number_range_variable.match(text)

            # <0x0000-0x1234>
            match_hex_range_variable = re_hex_range_variable.match(text)

            if match_number_range_variable:
                min_int = int(match_number_range_variable.group(1))
                max_int = int(match_number_range_variable.group(2))

            elif match_hex_range_variable:
                min_int = int(match_hex_range_variable.group(1), 16)
                max_int = int(match_hex_range_variable.group(2), 16)

            else:
                raise InvalidVariable("network-docopt does not yet support '%s' variables" % text)

            if match_number_in_range(argv_word, min_int, max_int):
                value = argv_word
                value_raw = argv_word
                matched = True

            # The user entered a number it is just out of range
            elif use_fuzzy and match_decimal(argv_word):
                value = argv_word
                value_raw = argv_word
                fuzzy = True
                matched = True

    elif text.isdigit():

        if match_decimal_value(text, argv_word):
            value = argv_word
            value_raw = argv_word
            matched = True

        elif for_tab_complete and match_decimal_value_startswith(text, argv_word):
            value = argv_word
            value_raw = argv_word
            matched = True

    else:
        if match_keyword(text, argv_word):
            value = argv_word
            value_raw = argv_word
            matched = True

        elif match_keyword_startswith(text, argv_word):
            value = text
            value_raw = argv_word
            matched = True

        elif for_keyword_help and match_keyword_contains(text, argv_word):
            value = text
            value_raw = argv_word
            matched = True

        elif use_fuzzy:

            if match_keyword_fuzzy_lower(text, argv_word):
                value = text
                value_raw = argv_word
                fuzzy = True
                matched = True

            elif match_keyword_fuzzy_swap_letter_pairs(text, argv_word):
                value = text
                value_raw = argv_word
                fuzzy = True
                matched = True

            elif match_keyword_fuzzy_drop_each_letter(text, argv_word):
                value = text
                value_raw = argv_word
                fuzzy = True
                matched = True

    return (matched, value, value_raw, fuzzy)



if __name__ == '__main__':
    import doctest

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(filename)16s %(levelname)8s: %(message)s')
    log = logging.getLogger(__name__)

    # Color the errors and warnings in red
    logging.addLevelName(logging.ERROR, "\033[91m   %s\033[0m" % logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.WARNING, "\033[91m %s\033[0m" % logging.getLevelName(logging.WARNING))

    doctest.testmod()
