#!/usr/bin/env python

# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
# Author: Daniel Walton, dwalton@cumulusnetworks.com

import logging
import unittest
from network_docopt.match import match_word, is_loose_matching_variable
from pprint import pformat

log = logging.getLogger(__name__)

# Colors for printing
red    = 91
green  = 92
yellow = 93
blue   = 94


class AmbiguousNode(Exception):
    """
    Raised if a Node has multiple loose matching variables such as (<text>|<wildcard>)
    """
    pass


def lists_have_common_entry(listA, listB):
    """
    lists_have_common_entry(['bart', 'homer', 'lisa', 'marge'], ['marge', 'bart'])
    True

    lists_have_common_entry(['marge', 'bart'], ['bart', 'homer', 'lisa', 'marge'])
    True

    lists_have_common_entry(['bart', 'homer', 'lisa', 'marge'], ['flanders', 'moe'])
    False

    lists_have_common_entry(['flanders', 'moe'], ['bart', 'homer', 'lisa', 'marge'])
    False
    """
    assert isinstance(listA, (list, tuple)), "listA '%s' is %s, it must be a list or tuple" % (pformat(listA), type(listA))
    assert isinstance(listB, (list, tuple)), "listB '%s' is %s, it must be a list or tuple" % (pformat(listB), type(listB))
    return any(x in listA for x in listB)


def string_length_delta(stringA, stringB):
    """
    Return the length delta of stringA vs stringB

    >>> string_length_delta('zelda', 'ganon')
    0

    >>> string_length_delta('zelda', 'dodongo')
    2

    >>> string_length_delta('dodongo', 'zelda')
    2
    """
    return abs(len(stringA) - len(stringB))


def _tree_nodes(node):
    nodes = [node,]

    for child_node in list(node.children.values()):
        nodes.extend(_tree_nodes(child_node))

    return nodes


def tree_nodes(node):
    """
    Recursively walk the tree and return a tuple of all Node objects in the tree
    """
    return tuple(_tree_nodes(node))


# This isn't used for much other than backwards compatibility
class Word(object):

    def __init__(self, text=None):
        self.text = text
        self.value = None
        self.value_raw = None
        self.fuzzy = False

    def __str__(self):
        return self.value


def node_prefer_more_specific_matches(matches):
    """
    Scenario #1 - loose variable vs strict variable
    node_text is (<text>|<ipv4>) and argv_text is 1.1.1.1. <text> will match
    on anything without a whitespace so 1.1.1.1 will match both <ipv4> and
    <text>. We want the end result to be that 1.1.1.1 matched against <ipv4>
    as that is a more specific variable type than <text>.

    >>> node_prefer_more_specific_matches([['<ipv4>', '1.1.1.1', '1.1.1.1', False], ['<text>', '1.1.1.1', '1.1.1.1', False]])
    [('<ipv4>', '1.1.1.1', '1.1.1.1', False)]


    Scenario #2 - loose variable vs keyword
    node_text is (<text>|zelda) and argv_text is 'zel'. We want 'zel' to match
    against 'zelda' since a keyword is more specific than the loose matching
    varible <text>.
    >>> node_prefer_more_specific_matches([['<text>', 'zel', 'zel', False], ['zelda', 'zelda', 'zel', False]])
    [('zelda', 'zelda', 'zel', False)]


    Scenario #3 - strict variable vs keyword
    node_text is (<ipv4>|1.1.1.1) and argv_text is '1.1.1.1'. We want '1.1.1.1' to match
    against '1.1.1.1' since a keyword is more specific than the strict matching
    varible <ipv4>.
    >>> node_prefer_more_specific_matches([['<ipv4>', '1.1.1.1', '1.1.1.1', False], ['1.1.1.1', '1.1.1.1', '1.1.1.1', False]])
    [('1.1.1.1', '1.1.1.1', '1.1.1.1', False)]


    Scenario #4 - multiple keywords match
    # upstream|upstream-join-desired|upstream-rpf
    >>> node_prefer_more_specific_matches([['upstream', 'upstream', 'upstr', False], ['upstream-join-desired', 'upstream-join-desired', 'upstr', False], ['upstream-rpf', 'upstream-rpf', 'upstr', False]])
    [('upstream', 'upstream', 'upstr', False), ('upstream-join-desired', 'upstream-join-desired', 'upstr', False), ('upstream-rpf', 'upstream-rpf', 'upstr', False)]


    Scenario #5 - 'matches' only has one entry
    >>> node_prefer_more_specific_matches([['<ipv4>', '1.1.1.1', '1.1.1.1', False], ])
    [('<ipv4>', '1.1.1.1', '1.1.1.1', False)]
    """

    if len(matches) == 1:
        return [tuple(matches[0])]

    var_strict_matches = []
    var_loose_matches = []
    keyword_matches = []

    # Find how many of the matches were against loose matching variables such as a variable like <ip>, <text>, etc
    for (word_text, value, value_raw, fuzzy) in matches:
        if word_text.startswith('<'):
            if is_loose_matching_variable(word_text):
                var_loose_matches.append((word_text, value, value_raw, fuzzy))

            else:
                var_strict_matches.append((word_text, value, value_raw, fuzzy))
        else:
            keyword_matches.append((word_text, value, value_raw, fuzzy))

    # If we have either of these ignore the loose matches
    if keyword_matches:
        result = keyword_matches

    elif var_strict_matches:

        # If there is more than one var_strict_matches it means the node has
        # something like "(<ipv4>|<bgppeer>)" where both of those will match
        # against an IP address. In this scenario we want to prefer "<bgppeer>"
        # since that is more specific than "<ipv4>".
        #
        # I don't have a great solution here right now...short term will be to
        # look for the few scenarios where we know this happens and handle them. We
        # will raise an exception if it is a scenario we haven't coded support
        # for yet.
        if len(var_strict_matches) > 1:
            result = []
            word_texts = []

            for (word_text, value, value_raw, fuzzy) in var_strict_matches:
                word_texts.append(word_text)

                if word_text == '<bgppeer>':
                    result = [(word_text, value, value_raw, fuzzy),]
                    break
            else:
                raise AmbiguousNode("Node contains multiple variables %s that match '%s'" % (','.join(word_texts), value_raw))

        else:
            result = var_strict_matches

    else:
        if len(var_loose_matches) > 1:

            # If the node has multiple loose matching variables we have no way to
            # choose which should be the one that matches.  For instance if node_text
            # is (<text>|<interface>) and argv_text is 'bond0' we cannot narrow that
            # scenario down to one obvious winner.
            word_texts = []

            for (word_text, _, value_raw, _) in var_loose_matches:
                word_texts.append(word_text)

            log.info("matches:\n%s\n\n" % pformat(matches))
            log.info("var_loose_matches:\n%s\n\n" % pformat(var_loose_matches))
            raise AmbiguousNode("Node contains multiple loose matching variables %s that are matched by %s" % (','.join(word_texts), value_raw))

        else:
            result = var_loose_matches

    return result


def node_calculate_ambiguity_score(word_text, value, value_raw):
    """
    Return the ambiguity score for a node

    >>> node_calculate_ambiguity_score('<0-100>', 37, 37)
    0

    >>> node_calculate_ambiguity_score('<text>', 'zelda', 'zelda')
    99

    >>> node_calculate_ambiguity_score('<ipv4>', '1.1.1.1', '1.1.1.1')
    0

    >>> node_calculate_ambiguity_score('zelda', 'zelda', 'zel')
    2
    """
    assert isinstance(word_text, str), "word_text '%s' is %s, it must be str or unicode" % (pformat(word_text), type(word_text))

    if isinstance(value_raw, int):
        ambiguity_score = 0

    elif word_text.startswith('<'):

        if is_loose_matching_variable(word_text):
            ambiguity_score = 99
        else:
            ambiguity_score = 0

    # keyword
    else:
        ambiguity_score = string_length_delta(value, value_raw)

    return ambiguity_score


def node_all_match_shortest_keyword(matches):
    """
    If a node has choices (upstream|upstream-join-desired|upstream-rpf)
    and the user enters 'upstr' we will have three matches...we should
    - find the shortest word among all choices
    - return False if any of the matches is a var like <ipv4>, etc
    - return True if all of our choices start with the shortest word

    >>> node_all_match_shortest_keyword([('upstream', 'upstream', 'upstr', False), ('upstream-join-desired', 'upstream-join-desired', 'upstr', False), ('upstream-rpf', 'upstream-rpf', 'upstr', False)])
    (True, 'upstream', 'upstream', 'upstr', False)

    Here 'point-to-point' is the shortest match but 'point-to-multipoint'
    does not start with 'point-to-point' so we return False
    >>> node_all_match_shortest_keyword([('point-to-point', 'point-to-point', 'point', False), ('point-to-multipoint', 'point-to-multipoint', 'point', False)])
    (False, None, None, None, None)

    If one of the matches is a variable return False
    >>> node_all_match_shortest_keyword([('upstream', 'upstream', 'upstr', False), ('upstream-join-desired', 'upstream-join-desired', 'upstr', False), ('<text>', 'upstr', 'upstr', False)])
    (False, None, None, None, None)

    If there is only one entry in matches we must return True
    >>> node_all_match_shortest_keyword([('upstream-rpf', 'upstream-rpf', 'upstr', False)])
    (True, 'upstream-rpf', 'upstream-rpf', 'upstr', False)
    """

    # Simple case where there is only one entry
    if len(matches) == 1:
        return tuple([True,] + list(matches[0]))

    all_match_shortest_keyword = True
    shortest_keyword_value_len = None
    shortest_keyword_value = None
    shortest_keyword_value_raw = None
    shortest_keyword_text = None
    shortest_keyword_fuzzy = None

    # First find the keyword that is the shortest match, so 'upstream' in our
    # example above.
    for (word_text, value, value_raw, fuzzy) in matches:
        if word_text.startswith('<'):
            all_match_shortest_keyword = False
            break

        value_len = len(value)

        if shortest_keyword_value_len is None or value_len < shortest_keyword_value_len:
            shortest_keyword_value_len = value_len
            shortest_keyword_value = value
            shortest_keyword_value_raw = value_raw
            shortest_keyword_text = word_text
            shortest_keyword_fuzzy = fuzzy

    # Now verify that all of our choices start with the keyword that is the shortest match
    if all_match_shortest_keyword:
        for (text, value, value_raw, fuzzy) in matches:
            if not value.startswith(shortest_keyword_value):
                all_match_shortest_keyword = False
                break

    if all_match_shortest_keyword:
        return (True, shortest_keyword_text, shortest_keyword_value, shortest_keyword_value_raw, shortest_keyword_fuzzy)
    else:
        return (False, None, None, None, None)


def node_all_start_with_same_text(argv_text, matches):
    """
    >>> node_all_start_with_same_text('upstr', [('upstream', 'upstream', 'upstr', False), ('upstream-join-desired', 'upstream-join-desired', 'upstr', False), ('upstream-rpf', 'upstream-rpf', 'upstr', False)])
    (True, 'upstream', 'upstream', 'upstr', False)
    """
    all_start_with_same_text = True
    shortest_keyword_value_len = None
    shortest_keyword_value = None
    shortest_keyword_value_raw = None
    shortest_keyword_text = None
    shortest_keyword_fuzzy = None

    for (word_text, value, value_raw, fuzzy) in matches:
        if word_text.startswith('<') or not word_text.startswith(argv_text):
            all_start_with_same_text = False
            break

        value_len = len(value)

        if shortest_keyword_value_len is None or value_len < shortest_keyword_value_len:
            shortest_keyword_value_len = value_len
            shortest_keyword_value = value
            shortest_keyword_value_raw = value_raw
            shortest_keyword_text = word_text
            shortest_keyword_fuzzy = fuzzy

    if all_start_with_same_text:
        return (True, shortest_keyword_text, shortest_keyword_value, shortest_keyword_value_raw, shortest_keyword_fuzzy)
    else:
        return (False, None, None, None, None)


def match_node(node_text, argv_text, use_fuzzy, keywords_only, for_tab_complete, for_keyword_help, tab_words):
    """
    Returns the following tuple
        (matched, required, word_text, value, value_raw, fuzzy, ambiguity_score)

    node_text has one keyword and the user enters that exact keyword
    >>> match_node('add', 'add', False, False, False, False, [])
    (True, True, 'add', 'add', 'add', False, 0)

    node_text has one keyword and the user enters part of the keyword
    >>> match_node('add', 'ad', False, False, False, False, [])
    (True, True, 'add', 'add', 'ad', False, 1)

    node_text has one keyword, the user typoes but use_fuzzy is False
    >>> match_node('add', 'dad', False, False, False, False, [])
    (False, True, 'add', None, None, False, 0)

    node_text has one keyword, the user typoes but use_fuzzy is True
    >>> match_node('add', 'dad', True, False, False, False, [])
    (True, True, 'add', 'add', 'dad', True, 0)

    node_text has two keywords (required) and the user enters that exact keyword
    >>> match_node('(add|del)', 'add', False, False, False, False, [])
    (True, True, 'add', 'add', 'add', False, 0)

    node_text has two keywords (required) and the user enters part of the keyword
    >>> match_node('(add|del)', 'ad', False, False, False, False, [])
    (True, True, 'add', 'add', 'ad', False, 1)

    node_text has two keywords (optional) and the user enters that exact keyword
    >>> match_node('[add|del]', 'add', False, False, False, False, [])
    (True, False, 'add', 'add', 'add', False, 0)

    node_text has two keywords (optional) and the user enters part of the keyword
    >>> match_node('[add|del]', 'ad', False, False, False, False, [])
    (True, False, 'add', 'add', 'ad', False, 1)
    """

    assert isinstance(argv_text, str), "argv_text '%s' is %s, it must be str or unicode" % (pformat(argv_text), type(argv_text))

    ambiguity_score = 0
    fuzzy = None
    required = False
    matched = False
    value = None
    value_raw = None
    word_text = None
    matches = []

    # Remove the [] or () and note if this node is required.
    # Only nodes with [] are not required.
    if node_text.startswith('['):
        required = False
        node_text = node_text[1:-1]

    elif node_text.startswith('('):
        required = True
        node_text = node_text[1:-1]

    else:
        required = True

    # Split the 'node_text' by | and evaluate each 'word_text' against 'argv_text'.
    # Store the comparisons that are a match in 'matches'.
    for word_text in node_text.split('|'):
        if keywords_only:
            if not word_text.startswith('<'):
                (matched, value, value_raw, fuzzy) = match_word(word_text, argv_text, use_fuzzy, for_tab_complete, for_keyword_help, tab_words)

                if matched:
                    matches.append((word_text, value, value_raw, fuzzy))
        else:
            (matched, value, value_raw, fuzzy) = match_word(word_text, argv_text, use_fuzzy, for_tab_complete, for_keyword_help, tab_words)

            if matched:
                matches.append((word_text, value, value_raw, fuzzy))

    if matches:
        matches = node_prefer_more_specific_matches(matches)

        # One or more 'word_text' matched
        (all_match_shortest_keyword, word_text, value, value_raw, fuzzy) = node_all_match_shortest_keyword(matches)

        if all_match_shortest_keyword:
            ambiguity_score = node_calculate_ambiguity_score(word_text, value, value_raw)
            return (True, required, word_text, value, value_raw, fuzzy, ambiguity_score)

        # All start with same text
        # Only do this if we are being called for TAB complition
        if for_tab_complete:
            (all_start_with_same_text, word_text, value, value_raw, fuzzy) = node_all_start_with_same_text(argv_text, matches)

            if all_start_with_same_text:
                ambiguity_score = node_calculate_ambiguity_score(word_text, value, value_raw)
                return (True, required, word_text, value, value_raw, fuzzy, ambiguity_score)

    return (False, required, word_text, value, value_raw, fuzzy, ambiguity_score)


def get_node_text_slow(text):
    """
    Most of the time whitespace can be used to split nodes
        (set|clear) <interface> clagd-enable (no|yes)

    nodes
    - (set|clear)
    - <interface>
    - clagd-enable
    - (no|yes)

    But if we are dealing with multiword keywords, such as "soft in", that approach
    does not work. We can only split on whitespaces if we are not inside a () or []
        bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]

    nodes:
    - bgp
    - (<ipv4>|<ipv6>|<interface>|*)
    - [soft in|soft out]

    In this scenario we go character by character so it is slower
    than get_node_text_fast() where it splits based on whitespace.

    >>> get_node_text_slow('(set|clear) <interface> clagd-enable (no|yes)')
    ('(set|clear)', '<interface>', 'clagd-enable', '(no|yes)')

    >>> get_node_text_slow('bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]')
    ('bgp', '(<ipv4>|<ipv6>|<interface>|*)', '[soft in|soft out]')
    """
    nodes = []
    node_index = 0
    node_text = []
    parens = 0
    brackets = 0

    for char in text:
        if char == ' ':
            if parens == 0 and brackets == 0:
                nodes.append(''.join(node_text))
                node_index += 1
                node_text = []
            else:
                node_text.append(char)
        else:
            if char == '(':
                parens += 1

            elif char == ')':
                parens -= 1

            elif char == '[':
                brackets += 1

            elif char == ']':
                brackets -= 1

            node_text.append(char)

    if node_text:
        nodes.append(''.join(node_text))

    return tuple(nodes)


def get_node_text_fast(text):
    """
    If the command does not have nodes with whitespace within the node then we
    can split on whitespace which is nice and fast.

    Return None if we find a node that contains whitespaces.

    >>> get_node_text_fast('(set|clear) <interface> clagd-enable (no|yes)')
    ('(set|clear)', '<interface>', 'clagd-enable', '(no|yes)')

    >>> get_node_text_fast('bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]') is None
    True
    """
    nodes = []
    node_index = 0

    for node_text in text.split():

        if node_text.startswith('[') and not node_text.endswith(']'):
            return None

        if node_text.startswith('(') and not node_text.endswith(')'):
            return None

        nodes.append(node_text)
        node_index += 1

    return tuple(nodes)


def get_node_text(text):
    """
    >>> get_node_text('(set|clear) <interface> clagd-enable (no|yes)')
    ('(set|clear)', '<interface>', 'clagd-enable', '(no|yes)')

    >>> get_node_text('bgp (<ipv4>|<ipv6>|<interface>|*) [soft in|soft out]')
    ('bgp', '(<ipv4>|<ipv6>|<interface>|*)', '[soft in|soft out]')
    """
    nodes = get_node_text_fast(text)

    if nodes is None:
        nodes = get_node_text_slow(text)

    return nodes


def expand_nodes_to_lines(nodes):
    """
    Given a list of nodes, expand them into the lines needed to build a tree
    for the various combinations that can be made from those nodes.

    >>> expand_nodes_to_lines(['zelda', ['ipv4 <ipv4>', 'ipv6 <ipv6>'], ['vrf <interface>', '']])
    ['zelda ipv4 <ipv4> vrf <interface>', 'zelda ipv4 <ipv4>', 'zelda ipv6 <ipv6> vrf <interface>', 'zelda ipv6 <ipv6>']

    >>> expand_nodes_to_lines(['zelda', '<acl-ipv4>', ['priority <number-range-list>', '']])
    ['zelda <acl-ipv4> priority <number-range-list>', 'zelda <acl-ipv4>']

    >>> expand_nodes_to_lines(['zelda', '<acl-ipv4>', ['priority <number-range-list>']])
    ['zelda <acl-ipv4> priority <number-range-list>']

    >>> expand_nodes_to_lines(['zelda', 'add', '[<ipv6>|all]', 'cl-support'])
    ['zelda add [<ipv6>|all] cl-support']

    >>> expand_nodes_to_lines([])
    ['']
    """
    lines = []
    line = []

    for (node_index, node) in enumerate(nodes):
        if isinstance(node, str):
            line.append(node)
        else:

            # 'node' is a list of strings
            for x in node:
                for rest_of_line in expand_nodes_to_lines(nodes[node_index+1:]):
                    # compress duplicate whitespaces
                    lines.append(' '.join(line + [x, ] + [rest_of_line, ]).strip().replace('  ', ' '))
            break
    else:
        line = ' '.join(line)
        lines.append(line)

    return lines


def expand_command_to_multiple_commands(command):
    """
    Originally I tried putting nodes on the tree that contained multiple
    keywords ('ipv4 unicast')...this made TAB completion a nightmare though.
    It is much cleaner to expand these options out and build a bigger tree.

    For 'show ip bgp [ipv4 unicast|ipv4 multicast|ipv6 unicast] summary'
    Expand this to the four lines that will cover each scenario

        show ip bgp ipv4 multicast summary
        show ip bgp ipv4 unicast summary
        show ip bgp ipv6 unicast summary
        show ip bgp summary

    For 'show ip bgp (ipv4 unicast|ipv4 multicast|ipv6 unicast) summary'
    Expand this out to the three lines that will cover each scenario

        show ip bgp ipv4 multicast summary
        show ip bgp ipv4 unicast summary
        show ip bgp ipv6 unicast summary

    Normal scenario where none of the Nodes contain a multi-word option
    >>> expand_command_to_multiple_commands('nes show zelda')
    ['nes show zelda']

    Scenarios where a Node has a required multi-word choice
    >>> expand_command_to_multiple_commands('nes show (mario hat|zelda sword)')
    ['nes show mario hat', 'nes show zelda sword']

    >>> expand_command_to_multiple_commands('show ip bgp (ipv4 unicast|ipv4 multicast|ipv6 unicast) summary')
    ['show ip bgp ipv4 unicast summary', 'show ip bgp ipv4 multicast summary', 'show ip bgp ipv6 unicast summary']

    Scenarios where a Node has an optional multi-word choice
    >>> expand_command_to_multiple_commands('show ip bgp [ipv4 unicast|ipv4 multicast|ipv6 unicast] summary')
    ['show ip bgp ipv4 unicast summary', 'show ip bgp ipv4 multicast summary', 'show ip bgp ipv6 unicast summary', 'show ip bgp summary']

    >>> expand_command_to_multiple_commands('nes load mario [mustache trimmer]')
    ['nes load mario mustache trimmer', 'nes load mario']
    """
    program = None
    nodes = []

    for node_text in get_node_text(command):
        if not program:
            program = node_text

        if ' ' in node_text:

            # Optional multi-word choice
            if node_text.startswith('['):
                nodes.append(node_text[1:-1].split('|') + [''])

            # Required multi-word choice
            elif node_text.startswith('('):
                nodes.append(node_text[1:-1].split('|'))

            else:
                raise Exception("'%s' has a space but no [] or ()" % node_text)

        # No spaces in the node text, add it as is
        else:
            nodes.append(node_text)

    lines = expand_nodes_to_lines(nodes)

    return lines


def get_first_ENTER_ok_node_index(node_texts):
    """
    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario'])
    2

    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario', '[belt]'])
    2

    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario', '[belt|hat]'])
    2

    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario', '(belt|hat)'])
    3

    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario', '[belt|hat] [mustache]'])
    2

    >>> get_first_ENTER_ok_node_index(['nes', 'load', 'mario', '[belt|hat]', 'mustache'])
    4
    """

    last_ENTER_ok_node_index = None

    for (node_index, node_text) in enumerate(node_texts):
        if not node_text.startswith('['):
            last_ENTER_ok_node_index = node_index

    return last_ENTER_ok_node_index


class Node(object):
    count = 0

    def __init__(self, depth, parent, text):
        assert isinstance(text, str), "text '%s' is %s, it must be str or unicode" % (pformat(text), type(text))
        self.depth = depth
        self.parent = parent
        self.children = {}
        self.tab_words = []
        self.ENTER_ok = False
        self.text = text
        Node.count += 1

        # [] means the Node is optional
        if text.startswith('['):
            self.required = False
            self.words = tuple(text[1:-1].split('|'))

        elif text.startswith('('):
            assert '|' in text, "Nodes with () must have multiple options such as (foo|bar), this node has %s" % text
            self.required = True
            self.words = tuple(text[1:-1].split('|'))

        else:
            self.required = True
            self.words = tuple(text.split('|'))

        self.reset()

    def __str__(self):
        return self.text

    def reset(self):
        """
        reset all variables that are populated via Node.match()
        """
        self.match_text = None
        self.match_value = None
        self.match_value_raw = None
        self.match_word_text = None
        self.match_fuzzy = False
        self.value = None
        self.ambiguity_score = 0

    def match(self, argv_text, use_fuzzy, keywords_only, for_tab_complete):
        (matched,
         required,
         match_text,
         match_value,
         match_value_raw,
         match_fuzzy,
         ambiguity_score) =\
            match_node(self.text, argv_text, use_fuzzy, keywords_only, for_tab_complete, False, self.tab_words)

        if matched:
            if match_text.startswith('<wildcard'):
                if self.match_value is not None:
                    self.match_value += ' ' + match_value
                else:
                    self.match_value = match_value

                if self.match_value_raw is not None:
                    self.match_value_raw += ' ' + match_value_raw
                else:
                    self.match_value_raw = match_value_raw
            else:
                self.match_value = match_value
                self.match_value_raw = match_value_raw

            self.match_text = match_text
            self.match_fuzzy = match_fuzzy
            self.ambiguity_score = ambiguity_score
            self.match_word_text = self.match_text

            # We no longer use the Word() class for matching but when we did we
            # would store the Word object that matched as the Node.value. We do
            # that here for backwards compatibility.
            word = Word()
            word.text = self.match_text
            word.value = self.match_value
            word.value_raw = self.match_value_raw
            word.fuzzy = self.match_fuzzy
            self.value = word
            return True

        else:
            self.reset()
            return False

    def has_word(self, word):
        """
        Return True if 'word' is in our list of words
        """
        return word in self.words

    def has_child(self, text):
        """
        Return True if we have a child Node for 'text'
        """
        return text in self.children

    def find_child(self, text):
        """
        Return the child Node for 'text' or None if we do not have that child
        """
        return self.children.get(text)

    def add_child(self, child_node):
        """
        Add a child Node
        """
        self.children[child_node.text] = child_node

    def find_node(self, text):
        """
        Return Node for 'text'
        """
        if self.text == text:
            return self
        else:
            for child_node in list(self.children.values()):
                node = child_node.find_node(text)

                if node is not None:
                    return node

            return None

    def leaf_node(self):
        """
        Return True if this node is a leaf node
        """
        return not self.children

    def root_node(self):
        """
        Return True if this node is the root node
        """
        return self.depth == 0

    def add_branch(self, command):
        assert self.root_node(), "add_branch can only be called for the root node"

        # Compress multiple spaces down to a single space
        command = ' '.join(command.strip().split())

        # If the line is blank return
        if not command:
            return

        # Remove spaces that are not needed
        command = command.replace(' | ', '|').replace('| ', '|').replace(' |', '|').replace('( ', '(').replace(' )', ')').replace('[ ', '[').replace(' ]', ']')

        commands = expand_command_to_multiple_commands(command)

        for command in commands:
            assert command.count('[') == command.count(']'), 'mis-matched []s in "%s"' % command
            assert command.count('(') == command.count(')'), 'mis-matched ()s in "%s"' % command
            parent = self
            depth = 1

            node_texts = get_node_text(command)
            first_ENTER_ok_node_index = get_first_ENTER_ok_node_index(node_texts)

            # The first node_text is the program name, ignore it
            for (node_index, node_text) in enumerate(node_texts):

                if node_index == 0:
                    continue

                # Does our parent node already have a child with this node_text?
                child_node = parent.find_child(node_text)

                if child_node:
                    parent = child_node

                    if node_index >= first_ENTER_ok_node_index:
                        child_node.ENTER_ok = True

                else:
                    child_node = Node(depth, parent, node_text)

                    if node_index >= first_ENTER_ok_node_index:
                        child_node.ENTER_ok = True

                    # Link this newly created child node to its parent
                    parent.add_child(child_node)
                    parent = child_node

                depth += 1

    def parent_nodes(self):
        """
        Return a tuple of our parent Nodes. These will be sorted starting
        from the top of the tree.
        """
        result = []
        node = self

        while node is not None:
            node = node.parent

            if node is not None:
                result.append(node)

        return tuple(reversed(result))

    def parent_nodes_matching_only(self):
        """
        Return a tuple of our parent Nodes that matched against argv. These
        will be sorted starting from the top of the tree.
        """
        result = []
        node = self

        while node is not None:
            node = node.parent

            if node is not None and node.match_value is not None:
                result.append(node)

        return tuple(reversed(result))

    def child_nodes(self):
        """
        Return a list of our child nodes (recursive)
        """
        result = []

        for child_node in list(self.children.values()):
            result.append(child_node)
            result.extend(child_node.child_nodes())

        return result

    def ascii(self, color=None):
        """
        Used to print this Node in an ascii tree
        """
        #extra_info = " -- leaf %s, required %s, last required %s" % (self.leaf_node(), self.required, self.ENTER_ok)
        extra_info = ""

        if color is None:
            return "%s%s%s" % ('    ' * self.depth, self.text, extra_info)
        else:
            return "%s\033[%dm%s\033[0m%s" % ('    ' * self.depth, color, self.text, extra_info)

    def has_optional_branch(self):
        if self.children:
            for child_node in list(self.children.values()):
                if child_node.has_optional_branch():
                    return True

            return False
        else:
            if self.required:
                return False
            else:
                return True

    def _populate_tab_words(self):
        """
        nes show [dodongo|zelda] [json]
        nes show (mario|luigi) json
        """
        tab_words = []

        for child_node in list(self.children.values()):
            tab_words.extend(child_node.words)

            # If the child is optional we must also include the choices for its children
            if not child_node.required:
                tab_words.extend(child_node._populate_tab_words())

        return tab_words

    def populate_tab_words(self):
        self.tab_words = tuple(sorted(self._populate_tab_words()))

    def command(self):
        """
        Traverse our parents to recontruct the command that birthed us
        """
        result = []
        node = self

        while node is not None:
            if node.match_fuzzy:
                result.append("\033[%dm%s\033[0m" % (blue, node.text))
            else:
                result.append(node.text)

            node = node.parent

        return ' '.join(reversed(result))

    def command_did_you_mean(self):
        """
        Traverse our parents to recontruct the command that birthed us
        - highlight nodes that fuzzy matched (print them in blue)
        - add output about what the command was looking for
        """
        result = ['    ' + self.command(), ]
        node = self

        vars_ip = ('<ip>', '<ipv4>', '<ipv6>')
        vars_ip_prefixlen = ('<ip/prefixlen>', '<ipv4/prefixlen>', '<ipv6/prefixlen>')
        vars_bgp_peer = ('<bgppeer>', )
        vars_aspath_list = ('<as-path-access-list>', )
        vars_community_list = ('<community-list>',)
        vars_extcommunity_list = ('<extcommunity-list>',)
        vars_large_community_list = ('<large-community-list>',)
        vars_prefix_list = ('<prefix-list>',)
        vars_route_map = ('<route-map>',)
        vars_interface = ('<interface>',)

        while node is not None:
            if node.match_fuzzy:
                add_newline = False

                if lists_have_common_entry(node.words, vars_ip):
                    result.append("        This command is looking for an IP address, not an IP/prefixlen")
                    add_newline = True

                elif lists_have_common_entry(node.words, vars_ip_prefixlen):
                    result.append("        This command is looking for an IP/prefixlen, not an IP address")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_bgp_peer):
                    result.append("        This command is looking for an existing BGP peer")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_aspath_list):
                    result.append("        This command is looking for an existing as-path access-list")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_community_list):
                    result.append("        This command is looking for an existing community-list")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_extcommunity_list):
                    result.append("        This command is looking for an existing extcommunity-list")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_large_community_list):
                    result.append("        This command is looking for an existing large-community-list")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_prefix_list):
                    result.append("        This command is looking for an existing prefix-list")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_route_map):
                    result.append("        This command is looking for an existing route-map")
                    add_newline = True

                if lists_have_common_entry(node.words, vars_interface):
                    result.append("        Interface names can be a maximum of 15 characters")
                    add_newline = True

                if add_newline:
                    result.append('')

            node = node.parent

        return '\n'.join(result)

    def command_ambiguity_score(self):
        """
        Return the total ambiguity_score of this node and all of our parent nodes
        """
        total_ambiguity_score = 0
        node = self

        while node is not None:
            total_ambiguity_score += node.ambiguity_score
            node = node.parent

        return total_ambiguity_score

    def command_populate_args(self, args):
        """
        Populate the args dictionary using this node and all of our
        parent nodes that matched
        """
        assert isinstance(args, dict), "args '%s' is %s, it must be a dict" % (pformat(args), type(args))
        node = self

        while node is not None:

            if node.match_value is not None:
                if isinstance(node.match_value, str) and node.match_value.isdigit():
                    match_value = int(node.match_value)
                else:
                    try:
                        match_value = str(node.match_value)
                    except UnicodeEncodeError:
                        match_value = node.match_value.encode('utf-8')

                # If a command uses multiple of the same variable:
                #   nes load mario <1-10> zelda <1-10>
                #
                # and the user enters
                #   nes load mario 2 zelda 6
                #
                # we convert the entry in args to a list to store all of the results
                if node.match_word_text in args:
                    if args[node.match_word_text] != match_value:
                        if isinstance(args[node.match_word_text], list):
                            args[node.match_word_text].append(match_value)
                        else:
                            args[node.match_word_text] = [args[node.match_word_text], match_value]
                else:
                    args[node.match_word_text] = match_value

            node = node.parent

        # Continuing with our `nes load mario 2 zelda 6` example, the value of
        # args['<1-20>'] will be [6, 2], the list is backwards because we start with
        # the node at the end of the command and work our way up the tree to the root.
        #
        # Reverse any lists to rememdy this.
        for (key, value) in list(args.items()):
            if isinstance(value, list):
                args[key] = list(reversed(value))

    def command_has_keywords(self, keywords):
        """
        Return True if our parent command contains all 'keywords'
        """
        assert self.leaf_node(), "command_has_keywords() must only be called on leaf nodes"
        used_nodes = set()

        for keyword in keywords:
            node = self

            while node is not None:

                if node not in used_nodes:
                    (matched, required, word_text, value, value_raw, fuzzy, ambiguity_score) =\
                        match_node(node.text, keyword, use_fuzzy=False, keywords_only=True, for_tab_complete=False, for_keyword_help=True, tab_words=node.tab_words)

                    if matched:
                        used_nodes.add(node)
                        break

                node = node.parent

            # This command does not contain 'keyword'
            else:
                return False

        return True

    def ambiguous_match(self, last_argv_word):
        (matched_for_tab_complete, _, _, _, _, _, _) =\
            match_node(self.text, last_argv_word, use_fuzzy=False, keywords_only=False, for_tab_complete=True, for_keyword_help=False, tab_words=self.tab_words)

        (matched_regular, _, _, _, _, _, _) =\
            match_node(self.text, last_argv_word, use_fuzzy=False, keywords_only=False, for_tab_complete=False, for_keyword_help=False, tab_words=self.tab_words)

        return matched_for_tab_complete and not matched_regular


class TestNode(unittest.TestCase):

    def setUp(self):
        self.root = Node(0, None, "nes")

        commands = """
            nes show [dodongo|zelda] [json]
            nes show (mario|luigi) json
            nes add (mario|martian|maverick) mustache
            nes add zelda sword
            nes del (mario hat|mario mustache|zelda sword) json
        """

        for command in commands.splitlines():
            self.root.add_branch(command)

        # Build a tuple of nodes and leaf nodes
        self.nodes = tree_nodes(self.root)
        self.leaf_nodes = tuple([node for node in self.nodes if node.leaf_node()])

    def test_root_node(self):
        self.assertTrue(self.root.root_node())

    def test_reset(self):
        self.root.match_text = 'zelda'
        self.assertEqual(self.root.match_text, 'zelda')

        self.root.reset()
        self.assertEqual(self.root.match_text, None)

    def test_find_node(self):
        mustache = self.root.find_node('mustache')
        self.assertTrue(mustache)

    def test_leaf_node(self):
        mustache = self.root.find_node('mustache')
        self.assertTrue(mustache.leaf_node())
        self.assertFalse(self.root.leaf_node())

    def test_required(self):
        show = self.root.find_node('show')
        dodongo_zelda = self.root.find_node('[dodongo|zelda]')
        mario_luigi = self.root.find_node('(mario|luigi)')

        self.assertTrue(show.required)
        self.assertTrue(mario_luigi.required)
        self.assertFalse(dodongo_zelda.required)

    def test_has_word(self):
        mmm = self.root.find_node('(mario|martian|maverick)')
        self.assertTrue(mmm.has_word('mario'))
        self.assertTrue(mmm.has_word('martian'))
        self.assertTrue(mmm.has_word('maverick'))
        self.assertFalse(mmm.has_word('zelda'))

    def test_has_child(self):
        zelda = self.root.find_node('zelda')
        self.assertTrue(zelda.has_child('sword'))
        self.assertFalse(zelda.has_child('rocket-launcher'))

    def test_find_child(self):
        zelda = self.root.find_node('zelda')
        self.assertTrue(zelda.find_child('sword'))
        self.assertEqual(zelda.find_child('rocket-launcher'), None)

    def test_parent_nodes(self):
        """
        nes add zelda sword
        """
        sword = self.root.find_node('sword')
        sword_parents = sword.parent_nodes()

        self.assertEqual(sword_parents[0], self.root)
        self.assertEqual(sword_parents[1], self.root.find_node('add'))
        self.assertEqual(sword_parents[2], self.root.find_node('zelda'))
        self.assertEqual(len(sword_parents), 3)

    def test_child_nodes(self):
        """
        nes show [dodongo|zelda] [json]
        nes show (mario|luigi) json
        """
        show = self.root.find_node('show')
        show_children = show.child_nodes()
        dodongo_zelda = self.root.find_node('[dodongo|zelda]')
        mario_luigi = self.root.find_node('(mario|luigi)')
        mustache = self.root.find_node('mustache')

        self.assertTrue(dodongo_zelda in show_children)
        self.assertTrue(mario_luigi in show_children)
        self.assertFalse(mustache in show_children)

    def test_populate_tab_words(self):
        """
        nes show [dodongo|zelda] [json]
        nes show (mario|luigi) json
        nes add (mario|martian|maverick) mustache
        nes add zelda sword
        nes del (mario hat|mario mustache|zelda sword) json
        """
        show = self.root.find_node('show')
        add = self.root.find_node('add')
        delete = self.root.find_node('del')
        self.assertEqual(show.tab_words, [])
        self.assertEqual(add.tab_words, [])

        for node in self.nodes:
            node.populate_tab_words()

        self.assertEqual(show.tab_words, ('dodongo', 'json', 'luigi', 'mario', 'zelda'))
        self.assertEqual(add.tab_words, ('mario', 'martian', 'maverick', 'zelda'))
        self.assertEqual(delete.tab_words, ('mario', 'zelda'))

    def test_ENTER_ok(self):
        """
        nes show [dodongo|zelda] [json]
        nes show (mario|luigi) json
        """
        show = self.root.find_node('show')
        dodongo_zelda = self.root.find_node('[dodongo|zelda]')
        mario_luigi = self.root.find_node('(mario|luigi)')

        self.assertTrue(show.ENTER_ok)
        self.assertFalse(dodongo_zelda.ENTER_ok)
        self.assertFalse(mario_luigi.ENTER_ok)

        self.assertTrue(show.ENTER_ok)
        self.assertFalse(dodongo_zelda.ENTER_ok)
        self.assertFalse(mario_luigi.ENTER_ok)

    def test_command(self):
        sword = self.root.find_node('sword')
        self.assertEqual(sword.command(), 'nes add zelda sword')

        mustache = self.root.find_node('mustache')
        self.assertEqual(mustache.command(), 'nes add (mario|martian|maverick) mustache')

    def test_command_has_keywords(self):
        sword = self.root.find_node('sword')

        self.assertTrue(sword.command_has_keywords(('zeld', 'ad')))
        self.assertTrue(sword.command_has_keywords(('zelda', 'add')))
        self.assertFalse(sword.command_has_keywords(('mario', 'add')))


if __name__ == '__main__':
    import doctest

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(filename)16s %(levelname)8s: %(message)s')
    log = logging.getLogger(__name__)

    # Color the errors and warnings in red
    logging.addLevelName(logging.ERROR, "\033[91m   %s\033[0m" % logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.WARNING, "\033[91m %s\033[0m" % logging.getLevelName(logging.WARNING))

    doctest.testmod()

    test_suites_to_run = []
    test_suites_to_run.append(TestNode)

    for test_suite in test_suites_to_run:
        suite = unittest.TestLoader().loadTestsFromTestCase(test_suite)
        unittest.TextTestRunner(verbosity=2).run(suite)

